# Copyright (c) 2019-2020  University of Houston

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/

from __future__ import absolute_import, print_function

import argparse
import glob
import logging
import sys
import subprocess

from datetime import datetime as dt
from operator import itemgetter
from tempfile import NamedTemporaryFile as ntf

import dateutil.parser
import requests
import six

from . import config
from .json import dumps, loads


def getUsers(course, instructors):
  url = '%s/course/%s/users?instructors=%s' % (config.LABSRV, course, instructors)
  users = requests.get(url).json()['value']
  return users


def getAttempt(exercise_type, euuid, user_urn, index):
  url = '%s/%s/%s/user/%s/attempt/%d' % (
      config.LABSRV, exercise_type, euuid, user_urn, index
  )
  res = requests.get(url)
  attempt = res.json()['value']
  return loads(attempt)


def getAttemptUntil(exercise_type, euuid, user_urn, due=None):
  if isinstance(due, (six.text_type, )):
    due = dateutil.parser.parse(due)

  url = '%s/%s/%s/user/%s/attempts/' % (config.LABSRV, exercise_type, euuid, user_urn)
  files = requests.get(url).json()['value']

  logging.debug('%s has %d attempts', user_urn, len(files))

  index = None
  if not files:
    return index

  if not due:
    index = files[-1][0]

  files = sorted(files, key=itemgetter(1))
  for i, ts in files:
    ts = dt.fromtimestamp(float(ts))
    if ts > due:
      break
    index = i

  if index is None:
    return index

  return getAttempt(exercise_type, euuid, user_urn, index)


def getModule(course, toponame):
  url = '%s/course/%s/modules/all' % (config.LABSRV, course)
  mods = requests.get(url).json()['value']
  for modid in mods:
    url = '%s/module/%s' % (config.LABSRV, modid)
    mod = requests.get(url).json()['value']
    if mod['topo-name'] == toponame:
      return mod


def loadFiles(pattern):
  files = glob.glob(pattern)
  if not files:
    return None
  fname = max(files)
  with open(fname) as f:
    return f.read()


def getDataPool(datapool, index, use_vlans):
  url = '%s/datapool/%s/get/%d' % (config.LABSRV, datapool, index)
  data = requests.get(url).json()['value']
  data['use-vlans'] = use_vlans
  return data


def getAnswerKey(akeyname):
  url = '%s/answer-key/name/%s' % (config.LABSRV, akeyname)
  res = requests.get(url).json()
  if 'value' not in res or res["value"] == []:
    logging.error('No Answer Key')
    sys.exit(-6)
  return res['value']


def loadData(xrcs, user_urn, attempt):
  data = {}
  for qid, val in enumerate(xrcs['question-data'], 1):
    qid = '%s' % (qid)
    if 'data-pool' in val:
      use_vlans = val.get('use-vlans', False)
      if "index" in attempt[qid]:
        data[qid] = getDataPool(val["data-pool"], attempt[qid]["index"], use_vlans)
      else:
        data[qid] = getDataPool(val["data-pool"], attempt[qid][0]["_uidata"]["index"], use_vlans)
    else:
      mod = getModule(xrcs['course'], xrcs['lab-topo-names'])
      data[qid] = {}
      data[qid]['listresources'] = loadFiles('%s/%s/%s-manifest*' %
                                             (config.FOLDER, mod['uuid'], user_urn))
      data[qid]['sliverstatus'] = loadFiles('%s/%s/%s-status*' %
                                            (config.FOLDER, mod['uuid'], user_urn))
  return data


def evalQuestions(data, keyTemplate, attempt):
  res = requests.post(config.GRADER, json={
      "data": data,
      "answerkey": keyTemplate,
      "attempt": attempt})
  res = res.json()
  if res["retcode"] != 0:
    logging.error(res["msg"])
  return res['value']


def invokeTemp(detail):
  with ntf(prefix='key-', suffix=".json") as tfk, ntf(prefix='att-', suffix=".json") as tfa:
    key = dumps(detail['key'], sort_keys=True, indent=2)
    tfk.write(key)
    tfk.flush()
    ans = dumps(detail['ans'], sort_keys=True, indent=2)
    tfa.write(ans)
    tfa.flush()

    subprocess.call(["vimdiff", tfk.name, tfa.name])


def gradeAll(course, toponame, users, date, hardstop, exam=False):
  result = {}

  if exam:
#    url = '%s/course/%s/exam/foo' % (config.LABSRV, course)
    url = '%s/course/name/%s/exams' % (config.LABSRV, course)
    if date:
      xrcs = requests.post(url, json = {"admin" : True, "date" : date}).json()['value']
    else:
#      exams = requests.post(url, json = {"admin" : True}).json()['value'] # 4421 exam01
      exams = requests.get(url).json()['value']
      for exam in exams:
        if exam['lab-topo-names'] == toponame:
          xrcs = exam
    exercise_type = 'exam'

  else:
    url = '%s/course/%s/exercise/%s' % (config.LABSRV, course, toponame)
    xrcs = requests.get(url).json()['value']
    exercise_type = 'exercise'

  if not xrcs:
    logging.error('No question file for %s' % exercise_type)
    sys.exit(-7)

  euuid = xrcs['uuid']

  if 'akey-name' not in xrcs:
    logging.error('Exercise doesn\'t have answer key name field!')
    sys.exit(-8)

  keyTemplate = getAnswerKey(xrcs['akey-name'])

  for user_urn in users:
    if hardstop:
      attempt = getAttemptUntil(exercise_type, euuid, user_urn, dateutil.parser.parse(hardstop))
    else:
      attempt = getAttemptUntil(exercise_type, euuid, user_urn, xrcs['hard-stop'])

    if not attempt:
      result[user_urn] = None
      continue
    data = loadData(xrcs, user_urn, attempt)
    try:
        result[user_urn] = evalQuestions(data, keyTemplate, attempt)
    except ValueError as e: # Stubax: this is to address issue grading 4421 routers exercise 10-2020
        print("error getting grade for user: %s" % user_urn)
        print(e)

  return result


def loadUsernameCSV(path):
  with open(path, "r") as f:
    users = {}
    for line in f:
      l = line.split(',')
      username = l[2].strip('\n')
      users[username] = {"first" : l[1], "last" : l[0]}

  return users


def namePadding(users):
  first_max = 0
  last_max = 0
  combined_max = 0
  uname_max = 0
  for username, user in users.items():
    if len(user["first"]) > first_max:
      first_max = len(user["first"])
    if len(user["last"]) > last_max:
      last_max = len(user["last"])
    if len(username) > uname_max:
      uname_max = len(username)
    if len(user["first"]) + len(user["last"]) > combined_max:
      combined_max = len(user["first"]) + len(user["last"]) 
  return (first_max, last_max, combined_max, uname_max)


def wordPadding(words):
  longest = 0
  for word in words:
    if len(word) > longest:
      longest = len(word)
  return longest


def printExerciseInfo(course, dates = False, exam = False):
  if exam:
    url = "%s/course/name/%s/exams" % (config.LABSRV, course)
    exes = requests.get(url).json()["value"]
    print("Exams for course: %s" % course)
  else:
    import grader.labserve
    lsp = grader.labserve.LabserveProxy(course)
    exes = lsp.exercises
    print("Exercise for course: %s" % course)
  
  exercises = {}
  names = []
  for ex in exes:
    name = ex["lab-topo-names"]
    if isinstance(name, list):
      short_name = str(name[0])
    else:
      short_name = str(name)
    names.append(short_name)
    ex["short-name"] = short_name
    exercises[ex["hard-start"]] = ex

  padding = max(wordPadding(names), 8)
  if dates:
    line = "+" + "-" * (padding + 2) + "+" + "-" * 18 + "+" + "-" * 18 + "+"
    print(line)
    print("| %s | %s | %s |" %(str.ljust("Ex. Name", padding), str.ljust("Start", 16), str.ljust("Stop", 16)))
    print(line)
  for date in sorted(list(exercises.keys())):
    ex = exercises[date]
    if dates:
      print("| %s | %s | %s |" % (str.ljust(ex["short-name"], padding), str(ex["hard-start"]), str(ex["hard-stop"])))
    else:
      print("%s" % ex["lab-topo-names"])
  if dates:
    print(line)


def parseArgs():
  parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

  parser.add_argument("--name", dest="name", type=str, help="Exercise name")
  parser.add_argument("--exam", dest="exam", action='store_true', default=False, 
                      help="If grading an exam")
  parser.add_argument("--date", dest="date", type=str, default=None, 
          help="Use with --exam to set the 'perspective date.' Format: YYYYMMDDHHmm")
  parser.add_argument("--hard-stop", dest="hardstop", type=str, default=None, help="Expected format: YYYY-MM-DD HH:mm")
  parser.add_argument("--course", dest="course", type=str, default='cn4421', 
                      help="Course name (cn4421, an4325, ns6313, etc.)")
  parser.add_argument("--instructors", dest="instructors", action='store_true')
  parser.add_argument("--graderurl", dest="grader", type=str, default=None)
  parser.add_argument("--stunames", dest="stunamespath", type=str, default=None, 
                      help="Path to the student names/ID map file (.csv)")
  parser.add_argument("--stuprint", dest="stuprint", action='store_true', default=False, 
                      help="Print student names for Teams grading. Requires path (--stunames)")
  parser.add_argument("--numeric", dest="numeric", action='store_true', default=False, 
                      help="Output format for each question is decimal value")
  parser.add_argument("--exnames", dest="exnames", action='store_true', default=False, 
                      help="List the exercise names for a given course")
  parser.add_argument("--exdates", dest="exdates", action='store_true', default=False, 
                      help="List the exercise names and start/stop dates for a given course")
  stu_list = parser.add_mutually_exclusive_group()

  stu_list.add_argument("-s", dest="student", type=str, default=None)

  detail_list = parser.add_mutually_exclusive_group()
  detail_list.add_argument("-v", dest="verbose", action='count', default=0, help='verbose flag')
  detail_list.add_argument("-q", dest="qid", type=str, help='Question ID')

  return parser.parse_args()

def client():
  opts = parseArgs()

  if opts.exnames or opts.exdates:
    printExerciseInfo(opts.course, opts.exdates, False)
    printExerciseInfo(opts.course, opts.exdates, True)
    sys.exit(0)
  
  users = getUsers(opts.course, opts.instructors)

  if opts.student:
    users = [u for u in users if opts.student in u]

  if opts.grader is not None:
    config.GRADER = opts.grader

  if opts.qid:
    if not opts.student:
      logging.error('Please provide a student name')
      sys.exit(-5)

  res = gradeAll(opts.course, opts.name, users, opts.date, opts.hardstop, opts.exam)

  if opts.verbose >= 2:
    print(dumps(res))
    sys.exit(0)

  if opts.qid:
    invokeTemp(res[users[0]][opts.qid])
    sys.exit(0)

  if opts.stuprint:
    if not opts.stunamespath:
      logging.error("Please provide path to csv with names/usernames")
      sys.exit(0)
    usernames = loadUsernameCSV(opts.stunamespath)
    first_padding, last_padding, combined_padding, username_padding = namePadding(usernames)

  output = []
  for user in sorted(res):
    val = res[user]
    username = user.split('+')[-1]

    if opts.stuprint:
      try:
        name = usernames[username]
        first, last = name["first"], name["last"]
      except KeyError:
        first, last = "", "Name not found"

    if isinstance(val, str):
      info = val
    elif isinstance(val, dict):
      total_score = "%4.1f" % (sum([val[q]['score'] for q in val])*100 / float(len(val)))
      info = str.rjust(total_score, 5)
      if opts.stuprint:
        scores = []
        for qid in sorted(val):
          if opts.exam or opts.numeric:
            score_str = '%.2f' % val[qid]['score']
          else:
            if val[qid]['score'] < 1.0:
              score_str = "Incorrect"
            else:
              score_str = "Correct"
          scores.append('Q%s: %s' % (qid, score_str))
        info = info + " " + ", ".join(scores)

      else:
        if opts.verbose == 1:
          scores = []
          for qid in sorted(val):
            scores.append('%s:%.2f' % (qid, val[qid]['score']))
          info = '%s %s' % (info, dumps(scores))
    else:
      info = 'No attempt found'

    if not opts.stuprint:
      print("%-10s %s" % (username, info))
    else:
      output.append("%s -- %s -- %s" % (str.ljust(last + ", " + first, combined_padding + 2), 
                                        str.ljust(str(username), username_padding),
                                        info))

  if opts.stuprint:
    for stu in sorted(output):
      print(stu)
      print()
