# Copyright (c) 2019-2020  University of Houston

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/

from __future__ import absolute_import

import base64
import dpkt
import re
from string import Template

from flask import current_app

from .solvers import *
from . import json


def parse_data(data):
  import geni.rspec.vtsmanifest as VTSM

  newdata = {}

  # { "<fname-root>" : (data, type), ...} except index, which is just an integer
  for k,v in data.items():
    if k == "index":
      continue
    elif k == "use-vlans":
      newdata[k] = v
    elif k == "listresources":
      newdata["manifest"] = VTSM.Manifest(xml=v[0])
    elif v[1] == "json":
      newdata[k] = json.loads(v[0])
    else:
      newdata[k] = base64.b64decode(v[0]) 	

  newdata['hosts'] = {}
  newdata['macs'] = {}
  newdata['ipaddrs'] = {}

  return newdata


def pop_uidata(obj):
  uidata = {}
  if isinstance(obj, (tuple, list)):
    # Look in widgettabs as lists
    if '_uidata' in obj[0]:
      uidata.update(obj[0]["_uidata"])
  else:
    if 'values' in obj:
      if '_uidata' in obj['values'][0]:
        res = obj['values'].pop(0)
        uidata.update(res['_uidata'])
      if 'brname' in obj['values'][0]:
        res = obj['values'].pop(0)
        uidata.update(res)
  return uidata


def eval_template(row, uidata):
  # Helper func
  def subs(item, data):
    #fix string issue
    item = re.sub(r'(\$[a-z]*)', r"'\1'", item)
    item = Template(item).substitute(data)
    return eval(item)

  res = None
  if isinstance(row, list):
    res = []
    for item in row:
      res.append(subs(item, uidata))
  else:
    res = subs(row, uidata)

  return res


def grade(data, key_template, attempt):
  scores = {}
  for qid in attempt:
    # Parse datapool values
    data[qid] = parse_data(data[qid])

    # pop uidata
    uidata = pop_uidata(attempt[qid])

    # Fill template and create objects
    answerkey = eval_template(key_template[qid], uidata)

    # score
    score = 0.

    # prep to iter
    cur_key = answerkey
    cur_ans = attempt[qid]
    if isinstance(cur_key, BaseTable):
      cur_key = [cur_key]
      cur_ans = [attempt[qid]['values']]
    elif isinstance(cur_key, MACAddress):
      cur_key = [cur_key]
      cur_ans = attempt[qid]
    elif isinstance(cur_key, ActionGroup):
      cur_key = [cur_key]
      cur_ans = [attempt[qid]['values']]
    elif isinstance(cur_key, BasePacket):
      cur_key = [cur_key]
      cur_ans = [attempt[qid]['values']]
    elif isinstance(cur_key, BaseTemplate):
      cur_key = [cur_key]
      cur_ans = attempt[qid]
    elif isinstance(cur_key, list):
      cur_key = cur_key
      cur_ans = attempt[qid]['values']

    for keyrow, ans in zip(cur_key, cur_ans):
      keyrow.resolve(data[qid])
      score += keyrow.check(ans)

    scores[qid] = {
        'score': score/float(len(cur_key)),
        'key': [keyrow.__json__() for keyrow in cur_key],
        'ans': cur_ans
    }

  return scores

# vim: et ts=2 sw=2
