# Copyright (c) 2017-2018  University of Houston

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from flask import Response, json

class APIEncoder(json.JSONEncoder):
  def default (self, obj): # pylint: disable=E0202
    if hasattr(obj, "__json__"):
      return obj.__json__()
    elif isinstance(obj, set):
      return list(obj)
    else:
      return json.JSONEncoder.default(self, obj)

def dumps (obj, sort_keys=False, indent=None):
  return json.dumps(obj, cls=APIEncoder, sort_keys=sort_keys, indent=indent)

def jsonify (data, code = 0, msg = ""):
  if isinstance(data, Exception) and hasattr(data, "code"):
    retstruct = data
  else:
    retstruct = { "retcode" : code, "msg" : msg, "value" : data }
  return Response(json.dumps(retstruct, cls=APIEncoder), mimetype='application/json')

loads = json.loads

class JSONInput(object):
  def validate (self, _):
    return False

  def construct (self, _):
    raise NotImplementedError()

class JSONValidationError(Exception):
  def __init__ (self, fields):
    super(JSONValidationError, self).__init__()
    self.code = 2
    self.__fields = fields

  def __json__ (self):
    return {"exception" : "JSON Parse Error: expected fields were missing and/or of the wrong type",
         "missing" : ["%s: %s" % (n,str(t)) for (p,n,t) in self.__fields if not p],
         "wrong-type" : ["%s: %s" % (n, str(t)) for (p,n,t) in self.__fields if p]}

  def __str__ (self):
    return "JSON Parse Error: required fields were missing and/or of the wrong type"

def jsonValidate (obj, fields, logger):
  fail = []
  out_objs = {}
  for (name, typ, opt) in parse(fields):
    if (not opt) and (not name in obj):
      fail.append((False, name, typ))
      continue
    elif opt and (not name in obj):
      continue
    if typ is not None:
      if isinstance(typ, tuple):
        if not isinstance(obj[name], typ):
          logger.warning("[%s] %s, expected %s" % (name, type(obj[name]), typ))
          fail.append((True, name, typ))
        else:
          out_objs[name] = obj[name]
      elif issubclass(typ, JSONInput):
        if not typ.validate(obj[name]):
          logger.warning("[%s] %s, expected %s" % (name, type(obj[name]), typ))
          fail.append((True, name, typ))
        else:
          out_objs[name] = typ.construct(obj[name])
      else:
        out_objs[name] = obj[name]
  if fail:
    raise JSONValidationError(fail)
  return out_objs

def parse (fields):
  for tup in fields:
    if len(tup) == 2:
      yield(tup[0], tup[1], False)
    elif len(tup) == 3:
      yield(tup[0], tup[1], tup[2])

def getdefault (obj, attr, val):
  try:
    return obj[attr]
  except KeyError:
    return val

OPTIONAL = True
