# Copyright (c) 2020-2021  University of Houston

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/

import datetime
import requests

from .json import loads

class MSTeamsProxy(object):
  def __init__ (self, lsp):
    self._lsp = lsp
    self._uuid = None
    self._users = None

  @property
  def uuid (self):
    if not self._uuid:
      url = "%s/msteams/course/%s/uuid" % (self._lsp._baseurl, self._lsp._course)
      self._uuid = requests.get(url).json()["value"]
    return self._uuid

  @uuid.setter
  def uuid (self, value):
    url = "%s/msteams/course/%s/uuid" % (self._lsp._baseurl, self._lsp._course)
    self._uuid = value
    requests.post(url, json={"uuid":value}).json()["value"]

  @property
  def users (self):
    if not self._users:
      url = "%s/msteams/course/%s/users" % (self._lsp._baseurl, self._lsp._course)
      self._users = requests.get(url).json()["value"]
    return self._users

  def setUserUUID (self, shortid, msuuid):
    url = "%s/msteams/%s/uuid/%s" % (self._lsp._baseurl, shortid, msuuid)
    self._users = None
    return requests.put(url).json()["value"]

  def getAssignmentUUID (self, euuid):
    url = "%s/msteams/course/%s/exercise/%s/assignment" % (self._lsp._baseurl, self._lsp._course,
                                                           euuid)
    return requests.get(url).json()["value"]

  def setAssignmentUUID (self, euuid, msuuid):
    url = "%s/msteams/course/%s/exercise/%s/assignment" % (self._lsp._baseurl, self._lsp._course,
                                                           euuid)
    return requests.post(url, json={"uuid":msuuid}).json()["value"]



class LabserveProxy(object):
  def __init__ (self, course = None, baseurl = "http://localhost:9008"):
    self._baseurl = baseurl
    self._course = course
    self._users = []
    self._exercises = []
    self._exams = []
    self.msteams = MSTeamsProxy(self)

  @property
  def course (self):
    return self._course

  @course.setter
  def course (self, val):
    if val != self._course:
      self._users = []
      self._exercises = []
    self._course = val

  @property
  def users (self):
    if not self._users:
      url = "%s/course/%s/users" % (self._baseurl, self.course)
      users = requests.get(url).json()["value"]
      self._users = users
    return self._users

  @property
  def exercises (self):
    if not self._exercises:
      url = "%s/course/name/%s/exercises" % (self._baseurl, self.course)
      exlist = requests.get(url).json()["value"]
      exlist.sort(key = lambda x: datetime.datetime.strptime(x["hard-stop"], "%Y-%m-%d %H:%M"))
      self._exercises = exlist
    return self._exercises

  @property
  def exams (self):
    if not self._exams:
      url = "%s/course/name/%s/exams" % (self._baseurl, self.course)
      exlist = requests.get(url).json()["value"]
      exlist.sort(key = lambda x: datetime.datetime.strptime(x["hard-stop"], "%Y-%m-%d %H:%M"))
      self._exams = exlist
    return self._exams

  def getAnswerKey (self, name):
    url = "%s/answer-key/name/%s" % (self._baseurl, name)
    return requests.get(url).json()["value"]

  def getAttempts (self, euuid, user_urn, exam = False):
    att_list = []
    if exam:
      url = "%s/exam/%s/user/%s/attempts" % (self._baseurl, euuid, user_urn)
    else:
      url = "%s/exercise/%s/user/%s/attempts" % (self._baseurl, euuid, user_urn)
    atts = requests.get(url).json()["value"]
    for index,ts in atts:
      if exam:
        atturl = "%s/exam/%s/user/%s/attempt/%d" % (self._baseurl, euuid, user_urn, index)
      else:
        atturl = "%s/exercise/%s/user/%s/attempt/%d" % (self._baseurl, euuid, user_urn, index)
      att_list.append((ts, loads(requests.get(atturl).json()["value"])))
    return att_list

  def getDatapoolIndex (self, datapool, index, use_vlans):
    url = "%s/datapool/%s/get/%d" % (self._baseurl, datapool, index)
    data = requests.get(url).json()["value"]
    data["use-vlans"] = use_vlans
    return data

  def getExam (self, edatestr, user_urn = "none"):
    url = "%s/course/%s/exam/%s" % (self._baseurl, self.course, user_urn)
    indata = {"admin" : True, "date" : edatestr}
    outdata = requests.post(url, json=indata).json()["value"]
    return outdata

  def getExerciseByToponame (self, name):
    for xrcs in self.exercises:
      if isinstance(xrcs["toponame"], (list, tuple)):
        if name in xrcs["toponame"]:
          return xrcs
      else:
        if name == xrcs["toponame"]:
          return xrcs
    
    

