# Copyright (c) 2019-2020  University of Houston

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/

from flask import current_app
import IPy

from .base import ActionGroup, MACAddress, splitHostIntf, BaseTemplate
from .actions import SendPacket, Forward, Flood, ReceivePacket, ReceiveAndIgnore, ARPRequest, ARPReply, EchoRequest, EchoReply
from .tables import resolveL2Table


def resolveFloodPorts(data, brname, inMAC):
  from uhgeni.graph.util import buildFromManifest

  manifest = data['manifest']
  l2table = resolveL2Table(data, brname)
  ports = set()
  vlan = None

  for row in l2table:
    if str(row[1]) == str(inMAC):
      inport = row[0]
      if data['use-vlans']:
        vlan = row[2]
        break

  # Different broadcast Domain?
  if inport is None:
    return None

  if "api:l2-switch:stp-info" in data:
    g = buildFromManifest(manifest, site="ignore")
    g.decorateSTPInfo(stpinfo=data["api:l2-switch:stp-info"])
    vertex = g.vertices[brname]
    for e in vertex.edges:
      if not e.port_a.up or not e.port_b.up:
        continue
      states = set()
      if e.port_a.stp: states.add(e.port_a.stp.state)
      if e.port_b.stp: states.add(e.port_b.stp.state)
      if len(states) > 1:
        continue
      if 'forwarding' not in states:
        continue
      if e.port_a.vertex == vertex:
        port = e.port_a
      if e.port_b.vertex == vertex:
        port = e.port_b
      if vlan:
        if port.vlanid in [0, vlan]:
          ports.add(port.name)
      else:
        ports.add(port.name)
  else:
    for port in manifest[brname].ports:
      for rport in manifest[port.remote_dpname].ports:
        if rport.client_id == port.remote_client_id:
          if vlan:
            if port.vlanid in [0, vlan]:
              ports.add(port.name)
          else:
            ports.add(port.name)

  ports.remove(inport)
  return list(ports)


def pathWalker(data, start, end, mac=False, flood=False):
  manifest = data['manifest']
  sHost, sIntf = splitHostIntf(start)
  eHost, eIntf = splitHostIntf(end)
  sMAC = manifest[sHost][sIntf].macaddress
  eMAC = manifest[eHost][eIntf].macaddress

  path = []
  cur = manifest[sHost][sIntf]
  while cur.remote_dpname != eHost:
    brname = cur.remote_dpname
    rport = cur.remote_client_id.split(':')[-1]

    outport = None
    vlan = None
    l2table = resolveL2Table(data, brname)

    for row in l2table:
      if str(row[1]) == sMAC:
        inport = row[0]
      if str(row[1]) == eMAC:
        outport = row[0]
        if data['use-vlans']:
          vlan = row[2]

    curAction = None
    if flood:
      ports = resolveFloodPorts(data, brname, inMAC=sMAC)
      curAction = Flood(brname, ports)
    else:
      kwargs = {}
      if mac:
        kwargs['destmac'] = eMAC
      if data['use-vlans']:
        kwargs['vlan'] = vlan
      if manifest[brname][outport].remote_dpname == eHost:
        kwargs['vlan'] = '[none]'
      curAction = Forward(brname, outport, **kwargs)

    path.append(curAction)

    cur = manifest[brname][outport]

  return path


class L2PathWalker(ActionGroup):
  def __init__(self, start, end, flood=False):
    super(L2PathWalker, self).__init__()
    self._start = start
    self._end = end
    self._flood = flood

  def __repr__ (self):
    return("{0}: {1}, {2}, {3}".format(self.__class__.__name__, 
                           str(self._start), str(self._end), str(self._flood)))

  def resolve(self, data):
    self.values += [SendPacket(self._start, MACAddress(self._end))]
    self.values += pathWalker(data, self._start, self._end, flood=self._flood)
    self.values += [ReceivePacket(self._end)]

    super(L2PathWalker, self).resolve(data)


class ARPTravel(ActionGroup):
  def __init__(self, host, ip, mac=False):
    super(ARPTravel, self).__init__()
    self._host = host
    self._ip = ip
    self._mac = mac

  def __repr__ (self):
    return("{0}: {1}, {2}, {3}".format(self.__class__.__name__, 
                         self._host, self._ip, self._mac))

  def resolve(self, data):
    manifest = data['manifest']
    status = data['sliverstatus']

    sHost = self._host
    sIntf = None
    eHost = None
    eIntf = None

    for container in status['vts_containers']:
      for port in container['ports']:
        for addr in port['ipv4-addresses']:
          ip = addr.split('/')[0]
          if self._ip == ip:
            eHost = container['client-id']
            eIntf = port['name']
          if container['client-id'] == self._host:
            subnet = IPy.IP(addr, make_net=True)
            if self._ip in subnet:
              sIntf = port['name']

    sMAC = manifest[sHost][sIntf].macaddress
    eMAC = manifest[eHost][eIntf].macaddress

    self.values += [ARPRequest((sHost, sIntf), self._ip)]
    self.values += pathWalker(data, (sHost, sIntf), (eHost, eIntf), flood=True, mac=self._mac)
    self.values += [
        ReceivePacket((eHost, eIntf)),
        ARPReply((eHost, eIntf), eMAC)
    ]
    self.values += pathWalker(data, (eHost, eIntf), (sHost, sIntf), mac=self._mac)
    self.values += [ReceivePacket((sHost, sIntf))]

    # Add Flood
    brnames = [action.vertex for action in self.values if isinstance(action, Flood)]
    brnames = [dp.client_id for dp in manifest.datapaths if dp.client_id not in brnames]
    for brname in brnames:
      found = False
      for rows in resolveL2Table(data, brname):
        if str(rows[1]) == sMAC:
          found = True
      if found:
        ports = resolveFloodPorts(data, brname, inMAC=sMAC)
        self.values += [Flood(brname, ports)]

    # Add Receive-and-Ignore
    # Assumes all hosts are in the same broadcast domain
    # Hosts can only have 'eth1' !!!
    brname = manifest[sHost][sIntf].remote_dpname
    for rows in resolveL2Table(data, brname):
      if str(rows[1]) in [sMAC, eMAC]:
        continue
      else:
        self.values += [ReceiveAndIgnore(mac2HostPort(data, rows[1]))]

    super(ARPTravel, self).resolve(data)


def resolveRoute(data, host, dstip):
  # Remove prefix len
  if '/' in dstip:
    dstip = dstip.split('/')[0]

  # Check if it is a quagga instance
  if 'vts:uh.quagga:get-route-table' in data.keys():
    if host in data['vts:uh.quagga:get-route-table']:
      routes = data['vts:uh.quagga:get-route-table'][host]

      for route in routes:
        if route['interface'] == 'lo':
          continue
        if not route['network']:
          continue
        if not route['selected']:
          continue
        net = route['network']
        if ' ' in net:
          net = net.split(' ')[0]
        net = IPy.IP(net, make_net=True)
        if dstip in net:
          return route['interface'], route['next-hop']

  # Check the host kernel router table
  if 'api:uh.host:get-route-table' in data.keys():
    if host in data['api:uh.host:get-route-table']:
      routes = data['api:uh.host:get-route-table']
      routes = routes[host]
      for route in routes:
        subnet = '%s/%s' % (route['destination'], route['mask'])
        subnet = IPy.IP(subnet, make_net=True)
        if dstip in subnet:
          return route['device'], route['gateway']
      return None

  return None


def ip2HostPort(data, tip):
  st = data['sliverstatus']
  for c in st['vts_containers']:
    for p in c['ports']:
      for addr in p['ipv4-addresses']:
        if tip in addr:
          host = c['client-id']
          intf = p['name']
          return (host, intf)


def mac2HostPort(data, mac):
  mac = str(mac)
  manifest = data['manifest']
  for port in manifest.ports:
    if getattr(port, 'macaddress', '') == mac:
      return port.dpname, port.name


def hostPort2ip(data, hostIntf, targetip=None):
  host, intf = splitHostIntf(hostIntf)

  for c in data['sliverstatus']['vts_containers']:
    if c['client-id'] == host:
      for p in c['ports']:
        for addr in p['ipv4-addresses']:
          if targetip:
            subnet = IPy.IP(addr, make_net=True)
            if targetip in subnet:
              return addr
          return addr
      break


class RouteInterface(BaseTemplate): 	
  def __init__ (self, sending_host, dest_ip):
    super(RouteInterface, self).__init__() 	
    self.dest_ip = dest_ip 	
    self.sending_host = sending_host 	
    self.intf_dev = None 	
    self.gateway = None
 
  def __repr__ (self):
    return("{0}: {1}, {2}, {3}, {4}".format(self.__class__.__name__, 
                           str(self.sending_host), str(self.dest_ip), 
                           str(self.intf_dev), str(self.gateway)))

  def resolve (self, data): 	
    self.intf_dev, self.gateway = resolveRoute(data, self.sending_host, self.dest_ip) 	

  def cleanAnswer (self, ans):
    intf = ans["interface"]
    parts = intf.split(".")
    if len(parts) == 2:
      if parts[0] == self.sending_host:
        ans["interface"] = parts[1]
    ans["interface"] = ans["interface"].strip()

  def check (self, other):
    self.cleanAnswer(other)
    if self.intf_dev == other["interface"]: 	
      return 1. 	
    return 0.


def L3PathWalker(data, start, targetIP, mac=True, arp=False):
  path = []
  manifest = data['manifest']

  sHost, sIntf = splitHostIntf(start)
  tHost, tIntf = ip2HostPort(data, targetIP)
  tMAC = manifest[tHost][tIntf].macaddress

  curHost, curIntf = sHost, sIntf

  while curHost != tHost:
    #Find next hop
    curIntf, nIP = resolveRoute(data, curHost, targetIP)
    if nIP == 'directly connected':
      nHost, nIntf = tHost, tIntf
      nIP = targetIP
    else:
      nHost, nIntf = ip2HostPort(data, nIP)

    nMAC = manifest[nHost][nIntf].macaddress

    if curHost != sHost:
      if arp:
        arpPath = ARPTravel(curHost, nIP, mac=True)
        arpPath.resolve(data)
        path += arpPath.values
      path += [Forward("%s.%s" % (curHost, curIntf), destmac=nMAC, destip = targetIP)]

    kwargs = {
        'data': data,
        'start': (curHost, curIntf),
        'end': (nHost, nIntf),
        'mac': mac,
        'flood': (nMAC == 'ff:ff:ff:ff:ff:ff')
    }
    path += pathWalker(**kwargs)
    path += [ReceivePacket((nHost, nIntf))]

    if (curHost, curIntf) == (nHost, nIntf):
      raise Exception
    curHost, curIntf = nHost, nIntf

  return path


class EchoTravel(ActionGroup):
  def __init__(self, host, ip, request=True, reply=False, arp=False):
    super(EchoTravel, self).__init__()
    self._host = host
    self._ip = ip
    self._request = request
    self._reply = reply
    self._arp = arp

  def __repr__ (self):
    return("{0}: {1}, {2}, {3}, {4}, {5}".format(self.__class__.__name__, 
                           str(self._host), str(self._ip), 
                           str(self._request),
                           str(self._reply), str(self._arp)))

  def resolve(self, data):
    manifest = data['manifest']
    status = data['sliverstatus']

    sHost = self._host
    sIntf, nIP = resolveRoute(data, self._host, self._ip)
    sIP = hostPort2ip(data, (sHost, sIntf))
    sMAC = manifest[sHost][sIntf].macaddress

    eHost, eIntf = ip2HostPort(data, self._ip)
    eIP = self._ip

    if self._request:
      nHost, nIntf = ip2HostPort(data, nIP)
      destmac = manifest[nHost][nIntf].macaddress

      if self._arp:
        arpPath = ARPTravel(self._host, nIP, mac=True)
        arpPath.resolve(data)
        self.values += arpPath.values

      self.values += [EchoRequest((sHost, sIntf), destip=eIP, destmac=destmac)]

      self.values += L3PathWalker(data, (sHost, sIntf), self._ip, arp=self._arp)

    if self._reply:
      eIntf, nIP = resolveRoute(data, eHost, sIP)
      nHost, nIntf = ip2HostPort(data, nIP)
      destmac = manifest[nHost][nIntf].macaddress

      self.values += [EchoReply((eHost, eIntf), destip=sIP, destmac=destmac)]

      self.values += L3PathWalker(data, eHost, sIP, arp=False)

    super(EchoTravel, self).resolve(data)

