# Copyright (c) 2019-2020  University of Houston

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/

from __future__ import print_function

import re

from .base import BaseAction, MACAddress


class SendPacket(BaseAction):
  _ACTION = 'send-packet'

  def resolve(self, data):
    if isinstance(self.data, MACAddress):
      self.data.resolve(data)
      mac = self.data.mac
    else:
      mac = self.data
    self.data = {"Dest. MAC:": mac}


class ReceivePacket(BaseAction):
  _ACTION = 'receive-packet'


class ReceiveAndIgnore(BaseAction):
  _ACTION = 'receive-and-ignore'


class Flood(BaseAction):
  _ACTION = 'flood'

  def resolve(self, data):
    if "Ports:" in self.data:
      return
    self.data = {"Ports:": sorted(self.data)}

  def check(self, other):
    if 'Ports:' in other['data']:
      other['data']['Ports:'] = re.findall(r'\d+', other['data']['Ports:'])
      other['data']['Ports:'].sort()

    return super(Flood, self).check(other)


class Forward(BaseAction):
  _ACTION = 'forward'

  def __init__(self, vertex, port=None, vlan=None, destmac=None, destip=None):
    super(Forward, self).__init__(vertex, None)
    self._vlan = vlan
    self._port = port
    self._destmac = destmac
    self._destip = destip

  def __repr__ (self):
    return("{0}: {1}, {2}, {3}, {4}".format(self.__class__.__name__, 
                           str(self._vlan), str(self._port),
                           str(self._destmac), str(self._destip)))

  def resolve(self, data):
    self.data = {}

    if self._port:
      self.data["Port:"] = self._port

    if data['use-vlans']:
      self.data['VLAN:'] = self._vlan

    if self._destmac:
      self.data['Dest. MAC:'] = str(self._destmac)

    if self._destip:
      if '/' in self._destip:
        self._destip = self._destip.split('/')[0]
      self.data['Dest. IP:'] = str(self._destip)

  def check (self, other):
    value = float(0)
    myself = self.__json__()
    total = len(other)

    for key in other:
      if key == "data":
        total = total + len(other[key]) - 1
        for inner_key in other[key]:
          try:
            if other[key][inner_key] == myself[key][inner_key]:
              value += 1
          except KeyError:
            continue
      else:
        if other[key] == myself[key]:
          value += 1

    return value/total


class ARPRequest(BaseAction):
  _ACTION = 'arp-request'

  def __init__(self, vertex, target):
    super(ARPRequest, self).__init__(vertex, {'Target IP:': target})


class ARPReply(BaseAction):
  _ACTION = 'arp-reply'

  def __init__(self, vertex, source):
    super(ARPReply, self).__init__(vertex, {'Source MAC:': source})


class EchoRequest(BaseAction):
  _ACTION = 'echo-request'

  def __init__(self, vertex, destip, destmac):
    if '/' in destip:
      destip = destip.split('/')[0]
    data = {'Dest. IP:': destip, 'Dest. MAC:': destmac}

    super(EchoRequest, self).__init__(vertex, data)


class EchoReply(BaseAction):
  _ACTION = 'echo-reply'

  def __init__(self, vertex, destip, destmac):
    if '/' in destip:
      destip = destip.split('/')[0]
    data = {'Dest. IP:': destip, 'Dest. MAC:': destmac}

    super(EchoReply, self).__init__(vertex, data)

