# Copyright (c) 2019-2020  University of Houston

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/

import IPy


class BaseTemplate(object):
  def resolve(self, data):
    raise NotImplementedError()

  def check(self, other):
    raise NotImplementedError()

  def cleanAnswer (self, ans):
    return

  def __json__(self):
    return {k: v for k, v in self.__dict__.items() if k[0] != '_'}


class BasePacket(BaseTemplate): 	
  pass


def splitHostIntf(inp):
  if isinstance(inp, tuple):
    return inp[0], inp[1]
  if inp.find('.') == -1:
    # Assume it is always eth1, maybe figure out from manifest?
    inp += '.eth1'
  return inp.split('.')


class MACAddress(BaseTemplate):
  def __init__(self, inp):
    super(MACAddress, self).__init__()
    self._inp = inp
    self.mac = None

  def resolve(self, data):
    host, intf = splitHostIntf(self._inp)

    manifest = data['manifest']
    self.mac = str(manifest[host][intf].macaddress)

  def check(self, other):
    if self.mac == other['mac']:
      return 1.
    return 0.


class BaseAction(BaseTemplate):
  _ACTION = None

  def __init__(self, vertex, data=None):
    super(BaseAction, self).__init__()
    self.action = self._ACTION
    if isinstance(vertex, tuple):
      self.vertex = '.'.join(vertex)
    else:
      self.vertex = vertex
    if data is None:
      data = {'': ''}
    self.data = data

  def resolve(self, data):
    if isinstance(self.vertex, BaseTemplate):
      self.vertex.resolve(data)

  def check(self, other):
    value = 0.
    myself = self.__json__()

    for key in myself:
      if myself[key] == other[key]:
        value += 1

    return value/len(myself)


class BaseTable(BaseTemplate):
  def __init__(self):
    super(BaseTable, self).__init__()
    self.values = []

  def resolve(self, data):
    raise NotImplementedError()

  def cleanAnswerRow (self, row):
    return

  def check(self, other):
    self.cleanAnswer(other)
    value = 0.

    if not self.values:
      return value

    ignored = 0
    checked_values = []
    for row in other:
      self.cleanAnswerRow(row)

      # ignore empty rows and
      if "brname" in row or "  -" in row.values():
        ignored += 1
        continue
      if row in self.values and row not in checked_values:
        value += 1
        checked_values.append(row)

    if len(other) - ignored == 0:
      return 0.

    value = value / len(self.values)
    # make penalty float
    penalty = 1. * len(self.values) / (len(other) - ignored)
    # penalty can be maximum 1
    penalty = min(1, penalty)
    # apply penalty
    value = value * penalty

    return value

  def __json__(self):
    return self.values


class TableGroup(BaseTable):
  def __init__(self, kls=BaseTable):
    super(TableGroup, self).__init__()
    self.__kls = kls

  def resolve(self, data):
    new_vals = []
    for val in self.values:
      inst = self.__kls(val)
      inst.resolve(data)
      new_vals.append(inst)
    self.values = new_vals

  def check(self, other):
    self.cleanAnswer(other)

    value = 0.

    if not self.values:
      return value

    for val in self.values:
      for oth in other:
        if val == oth:
          value += val.check(oth)

    return value/len(self.values)


class ActionGroup(BaseTable):
  def resolve(self, data):
    for row in self.values:
      row.resolve(data)

  def check(self, other):
    value = 0.

    floodans = []
    floodkey = []
    restans = []
    restkey = []
    ignoreans = []
    ignorekey = []

    for row in other:
      if row.get('action', '') == 'flood':
        floodans.append(row)
      elif row.get('action', '') == 'receive-and-ignore':
        ignoreans.append(row)
      elif row.get('action', '') == '  -':
        continue
      else:
        restans.append(row)

    for row in self.values:
      if row.action == 'flood':
        floodkey.append(row)
      elif row.action == 'receive-and-ignore':
        ignorekey.append(row)
      else:
        restkey.append(row)

    floodkey.sort(key=lambda x: x.vertex)
    floodans.sort(key=lambda x: x['vertex'])
    ignorekey.sort(key=lambda x: x.vertex)
    ignoreans.sort(key=lambda x: x['vertex'])

    for x, y in zip([floodkey, restkey, ignorekey], [floodans, restans, ignoreans]):
      for kr, ans in zip(x, y):
        value += kr.check(ans)

    return value/len(self.values)

# vim: et ts=2 sw=2
