# Copyright (c) 2019-2020  University of Houston

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/

from __future__ import absolute_import

import dpkt
from flask import current_app
import IPy
import six

from uhnet.types import EthernetMAC
from uhnet.pcap.eth import parse as eth_parse
from uhnet.pcap.arp import parse as arp_parse

from .base import BaseTable, TableGroup
from ..exceptions import UnableToResolveMACError, SourcePortNotFoundError

RESERVED = set([EthernetMAC('ff:ff:ff:ff:ff:ff')])


def resolveL2Table(data, brname, hide=[]):
  from uhgeni.graph.util import buildFromManifest

  macs = []
  manifest = data['manifest']
  vlan = data['use-vlans']

  if not isinstance(hide, list):
    hide = [hide]

  if "api:l2-switch:stp-info" in data:
    g = buildFromManifest(manifest, site="ignore")
    g.decorateSTPInfo(stpinfo=data["api:l2-switch:stp-info"])
    macs = g[brname].computeMACTable(include_vlans=vlan, hide=hide)
    return macs

  for port in manifest[brname].ports:
    for rport in manifest[port.remote_dpname].ports:
      if port.remote_dpname in hide:
        continue
      if rport.client_id == port.remote_client_id:
        try:
          if vlan:
            macs.append((port.name, rport.macaddress, port.vlanid))
          else:
            macs.append((port.name, rport.macaddress))
        except AttributeError:
          continue
  return macs


def resolveMAC2HostIntf(data, reject_images = None):
  if not reject_images:
    reject_images = []
  reject_images = set(reject_images)

  mac2hostintf = {}
  manifest = data['manifest']

  for container in manifest.containers:
    if container.image in reject_images:
      continue
    for port in container.ports:
      if not hasattr(port, "macaddress"):
        continue
    mac = str(port.macaddress)
    mac2hostintf[mac] = '%s.%s' % (port.dpname, port.name)
    
  if mac2hostintf:
    return mac2hostintf
  raise UnableToResolveMACError()


class MACTable(BaseTable):
  def __init__(self, brname):
    super(MACTable, self).__init__()
    self.brname = brname

  def __repr__ (self):
    return("{0}: {1}".format(self.__class__.__name__, str(self.brname)))

  def resolve(self, data):
    macs = resolveL2Table(data, self.brname)
    for row in macs:
      res = {
          "Port": row[0],
          "MAC Address": str(row[1])
      }
      if len(row) == 3:
        res["VLAN"] = row[2]
      self.values.append(res)


class SubnetTable(BaseTable):
  def __init__(self, subnet):
    super(SubnetTable, self).__init__()
    if isinstance(subnet, str):
      subnet = IPy.IP(subnet, make_net=True)
    self.subnet = subnet

  def __repr__ (self):
    return("{0}: {1}".format(self.__class__.__name__, str(self.subnet)))

  def resolve(self, data):
    status = data['sliverstatus']
    for c in status['vts_containers']:
      for p in c['ports']:
        for i in p['ipv4-addresses']:
          ip = IPy.IP(i.split('/')[0])
          if ip in self.subnet:
            row = {'Interface': '%s.%s' % (c['client-id'], p['name'])}
            self.values.append(row)

  def check(self, other):
    return super(SubnetTable, self).check(other['table'])

  def __eq__(self, other):
    if str(self.subnet) == other['subnet'].strip():
      return True
    return False

  def __json__(self):
    return {"subnet": str(self.subnet), "table": self.values}


class SubnetGroup(TableGroup):
  def __init__(self, kls=SubnetTable):
    super(SubnetGroup, self).__init__(kls=kls)

  def resolve(self, data):
    subnets = set()
    status = data['sliverstatus']
    for c in status['vts_containers']:
      for p in c['ports']:
        for i in p['ipv4-addresses']:
          s = IPy.IP(i, make_net=True)
          subnets.add(s)
    self.values = list(subnets)
    super(SubnetGroup, self).resolve(data)


class ARPTable(BaseTable):
  def __init__(self, host, ip):
    super(ARPTable, self).__init__()
    self._host = host
    self._ip = ip
    self.subnet = None

  def __repr__ (self):
    return("{0}: {1}, {2}, {3}".format(self.__class__.__name__, 
                           str(self._host), str(self._ip), str(self.subnet)))

  def resolve(self, data):
    manifest = data['manifest']

    mac2hostintf = resolveMAC2HostIntf(data)
    target = None

    for container in data['sliverstatus']['vts_containers']:
      if container['client-id'] == self._host:
        target = container
        break

    # find the correct interface
    target_port = None
    for port in target['ports']:
      for addr in port['ipv4-addresses']:
        subnet = IPy.IP(addr, make_net=True)
        if self._ip in subnet:
          target_port = port
          self.subnet = str(subnet)
    if not target_port:
      raise SourcePortNotFoundError(self._ip, self._host)

    # find the bridge name
    hostPort = manifest[self._host][target_port['name']]
    if data['use-vlans']:
      vlan = manifest.findPort(hostPort.remote_client_id).vlan_id

    # get the L2 table
    macs = resolveL2Table(data, hostPort.remote_dpname)

    for row in macs:
      mac = str(row[1])
      if mac == target_port['mac-address']:
        continue
      if data['use-vlans']:
        if vlan != row[2]:
          continue
      self.values.append({'Interface': mac2hostintf[mac]})

  def check(self, other):
    value = 0.
    other = other[0]

    if self.subnet != other['subnet'].strip():
      return value

    return super(ARPTable, self).check(other['table'])

  def __json__(self):
    return {"subnet": str(self.subnet), "table": self.values}


class DHCPTable(SubnetTable):
  def __init__(self, inp):
    super(DHCPTable, self).__init__(inp[0])
    self._server = inp[1]
    self._intf = inp[2]

  def __repr__ (self):
    return("{0}: {1}, {2}".format(self.__class__.__name__, 
                           str(self._server), str(self._intf)))

  def resolve(self, data):
    manifest = data['manifest']

    # TODO: reject_images should get configured through the exercise/answerkey
    mac2hostintf = resolveMAC2HostIntf(data, reject_images = ["uh.simple-dhcpd"])

    hostPort = manifest[self._server][self._intf]
    macs = resolveL2Table(data, hostPort.remote_dpname)
    if data['use-vlans']:
      vlan = manifest.findPort(hostPort.remote_client_id).vlan_id
      macs = [str(row[1]) for row in macs if row[2] == vlan]
    else:
      macs = [str(row[1]) for row in macs]

    for mac in macs:
      try:
        self.values.append({'Interface': mac2hostintf[mac]})
      except KeyError: # We won't always find the MACs now
        continue


class DHCPGroup(TableGroup):
  def __init__ (self, kls=DHCPTable):
    super(DHCPGroup, self).__init__(kls=kls)

  def resolve (self, data):
    subnets = set()
    status = data['sliverstatus']
    for c in status['vts_containers']:
      if 'dhcp' in c['client-id']:
        for p in c['ports']:
          for i in p['ipv4-addresses']:
            subnets.add((
                IPy.IP(i, make_net=True),
                c['client-id'],
                p['name']
            ))
    self.values = list(subnets)
    super(DHCPGroup, self).resolve(data)


class L2PcapTable(MACTable):
  def __init__(self, brname, pcaps=[]):
    super(L2PcapTable, self).__init__(brname)
    self._pcaps = pcaps

  def __repr__ (self):
    return("{0}: {1}".format(self.__class__.__name__, str(self._pcaps)))

  def resolve(self, data):

    for name in self._pcaps:
      pcap_meta = data['pcap-roster'][name]
      pcap_name = pcap_meta['pcap-name']

      port = data['manifest'][pcap_meta['bridge']][pcap_meta['port']]

      fl = six.BytesIO(data[pcap_name])

      reader = dpkt.pcap.Reader(fl)
      for (ts, pkt) in reader:
        fields = eth_parse(pkt)

        if fields['src'] in RESERVED:
          break

        if fields['tag']:
          if port.vlan_id == 0: #Trunk
            vid = fields['tag']
          else:
            break
        else:
          vid = port.vlan_id

        row = {
            'Port': pcap_meta['port'],
            'MAC Address': str(fields['src'])
        }

        if data['use-vlans']:
          row['VLAN'] = vid

        self.values.append(row)


class PcapARPTable(BaseTable):
  def __init__(self, host, interface, pcaps):
    self.hostname = host
    self._intf = interface
    self._pcaps = pcaps
    self.values = []

  def __repr__ (self):
    return("{0}: {1}, {2}, {3}".format(self.__class__.__name__, 
                           str(self.hostname), str(self._intf), self._pcaps))

  def cleanAnswerRow (self, row):
    for k,v in row.items():
      try:
        row[k] = v.strip()
      except AttributeError:
        continue

  def resolve(self, data):
    # circular dependency!
    from .actiongroups import hostPort2ip
    table = {}
    if 'api:uh.host:get-arp-table' in data:
      for row in data['api:uh.host:get-arp-table'][self.hostname]:
        if row['device'] == self._intf:
          table[row['ip-address']] = row['hw-address']
    ip = hostPort2ip(data, (self.hostname, self._intf))
    ip = ip.split('/')[0]
    for pcap_name in self._pcaps:
      fl = six.BytesIO(data[pcap_name])
      reader = dpkt.pcap.Reader(fl)
      for (ts, pkt) in reader:
        eth = eth_parse(pkt)
        arp = arp_parse(eth)
        if arp['spa'] in table:
          table[str(arp['spa'])] = str(arp['sha'])
        else:
          if str(arp['tpa']) == ip:
            table[str(arp['spa'])] = str(arp['sha'])
    for ip, mac in table.items():
      self.values.append({
          'Interface': self._intf,
          'IP Address': ip,
          'MAC Address': mac
      })

  def check(self, other):
    value = 0.
    for item in other:
      if item['hostname'] != self.hostname:
        continue
      value = super(PcapARPTable, self).check(item['table'])
    return value

