# Copyright (c) 2017-2021  University of Houston

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import, print_function

import datetime
import functools
import getpass
import json
import os.path
import random
import subprocess
import sys
import types

# pylint: disable=F0401
import dateutil.parser
from IPython.core.getipython import get_ipython
from IPython.core.magic import Magics, magics_class, line_magic
from IPython.display import display, Javascript
import ipywidgets as widgets
import requests
import six

from geni.aggregate.exceptions import AMError
from geni.aggregate.apis import CreateSliverError
from geni.aggregate.frameworks import KeyDecryptionError
import geni.aggregate.instageni as IGAM
import geni.aggregate.vts as VTSAM
import geni.minigcf.config
import geni.rspec.vtsmanifest
from geni.support.ipython import am_exc_handler, NoContextError, replaceSymbol, gsh
import grader.client
from grader.labserve import LabserveProxy

from . import creds
from . import net
from . import vtswrapper
from .build import sliverExists
from .dplib.dns import ZonesDatapool
from .exercise import ExerciseManager
from .log import Logger
from .scenarios import ScenarioManager as SMGR
from .tables import ListTable
from .uhwidgets.dnsie import InteractiveExperience as DNSIE
from .uhwidgets.numbersie import InteractiveExperience as NUMIE
from .uhwidgets.practice import PracticeController
from .util import LabserveError, userHasSliceCred, issueUserSliceCred

edmodule = types.ModuleType("uhed_ipython")

geni.minigcf.config.HTTP.TIMEOUT = 90

class LabModule(object):
  def __init__ (self, modid, spec):
    self.modid = modid
    self.course = spec["course"]
    self.params = spec["params"]
    self.request = spec["request"]
    self.site_filters = spec["site-filters"]
    self.sort = spec["sort"]
    self.uiname = spec["uiname"]
    self.vertex_filters = spec["vertex-filters"]

    self.var_pool = []
    self.save_manifest = True
    self.pre_build = None
    self.post_build = None
    self.topo_name = None
    self.instructor_only = False
    if spec["version"] >= 2:
      self.var_pool = spec["var-pool"]
      self.save_manifest = spec["save-manifest"]
      self.post_build = spec["post-build"]
    if spec["version"] >= 3:
      self.topo_name = spec["topo-name"]
    if spec["version"] >= 4:
      self.pre_build = spec["pre-build"]
    if spec["version"] >= 5:
      self.instructor_only = spec["instructor-only"]
    if spec["version"] >= 8:
      self.slice_suffix = spec["slice-suffix"]
      self.depends = spec["depends"]

rkwargs = {"headers" : {"Content-Type" : "application/json"},
           "verify" : False}

def getExerciseInfo (course, topo_name):
  exinfo = requests.get("http://localhost:9008/course/%s/exercise/%s" % (course, topo_name)).json()["value"]
  return exinfo

def getSiteConfig (key):
  data = requests.get("http://localhost:9008/config/%s" % (key)).json()["value"]
  return data

def courseinfo (userurn, course, admin = False, pdate = None):
  modlist = []
  if admin and pdate:
    data = json.dumps({"date" : pdate})
    for modid in requests.post("http://localhost:9008/course/%s/modules/%s" % (course, userurn),
                               data, **rkwargs).json()["value"]:
      modlist.append(LabModule(modid, requests.get("http://localhost:9008/module/%s" % (modid)).json()["value"]))
  else:
    for modid in requests.get("http://localhost:9008/course/%s/modules/%s" % (course, userurn)).json()["value"]:
      modlist.append(LabModule(modid, requests.get("http://localhost:9008/module/%s" % (modid)).json()["value"]))
  return modlist

def pickSite (userurn, course):
  site = requests.get("http://localhost:9008/course/%s/site/%s" % (course, userurn)).json()["value"]
  return site

def hasCredentials (course):
  credexists = requests.get("http://localhost:9008/course/%s/creds/enabled" % (course)).json()["value"]
  return credexists

def issueUserCred (course, user, project, ch_url):
  Logger.log("Issuing user credential for course %s" % (course))
  creds.addUserEnv(user)
  creds.makeSSHKeys(user)
  usercred_bundle = requests.get("http://localhost:9008/course/%s/creds/issue-user/%s" % (course, user)).json()["value"]
  creds.writeUCreds(user, usercred_bundle)
  user_cert = creds.makeContext(user, project, ch_url)
  return user_cert

def submit_pbar (dropdown, modlist, _):
  f = widgets.FloatProgress(description="Loading: ", min=0, max=len(modlist))
  display(f)
  for mod in modlist:
    f.value += 1
    if mod.uiname == dropdown.value:
      f.bar_style="success"
      f.description = "Done!"
      f.value = len(modlist)
      return

def selectnet (dropdown, modlist, _):
  for mod in modlist:
    if mod.uiname == dropdown.value:
      build(mod)
      lbl = widgets.Label(value="Success!")
      display(lbl)
      display(Javascript('IPython.notebook.select_next()'))
      return

def depBuild (modname, uhm):
  Logger.log("%%lab: building dependency - %s" % (modname))

  url = "http://localhost:9008/course/%s/modules/topo/%s" % (uhm._course, modname)
  mod = LabModule(modname, requests.get(url).json()["value"])

  (r, vardata) = build(mod)

  if mod.pre_build:
    uhm.setLabProgressLabel("Pre-build setup...")
    prebuild(mod, vardata, uhm)

  _ip = get_ipython()
  context = _ip.user_ns["context"]

  if mod.slice_suffix:
    sname = "%s-%s" % (uhm._uid, mod.slice_suffix)
    if not userHasSliceCred(context, sname):
      uhm.setLabProgressLabel("Issuing new slice credential...")
      user_cert = creds.getUserCert(uhm._uid)
      sdata = {}
      sdata[uhm._course] = { uhm._uid : [sname, user_cert]}
      issueUserSliceCred(uhm._course, sdata)
  else:
    sname = uhm._slice

  uhm.setLabProgressLabel("Running createsliver...")
  m = uhm._site.createsliver(context, sname, r)

  if mod.save_manifest:
    data = json.dumps({"manifest" : m.text})
    requests.post("http://localhost:9008/module/%s/manifest/%s" % (mod.modid, context.userurn),
                  data, **rkwargs)

  if mod.post_build:
    uhm.setLabProgressLabel("Post-build setup...")
    postbuild(mod, vardata)

  uhm.addSliceInfo(sname, uhm._site, m)
  uhm.setLabProgressLabel("Finished dependent build.")


def buildnet (dropdown, modlist, uhm, _):
  from . import util
  _ip = get_ipython()

  if dropdown.value[0] == "*":
    value = dropdown.value[3:]
  else:
    value = dropdown.value

  context = _ip.user_ns["context"]

  for mod in modlist:
    if mod.uiname == value:
      Logger.log("%%lab: Building %s" % (value))

      uhm.mod = mod
      uhm.resetLabProgressBar()

      if mod.depends:
        for ssuf in mod.depends:
          sname = "%s-%s" % (uhm._uid, ssuf)
          if userHasSliceCred(context, sname) and sliverExists(sname, uhm):
            continue
          else:
            uhm.setLabProgressLabel("Building required dependency (%s)..." % (ssuf))
            depBuild(ssuf, uhm)


      (r, vardata) = build(mod)
      existing = False

      if mod.pre_build:
        uhm.setLabProgressLabel("Pre-build setup...")
        prebuild(mod, vardata, uhm)

      try:
        if mod.slice_suffix:
          sname = "%s-%s" % (uhm._uid, mod.slice_suffix)
          if not userHasSliceCred(context, sname):
            uhm.setLabProgressLabel("Issuing new slice credential...")
            user_cert = creds.getUserCert(uhm._uid)
            sdata = {}
            sdata[uhm._course] = { uhm._uid : [sname, user_cert]}
            issueUserSliceCred(uhm._course, sdata)
        else:
          sname = uhm._slice

        uhm.setLabProgressLabel("Running createsliver...", True)

        m = uhm._site.createsliver(context, sname, r)

        if mod.save_manifest:
          data = json.dumps({"manifest" : m.text})
          requests.post("http://localhost:9008/module/%s/manifest/%s" % (mod.modid, _ip.user_ns["context"].userurn),
                        data, **rkwargs)
        _ip.push({"manifest" : m})
        uhm.manifest = m
        uhm.incrementLabProgress()
        if mod.post_build:
          uhm.setLabProgressLabel("Post-build setup...")
          postbuild(mod, vardata)
      except CreateSliverError:
        uhm.setLabProgressLabel("Getting existing network...")
        m = uhm._site.listresources(_ip.user_ns["context"], sname)
        if mod.topo_name and m.toponame == mod.topo_name:
          _ip.push({"manifest" : m})
          uhm.manifest = m
#          sys.exc_clear()
        else:
          Logger.log("%lab: Mismatched topology found")
          uhm.labProgressFail("ERROR: Existing topology found. You must delete it before you can make a new topology.")
          return

      try:
        ss = uhm._site.sliverstatus(uhm.context, sname)
        uhm.sliverstatus = ss
        if mod.save_manifest and not existing:
          requests.post("http://localhost:9008/module/%s/status/%s" % (mod.modid,
                        _ip.user_ns["context"].userurn), json.dumps({"status" : json.dumps(ss)}), **rkwargs)
      except Exception:
        Logger.log("%lab: Failed to post sliverstatus")

      uhm.setLabProgressLabel("Writing SSH Config", True)
      util.writeSliceConfig(sname, m)

      uhm.addSliceInfo(sname, uhm._site, m)

#      lbl = widgets.Label(value="Success!")
#      display(lbl)

      Logger.log("%lab: Success")
      uhm.labProgressSuccess("Success!")
      return

def build (modspec):
  mod = compile(modspec.request, '', 'exec')
  _ip = get_ipython()
  env = {}
  env.update(globals())
  env.update(locals())
  env.update(_ip.user_ns)
  vdata = {}
  if modspec.var_pool:
    vdata = random.choice(modspec.var_pool)
    env.update(vdata)

  rstate = random.getstate()
  eval(mod, env) # pylint: disable=W0123
  random.setstate(rstate)

  _ip.push({"request" : env["r"]})

  return (env["r"], vdata)

def prebuild (modspec, vdata, uhm):
  mod = compile(modspec.pre_build, '', 'exec')
  _ip = get_ipython()
  env = {}
  env.update(globals())
  env.update(locals())
  env.update(_ip.user_ns)
  env.update(vdata)

  rstate = random.getstate()
  eval(mod, env) # pylint: disable=W0123
  random.setstate(rstate)

  if "tools" in env:
    uhm._tools = env["tools"]

def postbuild (modspec, vdata):
  mod = compile(modspec.post_build, '', 'exec')
  _ip = get_ipython()
  env = {}
  env.update(globals())
  env.update(locals())
  env.update(_ip.user_ns)
  env.update(vdata)

  rstate = random.getstate()
  eval(mod, env) # pylint: disable=W0123
  random.setstate(rstate)


@magics_class
class UHMagics(Magics):
  def __init__ (self, shell):
    super(UHMagics, self).__init__(shell)
    from . import util

    self.__site = None
    self.__course = None
    self.__uid = None
    self.__ordered = False
    self._lsurl = "http://localhost:9008"
    self._lsp = None
    self.context = get_ipython().user_ns["context"]
    self.manifest = None
    self.sliverstatus = None
    self.push({"_uhm" : self})
    self._cur_em = None
    self._cur_ex = None
    self._cur_pc = None
    self._toponame = None
    self._admin = False
    self._pdate = None
    self._viewroot = getSiteConfig("view-root")

    self._pbar = None
    self._plbl = None
    self._sliceinfo = {}

    # Default slice to username if not supplied
    self.__slice = self.context.uname
    self.push({"SLICE" : self._slice})
    self.push({"NET" : net.NetworkDispatcher(self)})

    # Place holder for tools
    self._tools = {}

  @property
  def toponame (self):
    if self._toponame:
      return self._toponame
    return self.manifest.toponame

  @property
  def _course (self):
    if not self.__course:
      from . import util
      (course, uid, url) = util.getLabCourse()
      self.__course = course
      self.__uid = uid
      self._lsurl = url
      self._lsp = LabserveProxy(self.__course, baseurl = self._lsurl)
    return self.__course

  @_course.setter
  def _course (self, val):
    self.__course = val
    self._lsp = LabserveProxy(val, baseurl = self._lsurl)

  @property
  def _uid (self):
    if not self.__uid:
      from . import util
      (course, uid, url) = util.getLabCourse()
      self.__uid = uid
    return self.__uid

  @property
  def _slice (self):
    return self.__slice

  @_slice.setter
  def _slice (self, val):
    self.__slice = val
    self.push({"SLICE" : self.__slice})

  @property
  def _site (self):
    if not self.__site:
      site = pickSite(self.context.userurn, self._course)
      if isinstance(site, (six.text_type)):
        self._setsite(site)
      elif isinstance(site, (tuple,list)):
        self._setExplicitSite("vts", site[0], site[1])
    return self.__site

  def push (self, attrs): # pylint: disable=R0201
    _ip = get_ipython()
    _ip.push(attrs)

  def _setsite (self, sitename):
    nd = VTSAM.name_to_aggregate()
    self.__site = nd[sitename]
    self.push({"SITE" : self._site})

  def _setExplicitSite (self, stype, sitename, sitehost):
    if stype == "vts":
      self.__site = VTSAM.VTS(sitename, sitehost)
    elif stype == "ig":
      self.__site = IGAM.IGCompute(sitename, sitehost)
    self.push({"SITE" : self.__site})

  def checkExam (self):
    url = "http://localhost:9008/course/%s/exam/%s" % (self._course, self.context.userurn)
    if self._admin:
      if self._pdate:
        pdate = dateutil.parser.parse(self._pdate)
      else:
        pdate = datetime.datetime.now()
      exdata = requests.post(url,
                             json.dumps({"admin" : self.context.userurn,
                                         "date" : pdate.strftime("%Y%m%d%H%M")}),
                             **rkwargs).json()
    else:
      exdata = requests.get(url).json()
    if exdata["value"] == None:
      return False
    return exdata["retcode"] == 0

  def getExamInfo (self):
    url = "http://localhost:9008/course/%s/exam/%s" % (self._course, self.context.userurn)
    if self._admin:
      if self._pdate:
        pdate = dateutil.parser.parse(self._pdate)
      else:
        pdate = datetime.datetime.now()
      exdata = requests.post(url,
                             json.dumps({"admin" : self.context.userurn,
                                         "date" : pdate.strftime("%Y%m%d%H%M")}),
                             **rkwargs).json()
    else:
      exdata = requests.get(url).json()

    if exdata["retcode"] != 0:
      raise LabserveError(exdata["msg"])

    exinfo = exdata["value"]
    if exinfo:
      exinfo["exam"] = True
    return exinfo

  def getCurrentExercise (self, course, sdate):
    datestr = sdate.strftime("%Y%m%d%H%M")
    url = "http://localhost:9008/exercises/%s/current/%s/user/%s" % (course, datestr,
                                                                     self.context.userurn)
    exinfo = requests.get(url).json()["value"]
    return exinfo

  def getExerciseList (self, course, pdate, include_current = False):
    if not pdate:
      pdate = datetime.datetime.now()
    else:
      pdate = dateutil.parser.parse(pdate)

    if include_current:
      pdate += datetime.timedelta(days=7)

    pdatestr = pdate.strftime("%Y%m%d%H%M")
    url = "http://localhost:9008/exercises/%s/stop-before/%s/user/%s" % (course, pdatestr,
                                                                         self.context.userurn)
    exlist = requests.get(url).json()["value"]
    exlist.sort(key=lambda x: datetime.datetime.strptime(x["hard-stop"], "%Y-%m-%d %H:%M"))
    return exlist

  def labProgressFail (self, value):
    self._plbl.value = value
    self._pbar.bar_style = "danger"

  def labProgressSuccess (self, value):
    self._plbl.value = value
    self._pbar.value = self._pbar.max
    self._pbar.bar_style = "success"

  def setLabProgressLabel (self, value, inc = False):
    self._plbl.value = value
    if inc:
      self.incrementLabProgress()

  def incrementLabProgress (self):
    self._pbar.value = self._pbar.value + 1

  def resetLabProgressBar (self, new_box = False):
    if self._pbar:
      self._pbar.bar_style = "info"
      self._pbar.value = 0
      self._plbl.value = ""
    else:
      self._pbar = widgets.IntProgress(value = 0, min = 0, max = 5, step = 1, description = "",
                                       bar_style = "info", orientation = "horizontal")
      self._plbl = widgets.Text("", layout = widgets.Layout(width="100%"))

  @line_magic
  def lab (self, _):
    Logger.log("%lab: Displaying picker")
    box = labpicker(self._course, self)
    display(box)

  @line_magic
  def dns (self, _):
    Logger.log("%%dns: Experiencing Awesomeness")
    r = requests.get("http://localhost:9008/datapool/dns-dplmod/acquire/%s" % (self.context.userurn))
    ret = r.json()
    if ret["retcode"] != 0:
      raise LabserveError(ret["msg"])
    dpdata = ret["value"]
    dplist = []
    for k,v in dpdata.items():
      if k == "index":
        Logger.log("%%dns: dpindex %d" % (v))
      elif v[1] == "json":
        dplist.append(json.loads(v[0]))
    zdp = ZonesDatapool(dplist)
    ie = DNSIE(zdp)
    return ie.display()

  @line_magic
  def number (self, _):
    Logger.log("%%number: You'd think this would be better than dns")
    ie = NUMIE()
    return ie.display()

  @line_magic
  def scenario (self, line):
    parts = line.strip().split()
    scen_name = None
    if len(parts) == 2:
      scen_name = parts[1]

    SMGR.uhm = self
    box = SMGR.picker(scen_name)
    display(box)

  @line_magic
  def perspective_date (self, line):
    if self._admin:
      self._pdate = line

  @line_magic
  def site (self, line):
    parts = line.strip().split()
    if len(parts) == 1:
      Logger.log("Setting site to %s" % (parts[0]))
      self._setsite(parts[0])
    elif len(parts) == 3:
      Logger.log("Setting explicit site: %s" % (line))
      self._setExplicitSite(*parts)

  @line_magic
  def slice (self, line):
    self._slice = line
    Logger.log("User set slice to %s" % (line))
    self.push({"SLICE" : self._slice})

  @line_magic
  def delete (self, _):
    Logger.log("%%delete (%s): %s" % (self._site.name, self._slice))
    self._site.deletesliver(self.context, self._slice)

  @line_magic
  def course (self, line):
    Logger.log("Setting course: %s" % (line))
    self._course = line

  @line_magic
  def exercise (self, _):
    if self._admin and self._pdate:
      pdate = dateutil.parser.parse(self._pdate)
    else:
      pdate = datetime.datetime.now()

    if self._cur_em:
#      if self._cur_em.topo_name == getCurrentExercise(self._course, pdate).get(["topo-name"]):
      Logger.log("%exercise: Redisplaying current exercise module")
      return self._cur_em.display()

    exinfo = self.getCurrentExercise(self._course, pdate)

    if not exinfo:
      Logger.log("%%exercise: No exercise found!")
      return widgets.Label("There is no exercise available.")
    self._cur_em = ExerciseManager(exinfo, self._exsubmit)
    Logger.log("%%exercise: displaying exercise for %s" % (self._cur_em.topo_name))
    return self._cur_em.display()

  @line_magic
  def practice (self, _):
    Logger.log("%practice")
    if self.checkExam():
      Logger.log("%practice: Exam period started, request refused")
      return widgets.Label("You may not access practice during an exam period")
    exlist = self.getExerciseList(self._course, self._pdate)
    self._cur_pc = PracticeController(exlist)
    return self._cur_pc.display()

  @line_magic
  def exam (self, _):
    Logger.log("%exam")
    if self._cur_ex:
      return self._cur_ex.display()
    exinfo = self.getExamInfo()
    if not exinfo:
      return widgets.Label("There is no exam currently available for this course.")
    self._cur_ex = ExerciseManager(exinfo, self._exsubmit)
    return self._cur_ex.display()

  def _exsubmit (self, exm, button):
    Logger.log("%exsubmit button triggered")
    _ip = get_ipython()
    answers = {}
    for k,v in exm._qpanel_data.items():
      answers[k] = []
      for row in v:
        if hasattr(row, "getAnswerData"):
          answers[k] = {"index" : row.dpindex, "values" : row.getAnswerData()}
        else:
          row_info = {}
          answers[k].append(row_info)
          for name,widget in row.items():
            if isinstance(widget, dict):
              row_info[name] = widget
            else:
              row_info[name] = widget.value
    if exm.exam:
      requests.post("http://localhost:9008/exam/%s/user/%s" % (
        exm.uuid, self.context.userurn), json.dumps(answers), **rkwargs)
    else:
      try:
        requests.post("http://localhost:9008/module/%s/exercise/%s/user/%s" % (
          self.mod.modid, exm.uuid, self.context.userurn), json.dumps(answers), **rkwargs)
      except AttributeError:
        requests.post("http://localhost:9008/module/%s/exercise/%s/user/%s" % (
          "0000-0000-0001", exm.uuid, self.context.userurn), json.dumps(answers), **rkwargs)
    if exm.exam:
      button.description = "Resubmit"
    else:
      button.description = "Already Submitted"
      button.disabled = True

  @line_magic
  def tools (self, line):
    _ip = get_ipython()
    if not self._tools:
      return "No tools defined"
    if line in self._tools:
      res = self._tools[line](self)
      data = json.dumps(res)
      requests.post("http://localhost:9008/module/%s/extra/%s" %
        (self.mod.modid, _ip.user_ns["context"].userurn), data, **rkwargs)
    return None

  @line_magic
  def next (self, line):
    # Post current for grading (what about exams?)
    # Get info on next thing
    Logger.log("%next")
    ready = requests.get("http://localhost:9008/course/name/%s/current-attempt/%s" % (
      self._course, self.context.userurn)).json()["value"]
    if not ready:
      Logger.log("%next: not ready for next stage")
      print("You may not move on to the next stage until you have attempted the current task.")
      return

    info = requests.get("http://localhost:9008/course/name/%s/next/%s" % (
      self._course, self.context.userurn)).json()["value"]
    if info["type"] == ".xrcs":
      print("Your next task is an exercise (%s)" % (info["ui-name"]))
    elif info["type"] == ".exam":
      print("Your next task is an EXAM.")
      if info["time-limit"]:
        print("There is a time limit of %d minutes to complete it once you load the exam with %exam." %
          (info["time-limit"]))
    self._cur_em = None
    self._cur_ex = None

  @line_magic
  def status (self, line):
    Logger.log("%%status: %s" % (self.context.userurn))
    info = requests.get("http://localhost:9008/course/name/%s/status/%s" % (self._course,
                                                                       self.context.userurn))
    data = info.json()["value"]
    cols = ["Stage", "Status"]
    row = "<tr><td><div style='text-align:left'>{name}</div></td><td>{status}</td></tr>"
    rowdata = []
    for idx,item in enumerate(data["order-summary"]):
      rd = {"status":""}
      if idx == data["user-pos"]:
        rd["status"] = "In Progress"
      elif idx < data["user-pos"]:
        if item["type"] == ".exam":
          rd["status"] = "Complete"
        else:
          scores = self.getExerciseScores(item["uuid"])
          rd["status"] = "Complete - (%s)" % (", ".join(["%0.2f"% (x) for x in scores]))

      if item["type"] == ".exam":
        rd["name"] = "Exam"
      elif item["type"] == ".xrcs":
        rd["name"] = item["ui-name"]
      rowdata.append(rd)

    lt = ListTable(cols, row, rowdata)
    lt.border = True

    return lt

  def getExamScores(self, euuid):
    xrcs = None
    for xrcs in self._lsp.exams:
      if xrcs["uuid"] == euuid:
        break

    if not xrcs:
      Logger.log("%getExerciseScores: uuid %s not found for course %s" % (euuid, self._lsp.course))
      return []

    akey = self._lsp.getAnswerKey(xrcs["akey-name"])
    atts = self._lsp.getAttempts(euuid, self.context.userurn, False)
    data = buildGraderData(self._lsp, xrcs, self.context.userurn, atts[-1][1])
    scores = grader.client.evalQuestions(data, akey, atts[-1][1])
    keys = sorted(scores.keys())
    return [scores[x]["score"] for x in keys]

  def getExerciseScores(self, euuid):
    xrcs = None
    for xrcs in self._lsp.exercises:
      if xrcs["uuid"] == euuid:
        break

    if not xrcs:
      Logger.log("%getExerciseScores: uuid %s not found for course %s" % (euuid, self._lsp.course))
      return []

    akey = self._lsp.getAnswerKey(xrcs["akey-name"])
    atts = self._lsp.getAttempts(euuid, self.context.userurn, False)
    data = buildGraderData(self._lsp, xrcs, self.context.userurn, atts[-1][1])
    scores = grader.client.evalQuestions(data, akey, atts[-1][1])
    keys = sorted(scores.keys())
    return [scores[x]["score"] for x in keys]

  def addSliceInfo (self, sname, site, manifest):
    sliver_uuid = None
    for dp in manifest.datapaths:
      sliver_uuid = dp.sliver_id.split("+")[-1]
      break
    if not sliver_uuid:
      for ctr in manifest.containers:
        sliver_uuid = ctr.sliver_id.split("+")[-1]
        break
    sinfo = SliceInfo(sname)
    sinfo.sliver_uuid = sliver_uuid
    sinfo.manifest = manifest

    self._sliceinfo[(sname, site.name)] = sinfo

  def getSliceInfo (self, sname, site = None):
    if not site:
      site = self._site

    if (sname, site.name) not in self._sliceinfo:
      if not userHasSliceCred(self.context, sname):
        return None

      try:
        m = site.listresources(self.context, sname)
        self.addSliceInfo(sname, site, m)
      except Exception:
        return None

    return self._sliceinfo[(sname, site.name)]

  def clearSliceInfo (self, sname, site = None):
    if not site:
      site = self._site

    if (sname, site.name) in self._sliceinfo:
      del self._sliceinfo[(sname, site.name)]


class SliceInfo(object):
  def __init__ (self, sname):
    self.name = sname
    self.manifest = None
    self.sliver_uuid = None


def buildGraderData (lsp, exercise, user, attempt):
  data = {}
  for qid,val in enumerate(exercise["question-data"], 1):
    qid = str(qid)
    use_vlans = val.get("use-vlans", False)
    if "index" in attempt[qid]:
      data[qid] = lsp.getDatapoolIndex(val["data-pool"], attempt[qid]["index"], use_vlans)
    else:
      data[qid] = lsp.getDatapoolIndex(val["data-pool"], attempt[qid][0]["_uidata"]["index"], use_vlans)
  return data


VPNCONFIG = """client
proto %(protocol)s
dev tap
remote %(local-ip)s %(tp-port)d
mute 20
cipher AES-256-CBC

<secret>
%(key)s
</secret>
"""

def buildVPNConfig (uhm, name):
  m = uhm._site.listresources(uhm.context, uhm._slice)
  cid = None
  for vpn in m.functions:
    if isinstance(vpn, geni.rspec.vtsmanifest.SSLVPNFunction):
      if (vpn.client_id == name) or (vpn.note == name):
        cid = vpn.client_id
        break
  if not cid:
    print("Unknown VPN name")

  info = uhm._site._apiv3.poa(uhm.context, uhm._site.urlv3, uhm._slice, "vts:vpn:get-info",
                             options = {"client-id" : cid})
  info["key"] = "\n".join([x for x in info["key"].split("\n") if (x and not x[0] == "#")])
  print(VPNCONFIG % (info))

setattr(edmodule, "buildVPNConfig", buildVPNConfig)


def labpicker (course, uhm = None):
  if not uhm:
    ml = courseinfo("none", course)
  else:
    ml = courseinfo(uhm.context.userurn, course, uhm._admin, uhm._pdate)
  ml.sort(key=lambda x: x.sort)

  labels = []
  for mod in ml:
    if mod.instructor_only:
      labels.append("** %s" % (mod.uiname))
    else:
      if mod.uiname:
        labels.append(mod.uiname)

  dd = widgets.Dropdown(options=labels)
  if uhm and uhm._site and uhm._slice:
    sb = widgets.Button(description="Build Network")
    sb.on_click(functools.partial(buildnet, dd, ml, uhm))
    uhm.resetLabProgressBar()
    pbox = widgets.VBox([uhm._pbar, uhm._plbl], layout = widgets.Layout(width="100%"), width='100%')
    outerbox = widgets.VBox([dd, sb, pbox])
  else:
    sb = widgets.Button(description="Select Network")
    sb.on_click(functools.partial(selectnet, dd, ml))
    outerbox = widgets.VBox([dd, sb])

  return outerbox

setattr(edmodule, "labpicker", labpicker)

def buildHSCyberContext ():
  import geni.util
  import geni._coreutil as GCU

  keygen = geni.util._find_ssh_keygen()
  subprocess.call("%s -t rsa -b 2048 -f ~/.ssh/cloudlab_rsa -N ''" % (keygen), shell = True)
  pkpath = os.path.expanduser("~/.ssh/cloudlab_rsa.pub")

  cert_path = os.path.expanduser("~/cloudlab.pem")
  user_urn = parseCert(cert_path)
  username = user_urn.split("+")[-1]

  geni.util._buildContext("emulab-ch2", cert_path, cert_path, username, user_urn, pkpath,
                          "uh-hs-cyber", "%s/context.json" % (GCU.getDefaultDir()))


# Copied from build-context
def parseCert (certpath):
  # Right now this just returns the user URN, but may be extended to get any info we need out of the cert
  from cryptography import x509
  from cryptography.hazmat.backends import default_backend

  cert = x509.load_pem_x509_certificate(open(certpath, "rb").read(), default_backend())

  san = cert.extensions.get_extension_for_oid(x509.OID_SUBJECT_ALTERNATIVE_NAME)
  uris = san.value.get_values_for_type(x509.UniformResourceIdentifier)
  for uri in uris:
    if uri.startswith("urn:publicid"):
      return uri
  return None


def load_ipython_extension (ipy):
  import geni.util
  from . import util

  course = None
  course,uid,url = util.getLabCourse()
  if not geni.util.hasDataContext():
    path = os.path.expanduser("~/omni.bundle")
    if os.path.exists(path):
      geni.util.buildContextFromBundle(path, pubkey_path = geni.util.MAKE_KEYPAIR)
    else:
      if hasCredentials(course):
      # userurn = urn:publicid:IDN+ch.<project>.local+user+<username>
        project = course
        user_cert = issueUserCred(course, uid, project, "ch.%s.local" % (project))
        slicename = uid
        slice_data = {}
        slice_data[project] = { uid : [slicename, user_cert] }
        issueUserSliceCred(course, slice_data)

  if not geni.util.hasDataContext():
    raise NoContextError()

  import geni._coreutil
  imports = geni._coreutil.shellImports()
  imports["uhed"] = edmodule
  imports["genish"] = gsh

  tries = 0
  need_pw = False

  try:
    context = geni.util.loadContext(key_passphrase = "ffffffffff")
  except TypeError:
    context = geni.util.loadContext()
  except KeyDecryptionError:
    need_pw = True

  if need_pw:
    while True:
      pw = getpass.getpass("Private Key Passphrase: ")
      tries += 1
      try:
        imports["context"] = geni.util.loadContext(key_passphrase = pw)
        break
      except KeyDecryptionError:
        if tries < 3:
          continue
        break
      except IOError:
        break
  else:
    imports["context"] = context

  ipy.push(imports)
  ipy.register_magics(UHMagics)
  ipy.set_custom_exc((AMError,LabserveError), am_exc_handler)
  Logger.log("Extension loaded for course %s" % (course))
