# Copyright (c) 2018-2020  University of Houston

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import print_function

import json
import os
import subprocess

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa

def output (data):
  """Quick whitespace stripper"""
  if data.strip():
    print(data)

def call (cmd, show_output = True, user = None):
  """Call `cmd`, optionally printing both the command and its output.  Run as `user` if
    specified."""
  if user:
    if user == "root" and os.getuid() != 0:
      cmd = "sudo -H %s" % (cmd)
    else:
      cmd = "sudo -u %s %s" % (user, cmd)

  #if show_cmd:
  #  print(cmd)

  p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
  (sout, serr) = p.communicate()
  if show_output:
    output(sout)
    output(serr)
  return sout

def addUserEnv (uname):
  # Make the user directories
  userhome = os.path.expanduser("~%s" % (uname))
  if not os.path.exists("%s/.ssh" % (userhome)):
    call("mkdir -p %s/.ssh/" % (userhome))
  if not os.path.exists("%s/.bssw" % (userhome)):
    call("mkdir -p %s/.bssw/geni/" % (userhome))
  return userhome

def userHome (uname):
  userhome = os.path.expanduser("~%s" % (uname))
  bssw_path = "%s/.bssw/geni/" % (userhome)
  return userhome, bssw_path

def getUserCert (uname):
  userhome, bssw_path = userHome(uname)
  with open("%s/%s-cert.pem" % (bssw_path, uname), "rb") as usercertfile:
    b_user_cert = x509.load_pem_x509_certificate(usercertfile.read(), backend = default_backend())
  user_cert = b_user_cert.public_bytes(serialization.Encoding.PEM)
  return user_cert

def writeUCreds (uname, bundle):
  userhome, bssw_path = userHome(uname)
  with open("%s/%s-cert.pem" % (bssw_path, uname), "w") as f:
    f.write(bundle["usercert"])
  with open("%s/%s-key.pem" % (bssw_path, uname), "w") as f:
    f.write(bundle["certkey"])
  with open("%s/gpo-ch2-%s-usercred.xml" % (bssw_path, uname), "w") as f:
    f.write(bundle["usercred"])

def writeSCreds (uname, bundle, project, sname = None):
  if not sname:
    sname = uname
  userhome, bssw_path = userHome(uname)
  with open("%s/%s-%s-cert.pem" % (bssw_path, project, sname), "w") as f:
    f.write(bundle["slicecert"])
  with open("%s/gpo-ch2-%s-%s-scred.xml" % (bssw_path, project, sname), "w") as f:
    f.write(bundle["slicecred"])

def makeContext (uname, project, ch_url):
  userhome, bssw_path = userHome(uname)
  obj = {"framework" : "gpo-ch2", "project" : project, "user-name" : uname}
  obj["cert-path"] = "%s/.bssw/geni/%s-cert.pem" % (userhome, uname)
  obj["key-path"] = "%s/.bssw/geni/%s-key.pem" % (userhome, uname)
  obj["user-pubkeypath"] = "%s/.ssh/geni_rsa.pub" % (userhome)
  obj["user-urn"] = "urn:publicid:IDN+%s+user+%s" % (ch_url, uname)
  with open("context.json", "w+") as cf:
    cf.write(json.dumps(obj))

  call("mv context.json %s/.bssw/geni/" % (userhome))
  return getUserCert(uname)

def makeSSHKeys (uname):
  userhome = os.path.expanduser("~%s" % (uname))
  pkey = rsa.generate_private_key(public_exponent=65537, key_size=2048, backend=default_backend())
  privpem = pkey.private_bytes(encoding = serialization.Encoding.PEM,
                               format = serialization.PrivateFormat.TraditionalOpenSSL,
                               encryption_algorithm = serialization.NoEncryption())
  with open("geni_rsa", "w+b") as priv_key:
    priv_key.write(privpem)

  pubkey = pkey.public_key()
  pubdata = pubkey.public_bytes(encoding = serialization.Encoding.OpenSSH,
                                format = serialization.PublicFormat.OpenSSH)
  with open("geni_rsa.pub", "w+b") as pubkeyfile:
    pubkeyfile.write(pubdata)

  call("mv geni_rsa* %s/.ssh/" % (userhome))
  call("chmod 0600 %s/.ssh/geni_rsa" % (userhome))
