# Copyright (c) 2020  University of Houston

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import random

class DNSError(Exception): pass

class NXDomain(DNSError):
  def __init__ (self, key):
    super(NXDomain, self).__init__()
    self._key = key

  def __str__ (self):
    return "The key '%s' is not known to this server." % (self._key)

class NoRecordTypeError(DNSError):
  def __init__ (self, key, rtype):
    super(NoRecordTypeError, self).__init__()
    self._key = key
    self._rtype = rtype

  def __str__ (self):
    return "There is no record of type '%s' for key '%s'" % (self._rtype, self._key)


class Nameserver(object):
  def __init__ (self, zonedp, data):
    self.name = data["name"]
    self.has_root = False

    self._data = data
    self._zone_datapool = zonedp
    self._v4address = None
    self._v6address = None
    self._records = {}
    self._loadRecords()

  def _loadRecords (self):
    for domain in self._data["domains"]:
      if domain["name"] == ".":
        self.has_root = True
      for key,records in domain["records"].items():
        for record in records:
          self._records.setdefault(key, {})[record["type"]] = record

  def lookup (self, key, rtype):
    # TODO: Need to segment properly to support SOA records
    if key not in self._records:
      raise NXDomain(key)
    if rtype not in self._records[key]:
      raise NoRecordTypeError(key, rtype)

    rdata = self._records[key][rtype]
    rdata["key"] = key

    if rtype == "NS":
      # Search for glue
      try:  
        newkey = ".%s" % (".".join(rdata["value"].split(".")[::-1]))
        gdata = self.lookup(newkey, "A")
        rdata["additional"] = gdata
      except Exception:
        pass
      
    # TODO: Fill ttl from SOA if not present
    return rdata

  @property
  def v4address (self):
    if not self._v4address:
      rname = ".%s" % (".".join(self.name.split(".")[::-1]))
      for ns in self._zone_datapool:
        try:
          self._v4address = ns.lookup(rname, "A")["value"]
          break
        except Exception:
          continue
    return self._v4address



class ZonesDatapool(object):
  def __init__ (self, nsdatalist):
    self._nsdatalist = nsdatalist
    self._nameservers = {}
    self._root = None
    self._load()

  def _load (self):
    for nsdata in self._nsdatalist:
      self._nameservers[nsdata["name"]] = Nameserver(self, nsdata)

  def __iter__ (self):
    return iter(self._nameservers.values())

  def getNSNames (self):
    return sorted(self._nameservers.keys())

  @property
  def nameservers (self):
    for ns in self._nameservers.values():
      yield ns

  def pickRandomKeys (self, rtype, num = 3, min_len = 3):
    if rtype == "PTR":
      min_len += 2

    valid = []
    for ns in self._nameservers.values():
      for key,rrs in ns._records.items():
        if key.count(".") < min_len:
          continue
        if rtype not in rrs:
          continue
        # TODO: Filter compressed v6 keys by min_len
        valid.append(key)

    ret = set([])
    if valid:
      iters = 0
      while len(ret) < num and iters < num * 10:
        iters += 1
        ret.add(random.choice(valid))
    return ret

  def getRoot (self):
    if not self._root:
      for ns in self._nameservers.values():
        if ns.has_root:
          self._root = ns
    return self._root

  def getNameserverByName (self, name):
    return self._nameservers[name]

  def getNameserverByIPv4 (self, addr):
    for ns in self._nameservers.values():
      if ns.v4address == addr:
        return ns
