# Copyright (c) 2018-2020  University of Houston Networking Lab

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import base64
import datetime
import functools
import getpass
import json
import os
import random

import dateutil.parser
import geni.rspec.vtsmanifest
import ipywidgets as widgets # pylint: disable=F0401
from ipywidgets.embed import embed_minimal_html
from IPython.core.getipython import get_ipython # pylint: disable=F0401
from IPython import display
import requests
import six


from . import inputtable
from .log import Logger
from .panels import PANELLAYOUTS
from .uhwidgets.arptabs import ARPTabManager
from .uhwidgets.packetfields import PacketFieldManager
from .uhwidgets.subnettabs import SubnetTabManager
from .util import LabserveError
from .widgetutils import AnswerManager, Layouts, getMACs, getBridgePorts, getVLANs, getVLANIP, getExternalIP, getSubnetIP

# {
#   "actions" : ["flood", "output", "send-packet"],
#   "require-vlans" : false,
#   "topo-name" : ,
#   "question-data" : [
#       { "type" : "answermanager" | "tables" | "widgets",
#         "title" : "",
#         "subtitle" : "",
#(widgets)"questions" : [
#             ["name", "What is your name", {"type" : "text", "lines" : 1}],
#             ["quest", "What is your quest", {"type" : "text", "lines" : 3}],
#             ["color", "What is your favorite color",
#                   {"type" : "dropdown", "values" : ["Blue", "Green", "Red", "Yellow"]}]
#           ],
#
#

class ExerciseManager(object):
  def __init__ (self, data, submit_callback, practice_question = None):
    self.topo_name = None
    self.uuid = None
    self.global_graph = True
    self.version = 1
    self.time_limit = None
    self.exam = False
    self._exptime = None
    self._practice_question = practice_question

    self._load(data)

    self.uhm = get_ipython().user_ns["_uhm"]
    self.genish = get_ipython().user_ns["genish"]

    self.outerbox = None
    self.wgraph = None
    self.wtab = widgets.Tab()
    self.wsubmit = widgets.Button(description=self.submitText, icon="check", layout=Layouts.AUTOWIDTH)
    self.wsubmit.on_click(functools.partial(submit_callback, self))
    if self.exam:
      if self._exptime:
        self.wetitle = widgets.HTML(value="<h2>Exam expires at %s</h2>" % (self._exptime.strftime("%H:%M, %B %d")))
      else:
        self.wetitle = widgets.HTML(value="<h2>This exam does not expire</h2>")
    elif self.practice:
      self.wetitle = widgets.HTML(value="<h2>Practice Exercise</h2>")
    else:
      self.wetitle = widgets.HTML(value="<h2>Lab Exercises</h2>")

    if self.uhm._viewroot:
      self.w_exbutton = widgets.Button(icon="external-link")
      self.w_exbutton.on_click(self.expandCallback)
    else:
      self.w_exbutton = None

    self._qpanel_data = {}
    self._displaypanels = {}

    self._buildTabs()
    self._buildBox()

  @property
  def submitText (self):
    if self.practice:
      return "Grade"
    else:
      return "Submit"

  @property
  def practice (self):
    return self._practice_question is not None

  def _load (self, data):
    self.version = data["version"]
    if self.version == 1:
      self.layout_engine = data["layout-engine"]
    if self.version >= 1:
      self.topo_name = data.get("topo-name")
      self.uuid = data["uuid"]
      self.qpanels = data["question-data"]
    if self.version >= 2:
      self.global_graph = data["global-graph"]
    if self.version >= 3:
      self.topo_name = data["lab-topo-names"]

    if data.get("exam"):
      self.exam = True
      self.time_limit = data["time-limit"]
      if self.time_limit:
        self._exptime = datetime.datetime.now() + datetime.timedelta(minutes=self.time_limit)
      if data["hard-stop"]:
        self._exptime = dateutil.parser.parse(data["hard-stop"])

  def expandCallback (self, button):
    root = self.uhm._viewroot % ({"user" : getpass.getuser()})
    JSSTR = '<script type="text/javascript">window.open("%s/view/_popouts/current.html", "_blank", "location=no,toolbar=no,width=1000,height=600");</script>' % (root)
    extab = widgets.Tab()
    idxs = sorted(self._displaypanels.keys())
    tabs = []
    for idx in idxs:
      tabs.append(self._displaypanels[idx])
    extab.children = tabs
    for idx,tabnum in enumerate(idxs):
      extab.set_title(idx, str(tabnum))

    pop_path = os.path.expanduser("~/_popouts")
    if not os.path.exists(pop_path):
      os.mkdir(pop_path)

    embed_minimal_html("%s/current.html" % (pop_path), views=[extab], title="Tab Data")
    display.display(display.HTML(JSSTR))

  def _buildTabs (self):
    tabs = []
    cur_panel = 0
    for idx,qpanel in enumerate(self.qpanels, 1):
      if self.practice and idx != self._practice_question:
        continue

      data = self.getPanelData(qpanel)
      qpanel["_data"] = data

      # We generate the dynamically-populated subtitle here and store it in _subtitle,
      # so each type doesn't have to handle this itself.  We have to also store stvars
      # on the panel so that grader can have access to it eventually.
      stvars = {}
      try:
        if isinstance(qpanel["subtitle"], dict):
          self.buildTitleDict(stvars, qpanel, data)
          subtitle = qpanel["subtitle"]["string"] % (stvars)
        else:
          subtitle = qpanel["subtitle"] % (stvars)
        qpanel["_subtitle"] = subtitle
      except KeyError:
        pass  # A few child tabs have custom handling for subtitle
      qpanel["_stvars"] = stvars

      cur_panel += 1
      if qpanel["type"] == "widgets":
        tabs.append(self._buildWidgetTab(qpanel, cur_panel))
      elif qpanel["type"] == "answermanager":
        tabs.append(self._buildAnsMgrTab(qpanel, cur_panel))
      elif qpanel["type"] == "table":
        tabs.append(self._buildTableTab(qpanel, cur_panel))
    self.wtab.children = tabs
    for n in range(len(tabs)):
      self.wtab.set_title(n, str(n+1))

  def getPanelData (self, qpanel):
    if "data-pool" in qpanel:
      data = {}
      data["use-vlans"] = qpanel["use-vlans"]
      r = requests.get("http://localhost:9008/datapool/%s/acquire/%s" % (qpanel["data-pool"],
                                                                         self.uhm.context.userurn))
      ret = r.json()
      if ret["retcode"] != 0:
        raise LabserveError(ret["msg"])
      dpdata = ret["value"]
      # { "<fname-root>" : (data, type), ...} except index, which is just an integer
      for k,v in dpdata.items():
        if k == "index":
          data["index"] = v
        elif k == "listresources":
          data["manifest"] = geni.rspec.vtsmanifest.Manifest(xml=v[0])
        elif v[1] == "json":
          data[k] = json.loads(v[0])
        else:
          data[k] = base64.b64decode(v[0])

    else:
      data = {"index" : None, "manifest" : self.uhm.manifest, "sliverstatus" : self.uhm.sliverstatus,
              "use-vlans" : qpanel["use-vlans"]}

    if "table-args" in qpanel:
      data.update(qpanel["table-args"])

    return data

  def buildTitleDict (self, stvars, qpanel, pdata):
    pickhost = None
    if "host-image-names" in qpanel:
      hosts = [x.client_id for x in pdata["manifest"].containers if x.image in qpanel["host-image-names"]]
    else:
      hosts = [x.client_id for x in pdata["manifest"].containers if len(x.ports) == 1]

    for k,v in qpanel["subtitle"]["vars"].items():
      if v == "$pickhost":
        pickhost = random.choice(hosts)
        stvars[k] = pickhost
        Logger.log("$pickhost = %s" % (pickhost))

    for k,v in qpanel["subtitle"]["vars"].items():
      if k in stvars:
        continue

      if v == "$pickvlanip":
        ip = getVLANIP(pdata["manifest"], pdata["sliverstatus"], pickhost, hosts)
        stvars[k] = ip
        Logger.log("$pickvlanip = %s" % (ip))
      elif v == "$pickexternalip":
        ip = getExternalIP(pdata["manifest"], pdata["sliverstatus"], pickhost, hosts)
        stvars[k] = ip
      elif v == "$picksubnetip":
        ip = getSubnetIP(pdata["manifest"], pdata["sliverstatus"], pickhost, hosts)
        stvars[k] = ip
      elif isinstance(v, dict):
        # TODO: This is too similar to what is in panels.buildKwargs
        dpsrc = pdata[v["source"]]
        stvars[k] = dpsrc
        for attrname in v["attribute"].split("."):
          stvars[k] = stvars[k][attrname]
      elif isinstance(v, (six.text_type)):
        stvars[k] = v


  def _buildWidgetTab (self, qpanel, panel_pos):
    answer_row_list = []

    children = []
    data = qpanel["_data"]
    if "graph-panels" in qpanel:
      children.append(PANELLAYOUTS[qpanel["graph-panels"]["type"]](qpanel["graph-panels"], data).widget())
      self._displaypanels[panel_pos] = children[-1]

    children.append(widgets.HTML(value="<h2>%s</h2><i>%s</i>" % (qpanel["title"], qpanel["_subtitle"])))
    children.append(widgets.HTML(value="<br>"))

    qs = []
    for key,qtxt,wtype in qpanel["questions"]:
      qs.append(widgets.HTML("<b>%s</b>" % (qtxt)))
      if wtype["type"] == "text":
        if wtype["lines"] == 1:
          w = widgets.Text(layout = widgets.Layout(width="80%"))
        elif wtype["lines"] > 1:
          w = widgets.Textarea(layout = widgets.Layout(width="60%"))
      elif wtype["type"] == "dropdown":
        w = widgets.Dropdown(options = wtype["values"], layout = Layouts.AUTOWIDTH)
      if w:
        answer_row_list.append({key:w})
        qs.append(w)
      continue

    answer_row_list[0]["_uidata"] = qpanel["_stvars"]
    answer_row_list[0]["_uidata"]["index"] = data["index"]

    children.append(widgets.VBox(qs))

    self._qpanel_data[panel_pos] = answer_row_list
    return widgets.VBox(children)

  def _buildAnsMgrTab (self, qpanel, panel_pos):
    children = []
    data = qpanel["_data"]
    if "graph-panels" in qpanel:
      children.append(PANELLAYOUTS[qpanel["graph-panels"]["type"]](qpanel["graph-panels"], data).widget())
      self._displaypanels[panel_pos] = children[-1]

    children.append(widgets.HTML(value="<h2>%s</h2><i>%s</i>" % (qpanel["title"], qpanel["_subtitle"])))
    children.append(widgets.HTML(value="<br>"))

    am = AnswerManager(data, qpanel["actions"])
    am.require_vlans = qpanel["use-vlans"]
    am._uidata = qpanel["_stvars"]
    children.append(am.display())

    self._qpanel_data[panel_pos] = [am]
    return widgets.VBox(children)

  def _computeValue (self, obj, data):
    # TODO: Once again, much like panels.buildKwargs
    if isinstance(obj, (six.text_type)):
      if obj == "$pickbridge":
        brlist = [x.client_id for x in data["manifest"].datapaths]
        return random.choice(brlist)
      else:
        return obj
    elif isinstance(obj, dict):
      if not "_type" in obj:
        return obj

      if obj["_type"] == "computed":
        ret = data[obj["source"]]
        for attrname in obj["attribute"].split("."):
          ret = ret[attrname]
        return ret


  def _buildTableTab (self, qpanel, panel_pos):
    children = []
    data = qpanel["_data"]
    if "graph-panels" in qpanel:
      children.append(PANELLAYOUTS[qpanel["graph-panels"]["type"]](qpanel["graph-panels"], data).widget())
      self._displaypanels[panel_pos] = children[-1]

    opts = {}
    if "opts" in qpanel:
      opts = qpanel["opts"]

    if qpanel["table-type"] == "l2-table":
      brname = self._computeValue(qpanel["table-client-id"], data)

      # Store for grader
      qpanel["_stvars"]["pickbridge"] = brname

      cols = []
      cols.append(inputtable.DropdownColumn("Port",
                                            values = getBridgePorts(data["manifest"], brname)))
      cols.append(inputtable.DropdownColumn("MAC Address", values = getMACs(data["manifest"])))
      if qpanel["use-vlans"]:
        if "vlan-input" in opts and opts["vlan-input"] == "manual":
          cols.append(inputtable.IntInputColumn("VLAN", min = 0, max = 4092))
        else:
          cols.append(inputtable.DropdownColumn("VLAN", values = getVLANs(data["manifest"])))
      tbl = inputtable.InputTable(cols, data, inputtable.TableOptions())
      tbl.brname = brname
      self._qpanel_data[panel_pos] = [tbl]


    elif qpanel["table-type"] == "subnets-table":
      tbl = SubnetTabManager(data)
      self._qpanel_data[panel_pos] = [tbl]

    elif qpanel["table-type"] == "arp-table":
      tbl = ARPTabManager(data)
      self._qpanel_data[panel_pos] = [tbl]

    elif qpanel["table-type"] == "packet-fields":
      tbl = PacketFieldManager(data, opts)
      self._qpanel_data[panel_pos] = [tbl]

    tbl._uidata = qpanel["_stvars"]

    # TODO: Duplicated for bridge handling, fix at some point
    if isinstance(qpanel["subtitle"], dict):
      subtitle = qpanel["subtitle"]["string"] % (qpanel["_stvars"])
    else:
      subtitle = qpanel["subtitle"] % (qpanel["_stvars"])
    qpanel["_subtitle"] = subtitle

    children.append(widgets.HTML(value="<h2>%s</h2><i>%s</i>" % (qpanel["title"], qpanel["_subtitle"])))
    children.append(widgets.HTML(value="<br>"))
    children.append(tbl.display())

    return widgets.VBox(children)

  def _buildBox (self):
    _uhm = get_ipython().user_ns["_uhm"]
    genish = get_ipython().user_ns["genish"]

    if self.w_exbutton:
      self.wetitle.layout = widgets.Layout(width="85%")
      title_h = widgets.HBox([self.wetitle, self.w_exbutton])
    else:
      title_h = self.wetitle

    if self.global_graph:
      topo_graph = widgets.Output(layout = {'border': '1px solid black'})
      with topo_graph:
        topo_graph.append_display_data(genish.showtopo(_uhm.manifest, engine=self.layout_engine))
      self.outerbox = widgets.VBox([self.wetitle, topo_graph, self.wtab, self.wsubmit])
    else:
      self.outerbox = widgets.VBox([title_h, self.wtab, self.wsubmit])

  def display (self):
    return self.outerbox
