# Copyright (c) 2017-2021  University of Houston

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import datetime

import graphviz
from IPy import IP

import geni.util

from ..tables import ListTable, AllInterfaces

# lmods will have an api whitelist or blacklist
# Each vertex will be loaded and then the set of methods
# will be computed and then added to the instance

# Base method list is BASE_APIS[image-name] +
# [BASE_APIS[api-name] for api-name in EXTRA_API_MAP[image-name]]

# Applied methods are WHITELIST & BASE_METHODS or
# BASE_METHODS - BLACKLIST

# XXX: None of the comments up there are true at the moment

def inspector_display (uiname):
  def decorator (target):
    target.uiname = uiname
    return target
  return decorator


class Port(object):
  ID = 0

  def __init__ (self, mobj, net):
    self._id = Port.ID
    Port.ID += 1

    self.client_id = mobj.client_id
    self._mobj = mobj
    self._net = net

  def _pcid_poa (self, name, options = {}):
    opts = {"port-client-id" : self.client_id}
    opts.update(options)
    return self._net.site._apiv3.poa(self._net._context, self._net.site.urlv3,
                                     self._net.slice, name, options = opts)

  def setDown (self):
    return self._pcid_poa("vts:port-down")

  def setUp (self):
    return self._pcid_poa("vts:port-up")



class Datapath(object):
  ID = 0

  def __init__ (self, mobj, net = None):
    self._id = Datapath.ID
    Datapath.ID += 1

    self.client_id = mobj.client_id
    self._net = net

    self.dotshape = "box"
    self.dotfontsize = 12.0


class Container(object):
  ID = 0

  def __init__ (self, mobj, net = None):
    self._id = Container.ID
    Container.ID += 1

    self.client_id = mobj.client_id
    self.image = mobj.image
    self._mobj = mobj
    self._net = net
    self._ports = {}

    self.dotshape = "oval"
    self.dotfontsize = 12.0

  def __hash__ (self):
    return hash(self.client_id)

  def __repr__ (self):
    return "<%d: %s>" % (self._id, self.client_id)

  @property
  def net (self):
    return self._net

  @net.setter
  def net (self, value):
    self._net = value

  def _poa (self, name, options = {}):
    opts = {"client-ids" : [self.client_id]}
    opts.update(options)
    return self._net.site._apiv3.poa(self._net._context, self._net.site.urlv3,
                                     self._net.slice, name, options = opts)

  def _mappoa (self, name, data, options = {}):
    opts = {"client-id-map" : {self.client_id : data}}
    opts.update(options)
    return self._net.site._apiv3.poa(self._net._context, self._net.site.urlv3,
                                     self._net.slice, name, options = opts)

  def __getitem__ (self, name):
    if name not in self._ports:
      self._ports[name] = Port(self._mobj[name], self.net)
    return self._ports[name]


class HostInfoTable(ListTable):
  COLS = ["Interface", "MAC Address"]
  ROW = "<tr><td>{name}</td><td>{macaddress}</td></tr>"
  def __init__ (self, ctr):
    super(HostInfoTable, self).__init__(HostInfoTable.COLS, HostInfoTable.ROW, [])
    for port in ctr._mobj.ports:
      self._data.append({"name":port.name, "macaddress" : port.macaddress})


class ServiceStatusTable(ListTable):
  COLS = ["Service", "Status", "Info"]
  ROW = "<tr><td>{0}</td><td>{1}</td><td>{2}</td></tr>"
  def __init__ (self, data):
    super(ServiceStatusTable, self).__init__(ServiceStatusTable.COLS, ServiceStatusTable.ROW, data)

class TextFileTable(ListTable):
  ROW = '<tr><td style="text-align:left">{0}</td></tr>'
  def __init__ (self, data):
    super(TextFileTable, self).__init__([], TextFileTable.ROW, data)


class AlpineContainer(Container):
  def __init__ (self, mobj, net = None):
    super(AlpineContainer, self).__init__(mobj, net)

  @property
  def hwinfo (self):
    return HostInfoTable(self)

  @inspector_display("Route Table")
  def getRouteTable (self):
    return self._net.site.Host.getRouteTable(self._net._context, self._net.slice, [self.client_id])

  @inspector_display("ARP Table")
  def getARPTable (self):
    return self._net.site.Host.getARPTable(self._net._context, self._net.slice, [self.client_id])

  @inspector_display("Interface Information")
  def getInterfaceInfo (self):
    return self._poa("api:uh.host:get-interface-info")

  @inspector_display("/etc/resolv.conf")
  def _getResolvConf (self):
    res = self._poa("api:uh.host:get-file-contents", options={"file-path" : "/etc/resolv.conf"})
    return TextFileTable(res[self.client_id].split("\n"))

  @inspector_display("iptables-dump")
  def _getIPTables (self):
    res = self._poa("api:uh.host:get-iptables-dump")
    return TextFileTable(res[self.client_id].split("\n"))

  def setIPv4Address (self, interface, address):
    try:
      IP(address.split("/")[0])
    except ValueError as e:
      return str(e)
    return self._poa("api:uh.host:set-ipv4-address",
                     options = {"interface" : interface, "address" : address})

  def addStaticRoute (self, network, nexthop):
    return self._poa("api:uh.host:add-ipv4-route",
                     options = {"network" : network, "gateway" : nexthop})

  @inspector_display("Service Status")
  def svcStatus (self):
    res = self._net.site.Host.svcStatus(self._net._context, self._net.slice, self.client_id)
    data = []
    for row in res[self.client_id].split("\n"):
      parts = row.split(maxsplit=2)
      if parts:
        data.append(parts)
    return ServiceStatusTable(data)
      

  def svcStart (self, service_name):
    return self._net.site._apiv3.poa(self._net._context, self._net.site.urlv3,
                                     self._net.slice, "api:uh.host:supervisor-start",
                                     options={"client-ids" : [self.client_id],
                                              "svc-name" : service_name})

  def svcStop (self, service_name):
    return self._net.site._apiv3.poa(self._net._context, self._net.site.urlv3,
                                     self._net.slice, "api:uh.host:supervisor-stop",
                                     options={"client-ids" : [self.client_id],
                                              "svc-name" : service_name})

  def svcRestart(self, service_name):
    return self._net.site._apiv3.poa(self._net._context, self._net.site.urlv3,
                                     self._net.slice, "api:uh.host:supervisor-restart",
                                     options={"client-ids" : [self.client_id],
                                              "svc-name" : service_name})

class QuaggaContainer(AlpineContainer):
  def __init__ (self, mobj, net = None):
    super(QuaggaContainer, self).__init__(mobj, net)

  @inspector_display("Route Table")
  def getRouteTable (self):
    return self._net.site.IPv4Router.getRouteTable(self._net._context, self._net.slice, [self.client_id])

  def addBGPNeighbor (self, local_as, neighbor, remote_as):
    data = (local_as, neighbor, remote_as)
    return self._mappoa("vts:uh.quagga:add-bgp-neighbor", data)

  def addiBGPNeighbor (self, local_as, neighbor):
    self.addBGPNeighbor(local_as, neighbor, local_as)
    self._setUpdateSource(local_as, neighbor, "lo")

  def _setUpdateSource (self, local_as, neighbor, source):
    data = [(local_as, neighbor, source)]
    return self._mappoa("vts:uh.quagga:set-bgp-neighbor-update-source", data)

  def addBGPNetwork (self, local_as, network):
    data = (local_as, network)
    return self._mappoa("vts:uh.quagga:add-bgp-nets", data)

  def removeBGPNeighbor (self, local_as, neighbor, remote_as):
    data = (local_as, neighbor, remote_as)
    return self._mappoa("vts:uh.quagga:remove-bgp-neighbor", data)

  def removeBGPInstance (self, local_as):
    return self._mappoa("vts:uh.quagga:remove-bgp-session", local_as)

  @inspector_display("BGP Neighbors")
  def getBGPNeighbors (self):
    return self._poa("vts:uh.quagga:get-bgp-neighbors")

  @inspector_display("BGP RIB")
  def getBGPRIB (self):
    result = self._poa("vts:uh.quagga:show-ip-bgp")
    if self.client_id in result:
      return result[self.client_id]

  @inspector_display("OSPF Neighbors")
  def getOSPFNeighbors (self):
    return self._net.site.IPv4Router.getOSPFNeighbors(self._net._context, self._net.slice, [self.client_id])

  def addOSPFNetworks(self, nets):
    return self._poa("vts:uh.quagga:add-ospf-nets", options = {"networks" : nets})

  def setBGPRedistribution (self, local_as, *protocols):
    data = []
    for protocol in protocols:
      data.append((local_as, protocol))
    return self._mappoa("vts:uh.quagga:set-bgp-redist", data)

  def clearBGPRedistribution (self, local_as):
    return self._mappoa("vts:uh.quagga:clear-bgp-redist", local_as)

  def clearOSPFRedistribution (self):
    return self._mappoa("vts:uh.quagga:clear-ospf-redist", None)

  def setOSPFRedistribution (self, protocol):
    return self._mappoa("vts:uh.quagga:set-ospf-redist", [protocol])

class HostRouterContainer(AlpineContainer):
  def __init__ (self, mobj, net = None):
    super(HostRouterContainer, self).__init__(mobj, net)

  def setSNAT (self, internal_subnet, external_interface):
    return self._net.site._apiv3.poa(self._net._context, self._net.site.urlv3,
                                     self._net.slice, "api:uh.hostrouter:set-snat",
                                     options={"client-ids" : [self.client_id],
                                              "int-net" : internal_subnet,
                                              "ext-intf" : external_interface})


class DNSDHCPContainer(AlpineContainer):
  def __init__ (self, mobj, net = None):
    super(DNSDHCPContainer, self).__init__(mobj, net)

  def setSubnet (self, subnet):
    return self._net.site._apiv3.poa(self._net._context, self._net.site.urlv3,
                                     self._net.slice, "api:uh.dhcp:set-subnet",
                                     options={"client-id-map" : {self.client_id : subnet}})
  def setUpstreamDNS (self, server):
    return self._net.site._apiv3.poa(self._net._context, self._net.site.urlv3,
                                     self._net.slice, "api:uh.dhcp:set-upstream-dns",
                                     options={"client-ids" : [self.client_id],
                                              "dns" : server})

  def setHostConfig (self, mac, ip, hostname):
    return self._net.site._apiv3.poa(self._net._context, self._net.site.urlv3,
                                     self._net.slice, "api:uh.dhcp:set-host-configuration",
                                     options={"client-ids" : [self.client_id], "restart" : True,
                                              "mac" : mac, "ipv4-addr" : ip, "hostname" : hostname})

  def setDomain (self, domain):
    return self._net.site._apiv3.poa(self._net._context, self._net.site.urlv3,
                                     self._net.slice, "api:uh.dhcp:set-domain",
                                     options={"client-ids" : [self.client_id],
                                              "name" : domain})

  @inspector_display("/data/dnsmasq.conf")
  def __getDNSMasqConf(self):
    res = self._poa("api:uh.host:get-file-contents", options={"file-path" : "/data/dnsmasq.conf"})
    return TextFileTable(res[self.client_id].split("\n"))


class DNSRootContainer(AlpineContainer):
  def __init__ (self, mobj, net = None):
    super(DNSRootContainer, self).__init__(mobj, net)

  def addResourceRecord (self, rec_name, rec_type, rec_value, rec_ttl = 600):
    return self._net.site._apiv3.poa(self._net._context, self._net.site.urlv3,
                                     self._net.slice, "vts:uh.dnsroot:add-resource-record",
                                     options = {"client-id" : self.client_id,
                                                "record-name" : rec_name,
                                                "record-type" : rec_type,
                                                "record-value" : rec_value,
                                                "record-ttl" : rec_ttl})

  def deleteResourceRecord (self, rec_name, rec_type):
    return self._net.site._apiv3.poa(self._net._context, self._net.site.urlv3,
                                     self._net.slice, "vts:uh.dnsroot:delete-resource-record",
                                     options = {"client-id" : self.client_id,
                                                "record-name" : rec_name,
                                                "record-type" : rec_type})

  @inspector_display("All RRs")
  def getAllResourceRecords (self):
    return self._net.site.getAllDNSResourceRecords(self._net._context, self._net.slice, self.client_id)

  @inspector_display("/data/named.conf")
  def __getNamedConf(self):
    res = self._poa("api:uh.host:get-file-contents", options={"file-path" : "/data/named.conf"})
    return TextFileTable(res[self.client_id].split("\n"))



def _containerFactory (container):
  if container.image == "uh.quagga":
    return QuaggaContainer(container)
  elif container.image == "uh.host-router":
    return HostRouterContainer(container)
  elif container.image == "uh.dnsdhcp":
    return DNSDHCPContainer(container)
  elif container.image == "uh.dnsroot":
    return DNSRootContainer(container)
  else:
    return AlpineContainer(container)


class NetworkDispatcher(object):
  def __init__ (self, uhm):
    self._netslices = {}
    self._uhm = uhm

  def __getitem__ (self, attr):
    if attr not in self._netslices:
      sinfo = self._uhm.getSliceInfo("%s-%s" % (self._uhm._uid, attr))
      self._netslices[attr] = NetworkWrapper(self._uhm, sinfo.manifest, sinfo.name, self)
    return self._netslices[attr]


class NetworkWrapper(object):
  def __init__ (self, uhm, manifest, sname, dispatch = None):
    self._uhm = uhm
    self._manifest = manifest
    self._dispatch = dispatch

    self._context = uhm.context
    self.site = uhm._site
    self.slice = sname

    self._containers = {}
    self._datapaths = {}

    self._next_ss = None
    self._sliverstatus = None

    self._build()

  def __getitem__ (self, key):
    if key in self._containers:
      return self._containers[key]
    elif key in self._datapaths:
      return self._datapaths[key]
    else:
      raise KeyError

  def _build (self):
    for dp in self._manifest.datapaths:
      self._datapaths[dp.client_id] = Datapath(dp, net = self)
    for ctr in self._manifest.containers:
      cobj = _containerFactory(ctr)
      cobj.net = self
      self._containers[cobj.client_id] = cobj

  def delete (self):
    if not self._dispatch:
      print("You cannot delete this sliver through this interface.")
      return
    self.site.deletesliver(self._context, self.slice)
    del self._dispatch._netslices[self.slice.split("-")[-1]]
    self._uhm.clearSliceInfo(self.slice, self.site)

  def ShowInterfaces (self):
    now = datetime.datetime.now()
    if not self._next_ss or now > self._next_ss:
      self._sliverstatus = self.site.sliverstatus(self._context, self.slice)
      self._next_ss = now + datetime.timedelta(minutes=2)
    return AllInterfaces(self._manifest, self._sliverstatus, next_update = self._next_ss)

  def showtopo (self, engine = "neato"):
#    dotstr = geni.util.builddot([self._manifest])
    dotstr = self._builddot()
    g = graphviz.Source(dotstr)
    g.engine = engine
    return g

  def _builddot (self):
    """Constructs a dotfile of the topology described in the passed in manifest list and returns it as a string."""
    # pylint: disable=too-many-branches

    from geni.rspec import vtsmanifest as VTSM
    from geni.rspec.pgmanifest import Manifest as PGM

    dot_data = []
    dda = dot_data.append # Save a lot of typing

    dda("graph {")

    for manifest in [self._manifest]:
      if isinstance(manifest, PGM):
        intf_map = {}
        for node in manifest.nodes:
          dda("\"%s\" [label = \"%s\"]" % (node.sliver_id, node.name))
          for interface in node.interfaces:
            intf_map[interface.sliver_id] = (node, interface)

        for link in manifest.links:
          label = link.client_id
          name = link.client_id

          if link.vlan:
            label = "VLAN\n%s" % (link.vlan)
            name = link.vlan

          dda("\"%s\" [label=\"%s\",shape=doublecircle,fontsize=11.0]" % (name, label))

          for ref in link.interface_refs:
            dda("\"%s\" -- \"%s\" [taillabel=\"%s\"]" % (
              intf_map[ref][0].sliver_id, name,
              intf_map[ref][1].component_id.split(":")[-1]))
            dda("\"%s\" -- \"%s\"" % (name, intf_map[ref][0].sliver_id))


      elif isinstance(manifest, VTSM.Manifest):
        for dp in manifest.datapaths:
          dda("\"%s\" [shape=rectangle];" % (dp.client_id))

        for ctr in manifest.containers:
          dda("\"%s\" [shape=oval];" % (ctr.client_id))

        dda("subgraph cluster_vf {")
        dda("label = \"SSL VPNs\";")
        dda("rank = same;")
        for vf in manifest.functions:
          if isinstance(vf, VTSM.SSLVPNFunction):
            dda("\"%s\" [label=\"%s\",shape=hexagon];" % (vf.client_id, vf.note))
        dda("}")

        edges = set([])
        # TODO: We need to actually go through datapaths and such, but we can approximate for now
        for port in manifest.ports:
          if isinstance(port, VTSM.GREPort):
            pass
          elif isinstance(port, VTSM.PGLocalPort):
            dda("\"%s\" -- \"%s\" [taillabel=\"%s\"]" % (port.dpname, port.shared_vlan,
                                                         port.name))
            dda("\"%s\" -- \"%s\"" % (port.shared_vlan, port.dpname))
          elif isinstance(port, VTSM.InternalPort):
            dp = manifest.findTarget(port.dpname)
            if dp.mirror == port.client_id:
              continue # The other side will handle it, oddly
            # TODO: Handle mirroring into another datapath
            if "%s-%s" % (port.remote_dpname, port.dpname) not in edges:
              rnode = manifest.findTarget(port.remote_dpname)
              rpname = ""
              if rnode:
                for rp in rnode.ports:
                  if rp.client_id == port.remote_client_id:
                    rpname = rp.name
                    break
              dda("\"%s\" -- \"%s\" [taillabel=\"%s\",headlabel=\"%s\"]" % (port.dpname,
                  port.remote_dpname, port.name, rpname))
              edges.add("%s-%s" % (port.dpname, port.remote_dpname))
          elif isinstance(port, VTSM.InternalContainerPort):
            # Check to see if the other side is a mirror into us
            dp = manifest.findTarget(port.remote_dpname)
            if isinstance(dp, VTSM.ManifestDatapath):
              if port.remote_client_id == dp.mirror:
                remote_port_name = port.remote_client_id.split(":")[-1]
                dda("\"%s\" -- \"%s\" [headlabel=\"%s\",taillabel=\"%s\",style=dashed]" % (
                    port.remote_dpname, port.dpname, port.name, remote_port_name))
                continue

            # No mirror, draw as normal
            if "%s-%s" % (port.remote_dpname, port.dpname) not in edges:
              rnode = manifest.findTarget(port.remote_dpname)
              rpname = ""
              if rnode:
                for rp in rnode.ports:
                  if rp.client_id == port.remote_client_id:
                    rpname = rp.name
                    break
              dda("\"%s\" -- \"%s\" [taillabel=\"%s\",headlabel=\"%s\"]" % (port.dpname,
                  port.remote_dpname, port.name, rpname))
              edges.add("%s-%s" % (port.dpname, port.remote_dpname))

          elif isinstance(port, VTSM.VFPort):
            dda("\"%s\" -- \"%s\"" % (port.dpname, port.remote_client_id))
            dda("\"%s\" -- \"%s\"" % (port.remote_client_id, port.dpname))

          elif isinstance(port, VTSM.GenericPort):
            pass
          else:
            continue ### TODO: Unsupported Port Type


    dda("}")

    return "\n".join(dot_data)
