# Copyright (c) 2017-2020  University of Houston

from __future__ import absolute_import, print_function

import copy
import graphviz
import itertools
import math

import IPy
import six

import geni.rspec.vts as VTS
from uhnet.types import DPID, EthernetMAC

# Why IPy doesn't offer this method, I have no idea
def hostiter(net):
  for ip in net:
    if ip == net.net() or ip == net.broadcast():
      continue
    yield ip

class Graph(object):
  def __init__ (self):
    self.name = None
    self.vertices = {} 
    self.edges = []
    self.site = None # Some AM
    self.slicename = None
    self._decorations = set()

    # Clear these before copy
    self._edge_map = {}
    self._path = set()
    self._ipv4macmap = {}

    # Cred stuff if you want to perform actions
    self.context = None
    self.pkey_path = None
    self.pkey_passphrase = None

  def __getitem__ (self, key):
    if not key.count(":"):
      return self.vertices[key]
    vkey = key.split(":")[0]
    v = self.vertices[vkey]
    return v.ports[key]
      
  def getDatapathByDPID (self, dpid):
    for v in self.vertices.values():
      if isinstance(v, L2Datapath):
        if v.dpid == dpid:
          return v

  def getHostByMAC (self, mac):
    for v in self.vertices.values():
      if isinstance(v, ContainerVertex):
        if v.mac == mac:
          return v

  def clearPaths (self):
    self._path = set()

  def clearDecorations (self):
    self._decorations = set()

  def applyManifest (self, manifest):
    if "manifest" in self._decorations:
      return

    self.buildEdgeMap()
    for dp in manifest.datapaths:
      try:
        dpv = self.vertices[dp.client_id]
      except KeyError:
        # TODO: Log me
        print("Skipping missing datapath (%s)" % (dp.client_id))
        continue

      dpv._mdata = dp
      for port in dp.ports:
        p = BridgePort()
        p.vertex = dpv
        p.vertex.ports[port.client_id] = p
        p._mdata = port
        if port.client_id == dp.mirror:
          p.mirror = True

    for cntr in manifest.containers:
      try:
        cntrv = self.vertices[cntr.client_id]
      except KeyError:
        # TODO: Log me
        print("Skipping missing container (%s)" % (cntr.client_id))
        continue

      cntrv._mdata = cntr

#      cntrv.ssh = sshutil.VTSSSHInfo()
#      cntrv.ssh.graph = self
#      cntrv.ssh.hostname = cntr.logins[0].hostname
#      cntrv.ssh.username = cntr.logins[0].username

      for port in cntr.ports:
        p = Interface()
        p.vertex = cntrv
        p.vertex.ports[port.client_id] = p
        p._mdata = port
        p.name = port.name
        p.mac = EthernetMAC(port.macaddress)

    for v in self.vertices.values():
      for port in v.ports.values():
        if port.edge:
          continue
        remote_v = self.vertices[port._mdata.remote_dpname]
        remote_p = remote_v.ports[port._mdata.remote_client_id]
        edge = self.findEdge(v, remote_v, "manifest")
        port.edge = edge
        remote_p.edge = edge

        if edge.vertex_a == v:
          edge.port_a = port
          edge.port_b = remote_p
        elif edge.vertex_b == v:
          edge.port_b = port
          edge.port_a = remote_p
        edge.mark("manifest")

    self._decorations.add("manifest")

  def applySliverStatus (self, sliverstatus = None):
    if not sliverstatus:
      sliverstatus = self.site.sliverstatus(self.context, self.slicename)
    dplist = sliverstatus["vts_datapaths"]
    for dp in dplist:
      v = self.vertices[dp["client-id"]]

      for port in dp["ports"]:
        gp = v.ports[port["client-id"]]
        gp.number = port["number"]
        gp.up = port["up"]

    clist = sliverstatus["vts_containers"]
    for cntr in clist:
      v = self.vertices[cntr["client-id"]]
      for port in cntr["ports"]:
        cp = v.ports[port["client-id"]]
        cp.v4addresses = port["ipv4-addresses"]

  def getIPv4MACMap (self):
    map = {}
    if not self._ipv4macmap:
      for vertex in self.vertices.values():
        if not isinstance(vertex, ContainerVertex):
          continue
        for port in vertex.ports.values():
          if not isinstance(port, Interface):
            continue
          for addr in port.v4addresses:
            map[addr] = port.mac
      self._ipv4macmap = map
    return self._ipv4macmap
    
  def decorateSTPInfo (self, stpinfo = None):
    dplist = [x.client_id for x in self.vertices.values()
                if isinstance(x, L2Datapath)]
    if stpinfo is None:
      stpinfo = self.site.getSTPInfo(self.context, self.slicename, dplist)

    # Handle being run inside notebook
    if isinstance(stpinfo, dict):
      stpinfo = stpinfo.values()

    for obj in stpinfo:
      v = self.vertices[obj["client-id"]]
      v.stp = BridgeSTPInfo()
      v.stp.bridge_id = obj["stp_bridge_id"]
      v.stp.designated_root = obj["stp_designated_root"]
      v.stp.root_path_cost = int(obj["stp_root_path_cost"])

      for pobj in obj["ports"]:
        p = v.ports[pobj["client-id"]]
        if p.mirror:
          continue

        p.stp = PortSTPInfo()
        p.up = pobj["up"]
        info = pobj["info"]
        p.stp.port_id = info["stp_port_id"]
        p.stp.state = info["stp_state"]
        p.stp.secs_in_state = int(info["stp_sec_in_state"])
        p.stp.role = info["stp_role"]
        p.stp.rx_count = info["stp_rx_count"]
        p.stp.tx_count = info["stp_tx_count"]
        p.stp.error_count = info["stp_error_count"]

    self._decorations.add("stp")

  def getDotString (self, add_macs = False, hostlabels = True):
    dot_data = []
    dda = dot_data.append # Save a lot of typing
    dda("graph {")

    for e in self.edges:
      extra = ""
      einfo = {}
      if hostlabels or isinstance(e.port_a, BridgePort) or (len(e.vertex_a.ports) > 1):
        einfo["taillabel"] = e.port_a._mdata.name
      if hostlabels or isinstance(e.port_b, BridgePort) or (len(e.vertex_b.ports) > 1):
        einfo["headlabel"] = e.port_b._mdata.name

      # No STP for mirror ports, indicate their simplex nature
      skipstp = False
      if isinstance(e.port_a, BridgePort):
        if e.port_a.mirror:
          einfo["style"] = "dotted"
          einfo["dir"] = "forward"
          skipstp = True
      if isinstance(e.port_b, BridgePort):
        if e.port_b.mirror:
          einfo["style"] = "dotted"
          einfo["dir"] = "back"
          skipstp = True

      if "stp" in self._decorations and not skipstp:
        states = []
        if e.port_a.stp: states.append(e.port_a.stp.state)
        if e.port_b.stp: states.append(e.port_b.stp.state)

        ttlist = []
        if isinstance(e.vertex_a, L2Datapath):
          if e.port_a.up: state = "Up"
          else: state = "Down"
          ttlist.append("%s\t%s (%s)" % (e.port_a._mdata.client_id,
                                         state, e.port_a.stp.state))
        if isinstance(e.vertex_b, L2Datapath):
          if e.port_b.up: state = "Up"
          else: state = "Down"
          ttlist.append("%s\t%s (%s)" % (e.port_b._mdata.client_id,
                                         state, e.port_b.stp.state))
        einfo["tooltip"] = "\n".join(ttlist)

        if not e.port_a.up or not e.port_b.up:
          pass # This edge is down
        elif "blocking" in states:
          einfo["style"] = "dashed"
          einfo["penwidth"] = "1"
          einfo["color"] = "red"
        elif "learning" in states:
          einfo["style"] = "solid"
          einfo["penwidth"] = "2"
          einfo["color"] = "orange"
        elif "listening" in states:
          einfo["style"] = "dashed"
          einfo["penwidth"] = "2"
          einfo["color"] = "orange"
        elif "forwarding" in states:
          einfo["style"] = "solid"
          einfo["penwidth"] = "2"
          einfo["color"] = "green"

      if self._path:
        if e in self._path:
          einfo["style"] = "solid"
          einfo["penwidth"] = "3"
          einfo["color"] = "blue"

      if not e.port_a.up or not e.port_b.up:
        einfo["style"] = "dotted"
        einfo["penwidth"] = "1"
        einfo["color"] = "black"


      if einfo:
        attrlist = ['%s="%s"' % (k,v) for k,v in einfo.items()]
        extra = """[%s]""" % (",".join(attrlist))

      dda(""""%d" -- "%d" %s""" % (e.vertex_a._id, e.vertex_b._id, extra))

    for v in self.vertices.values():
      nextra = ""
      ninfo = []

      ninfo.append('label="%s",shape=%s,fontsize=%0.2f' % (v.client_id, v.dotshape,
                                                           v.dotfontsize))
#      ninfo.append('href="localhost://#"')
      ninfo.append('tooltip="%s"' % (v.tooltip))
      if ninfo:
        nextra = """[%s]""" % (",".join(ninfo))

      dda(""""%d" %s""" % (v._id, nextra))

    dda("}")
    return "\n".join(dot_data)

  def show (self, engine = "circo", hostlabels = True):
    dot = graphviz.Source(self.getDotString(hostlabels = hostlabels))
    dot.engine = engine
    return dot

  def highlightPath (self, edges):
    self._path = set()
    for edge in edges:
      self._path.add(edge)

  def addVertex (self, vertex):
    self.vertices[vertex.client_id] = vertex
    vertex.graph = self
    return vertex

  def addEdge (self, vertex_a, vertex_b):
    obj = GENIWire()
    obj.vertex_a = vertex_a
    obj.vertex_b = vertex_b
    self.edges.append(obj)

    vertex_a.edges.append(obj)
    vertex_b.edges.append(obj)
    return obj

  def buildEdgeMap (self):
    # { <vertex> : { <vertex> : [edge, ...] } }
    # { <dp0> : { <nf0> : [ e1, e2, e3 ] } }
    # { <nf0> : { <dp0> : [ e1, e2, e3 ] } } 
    self._edge_map = {}
    for edge in self.edges:
      self._edge_map.setdefault(
              edge.vertex_a, {}).setdefault(
                edge.vertex_b, []).append(edge)
      self._edge_map.setdefault(
              edge.vertex_b, {}).setdefault(
                edge.vertex_a, []).append(edge)

  def findEdge (self, vertex_a, vertex_b, unmark_val):
    v1 = self._edge_map[vertex_a]
    edge_list = v1[vertex_b]
    for edge in edge_list:
      if not edge.has_mark(unmark_val):
        return edge
    raise KeyError


class PortSTPInfo(object):
  def __init__ (self):
    self.port_id = None
    self.state = None
    self.secs_in_state = None
    self.role = None

    self.rx_count = None
    self.tx_count = None
    self.error_count = None

class Port(object):
  def __init__ (self):
    self.vertex = None
    self.edge = None
    self._name = None
    self.number = None
    self.stp = None
    self.up = True
    self._mdata = None

  @property
  def name (self):
    return self._name

  @name.setter
  def name (self, value):
    self._name = value

  def setDown (self, context = None):
    g = self.vertex.graph
    if not context:
      context = g.context
    g.site.portDown(context, g.slicename, self._mdata.client_id)

  def setUp (self, context = None):
    g = self.vertex.graph
    if not context:
      context = g.context
    g.site.portUp(context, g.slicename, self._mdata.client_id)


class BridgePort(Port):
  def __init__ (self):
    super(BridgePort, self).__init__()
    self.mirror = False

  @property
  def name (self):
    return self._mdata.name

  @property
  def vlanid (self):
    return self._mdata.vlan_id


class Interface(Port):
  def __init__ (self):
    super(Interface, self).__init__()
    self.mac = None
    self.v4addresses = []


class GENIVertex(object):
  ID = 0

  def __init__ (self, client_id):
    self._id = GENIVertex.ID
    GENIVertex.ID += 1

    self.graph = None
    self.edges = []
    self.client_id = client_id

    # { "dp:pnum" : Port }
    self.ports = {}
    self._mdata = None

    self.dotshape = "oval"
    self.dotfontsize = 12.0

  def __hash__ (self):
    return hash(self.client_id)

  def __repr__ (self):
    return "<%d: %s>" % (self._id, self.client_id)

  def __getitem__ (self, key):
    if isinstance(key, int):
      return self.ports["%s:%d" % (self.client_id, key)]
    elif isinstance(key, (six.text_type)):
      return self.ports["%s:%s" % (self.client_id, key)]

  @property
  def tooltip (self):
    return ""

  def copyUnconnected (self):
    v = self.__class__(self.client_id)
    v.ports = copy.copy(self.ports)
    v._mdata = self._mdata
    return v

  def detach (self):
    for edge in self.edges:
      if edge.vertex_a == self: other_vertex = edge.vertex_b
      elif edge.vertex_b == self: other_vertex = edge.vertex_a

      other_vertex.disconnectEdge(edge)
      self.graph.edges.remove(edge)

  def disconnectEdge (self, edge):
    for port in self.ports.values():
      if port.edge == edge:
        port.edge = None
        break

    self.edges.remove(edge)

  def clearEdges (self):
    self.edges = []


class ContainerVertex(GENIVertex):
  def __init__ (self, client_id):
    super(ContainerVertex, self).__init__(client_id)

  @property
  def tooltip (self):
    rlist = []
    for intf in self.ports.values():
      rlist.append("[%s] %s" % (intf.name, intf.mac))
      for addr in intf.v4addresses:
        rlist.append("%s %s" % (" " * (len(intf.name) + 2), addr))
    return "\n".join(rlist)

  def computeARPTable (self):
    ipmap = self.graph.getIPv4MACMap()
    arptable = []
    for intf in self.ports.values():
      for addr in intf.v4addresses:
        subnet = IPy.IP(addr, make_net = True)
        for k, v in ipmap.items():
          iface_subnet = IPy.IP(k, make_net = True)
          if subnet == iface_subnet and k != addr:
            arptable.append((k.split("/")[0], v, intf.name))
    return arptable

  def generateSpanningTree (self):
    vvids = set()
    ceids = set()
    next_vertices = [self]

    st = Graph()

    while next_vertices:
      cv = next_vertices.pop()
      vvids.add(cv._id)

      if cv.client_id not in st.vertices:
        stv = cv.copyUnconnected()
        st.addVertex(stv)
      else:
        stv = st.vertices[cv.client_id]

      for edge in cv.edges:
        ov = edge.getOther(cv)
        if edge._id in ceids:
          st.addEdge(stv, st.vertices[ov.client_id])

        if ov._id in vvids:
          continue

        ceids.add(edge._id)
        next_vertices.append(other)

    return st


class AlpineContainer(ContainerVertex):
  def __init__ (self, client_id):
    super(AlpineContainer, self).__init__(client_id)

  def _exec (self, cmd):
    ret = self.graph.site.Host.execcmd(self.graph.context, self.graph.slicename, [self.client_id], cmd)
    return ret[self.client_id]

  def ps (self):
    ret = self._exec("ps")
    return ret["stdout"]

  def dhclient (self, intfname):
    ret = self._exec("dhclient -v %s" % (intfname))
    return ret["stderr"]

class QuaggaContainer(AlpineContainer):
  def __init__ (self, client_id):
    super(QuaggaContainer, self).__init__(client_id)


class BridgeSTPInfo(object):
  def __init__ (self):
    self.bridge_id = None
    self.designated_root = None
    self.root_path_cost = None


class L2Datapath(GENIVertex):
  def __init__ (self, client_id):
    super(L2Datapath, self).__init__(client_id)
    self.dotshape = "box"
    self.dpid = None
    self.stp = None

  @property
  def tooltip (self):
    if not self.stp:
      return ""
    rlist = []
    rlist.append("Bridge ID: %s" % (self.stp.bridge_id))
    rlist.append("Designated Root: %s" % (self.stp.designated_root))
    return "\n".join(rlist)

  def createGENIResource (self):
    rsrc = VTS.Datapath(VTS.OVSL2Image(), self.client_id)
    return rsrc

  def _getAttachedMACs (self, pnum = None, vlan = None, up_only = True, hide = []):
    data = []
    for port in self.ports.values():
      if port._mdata.remote_dpname in hide:
        continue
      rp = self.graph[port._mdata.remote_client_id]
      if isinstance(rp, Interface):
        if rp.up or not up_only:
          if pnum:  # We're asking from the view of another bridge
            if vlan:
              data.append((pnum, rp.mac, port.vlanid))
            else:
              data.append((pnum, rp.mac))
          else:
            if vlan:
              data.append((port.name, rp.mac, port.vlanid))
            else:
              data.append((port.name, rp.mac))
    return data

  def computeMACTable (self, limit = 50, include_vlans = False, hide = []):
    data = []
    bridges = [(None, self, None)]
    idx = 0
    while bridges:
      idx += 1
      if idx > limit:
        break

      pnum,br,rpcid = bridges.pop()
      data.extend(br._getAttachedMACs(pnum, vlan = include_vlans, hide = hide))

      for port in br.ports.values():
        if not port.up or port.stp.state != "forwarding" or port._mdata.client_id == rpcid:
          continue
        rv = br.graph[port._mdata.remote_dpname]
        if not isinstance(rv, L2Datapath):
          continue
        if br.graph[port._mdata.remote_client_id].stp.state != "forwarding":
          continue
        if pnum:
          bridges.append((pnum, rv, port._mdata.remote_client_id))
        else:
          bridges.append((port.name, rv, port._mdata.remote_client_id))
    return data

  def getMACTable (self, context = None):
    if self.exercise:
      return None
    else:
      if not context:
        context = self.graph.context
      return self.graph.site.dumpMACs(context, self.graph.slicename, [self.client_id])

  def getSTPInfo (self, context = None):
    if not context:
      context = self.graph.context
    info = self.graph.site.getSTPInfo(context, self.graph.slicename, [self.client_id])
    if isinstance(info, dict):
      return info[self.client_id]
    return info


class Edge(object):
  ID = 0

  def __init__ (self):
    self._id = Edge.ID
    Edge.ID += 1

    self._marks = set()

    self.vertex_a = None
    self.vertex_b = None

  def mark (self, val):
    self._marks.add(val)

  def has_mark (self, val):
    return val in self._marks

  def getOther (self, vertex):
    if self.vertex_a == vertex:
      return self.vertex_b
    elif self.vertex_b == vertex:
      return self.vertex_a
    else:
      # TODO: Raise an exception
      return None

  @property
  def tooltip (self):
    return ""


class GENIWire(Edge):
  def __init__ (self):
    super(GENIWire, self).__init__()
    self.port_a = None
    self.port_b = None


def generateRequest (graph):
  r = VTS.Request()

  rsrc_map = {}
  for v in graph.vertices.values():
    rsrc = v.createGENIResource()
    rsrc_map[v] = rsrc
    r.addResource(rsrc)
   
  for e in graph.edges:
    rsrc_a = rsrc_map[e.vertex_a]
    rsrc_b = rsrc_map[e.vertex_b]
    VTS.connectInternalCircuit(rsrc_a, rsrc_b)

  return r
    
