# Copyright (c) 2018-2020  University of Houston Networking Lab

from __future__ import absolute_import

import copy

import ipywidgets as widgets # pylint: disable=F0401
from IPython.core.getipython import get_ipython # pylint: disable=F0401
import six

from .oldgraph.util import buildFromManifest

from . import tables
from .uhwidgets import graph

PANELTYPES = {}
PANELLAYOUTS = {}

def layout (typ):
  def decorator(k):
    PANELLAYOUTS[typ] = k
    return k
  return decorator

def panel (typ):
  def decorator(k):
    PANELTYPES[typ] = k
    return k
  return decorator


@layout("65-split")
class SplitSixtyFive(object):
  def __init__ (self, spec, data):
    self.lpanel = None
    self.rpanel = None
    self._spec = spec
    self._data = data

    self._buildPanels()

  def _buildPanels (self):
    lpdata = copy.copy(self._data)
    lpdata["_wlayout"] = {"width" : "65%", "border" : "1px solid black"}
    if isinstance(self._spec["lpanel"], (six.text_type)):
      args = self._spec["lpanel"].split(":")
      pkls = PANELTYPES[args[0]]
      self.lpanel = pkls(args[1:], lpdata)
    elif isinstance(self._spec["lpanel"], dict):
      pkls = PANELTYPES[self._spec["lpanel"]["type"]]
      self.lpanel = pkls(self._spec["lpanel"]["arg-sets"], lpdata)

    rpdata = copy.copy(self._data)
    rpdata["_wlayout"] = {"width" : "34%", "border" : "1px solid black"}
    if isinstance(self._spec["rpanel"], (six.text_type)):
      args = self._spec["rpanel"].split(":")
      pkls = PANELTYPES[args[0]]
      self.rpanel = pkls(args[1:], rpdata)
    elif isinstance(self._spec["rpanel"], dict):
      pkls = PANELTYPES[self._spec["rpanel"]["type"]]
      self.rpanel = pkls(self._spec["rpanel"]["arg-sets"], rpdata)

  def widget (self):
    hbox = widgets.HBox([self.lpanel.widget(), self.rpanel.widget()])
    return hbox


@layout("2-row")
class TwoRow(object):
  def __init__ (self, spec, data):
    self.tpanel = None
    self.bpanel = None
    self._spec = spec
    self._data = data

    self._buildPanels()

  def _buildPanels (self):
    tpdata = copy.copy(self._data)
    tpdata["_wlayout"] = {"width" : "100%", "border" : "1px solid black"}
    if isinstance(self._spec["tpanel"], (six.text_type)):
      args = self._spec["tpanel"].split(":")
      pkls = PANELTYPES[args[0]]
      self.tpanel = pkls(args[1:], tpdata)
    elif isinstance(self._spec["tpanel"], dict):
      pkls = PANELTYPES[self._spec["tpanel"]["type"]]
      self.tpanel = pkls(self._spec["tpanel"]["arg-sets"], tpdata)

    bpdata = copy.copy(self._data)
    bpdata["_wlayout"] = {"width" : "100%", "border" : "1px solid black"}
    if isinstance(self._spec["bpanel"], (six.text_type)):
      args = self._spec["bpanel"].split(":")
      pkls = PANELTYPES[args[0]]
      self.bpanel = pkls(args[1:], bpdata)
    elif isinstance(self._spec["bpanel"], dict):
      pkls = PANELTYPES[self._spec["bpanel"]["type"]]
      self.bpanel = pkls(self._spec["bpanel"]["arg-sets"], bpdata)

  def widget (self):
    vbox = widgets.VBox([self.tpanel.widget(), self.bpanel.widget()])
    return vbox


@layout("graph-2column")
class GraphTwoColumn(object):
  def __init__ (self, spec, data):
    self.top = None
    self.cleft = None
    self.cright = None

    self._spec = spec
    self._data = data

    self._buildPanels()

  def _buildPanels (self):
    topdata = copy.copy(self._data)
    topdata["_wlayout"] = {"width" : "100%", "border" : "1px solid black"}
    if isinstance(self._spec["toppanel"], (six.text_type)):
      args = self._spec["toppanel"].split(":")
      pkls = PANELTYPES[args[0]]
      self.top = pkls(args[1:], topdata)
    elif isinstance(self._spec["toppanel"], dict):
      pkls = PANELTYPES[self._spec["toppanel"]["type"]]
      self.top = pkls(self._spec["toppanel"]["arg-sets"], topdata)

    cdata = copy.copy(self._data)
    cdata["_wlayout"] = {"width" : "50%", "border" : "1px solid black"}

    if isinstance(self._spec["cleftpanel"], (six.text_type)):
      args = self._spec["cleftpanel"].split(":")
      pkls = PANELTYPES[args[0]]
      self.cleft = pkls(args[1:], cdata)
    elif isinstance(self._spec["cleftpanel"], dict):
      pkls = PANELTYPES[self._spec["cleftpanel"]["type"]]
      self.cleft = pkls(self._spec["cleftpanel"]["arg-sets"], cdata)

    if isinstance(self._spec["crightpanel"], (six.text_type)):
      args = self._spec["crightpanel"].split(":")
      pkls = PANELTYPES[args[0]]
      self.cright = pkls(args[1:], cdata)
    elif isinstance(self._spec["crightpanel"], dict):
      pkls = PANELTYPES[self._spec["crightpanel"]["type"]]
      self.cright = pkls(self._spec["crightpanel"]["arg-sets"], cdata)

  def widget (self):
    hbox = widgets.HBox([self.cleft.widget(), self.cright.widget()])
    vbox = widgets.VBox([self.top.widget(), hbox])
    return vbox


def coerceBool (val):
  if val.lower().strip() == "true":
    return True
  elif val.lower().strip() == "false":
    return False
  return None

def buildKwargs (aset, data):
  kwargs = {}
  target = None
  if isinstance(aset, (list, tuple)):
    target = aset[0]
    if len(aset) >= 2:
      arg_pairs = aset[1].split(";")
      for pair in arg_pairs:
        k,v = pair.split(",")
        kwargs[k] = v

  elif isinstance(aset, (six.text_type)):
    args = aset.split(":")
    target = args[0]
    if len(args) >= 2:
      arg_pairs = args[1].split(";")
      for pair in arg_pairs:
        k,v = pair.split(",")
        kwargs[k] = v

  elif isinstance(aset, dict):
    target = aset["target"]
    for name,val in aset["args"].items():
      if not isinstance(val, dict):
        kwargs[name] = val
        continue

      if not "_type" in val:
        kwargs[name] = val
        continue

      if val["_type"] == "computed":
        dpsrc = data[val["source"]]
        kwargs[name] = dpsrc
        for attrname in val["attribute"].split("."):
          kwargs[name] = kwargs[name][attrname]

  return target,kwargs
  

@panel("graph")
class GraphPanel(object):
  def __init__ (self, args, data):
    self._type = args[0]
    self._engine = "neato"
    self._hostlabel = True
    self._data = data
    self._parse_args(args)

  def _parse_args (self, args):
    arg_pairs = args[1].split(";")
    for pair in arg_pairs:
      k,v = pair.split(",")
      if k == "engine":
        self._engine = v
      elif k == "host-label":
        self._hostlabel = coerceBool(v)

  def widget (self):
    genish = get_ipython().user_ns["genish"]

    gw = widgets.Output(layout = self._data["_wlayout"])
    if self._type == "default":
      with gw:
        gw.append_display_data(genish.showtopo(self._data["manifest"], engine = self._engine))
    elif self._type == "stp":
      g = buildFromManifest(self._data["manifest"], site = "ignore")
      g.decorateSTPInfo(stpinfo = self._data["api:l2-switch:stp-info"])
      with gw:
        gw.append_display_data(g.show(engine = self._engine, hostlabels = self._hostlabel))

    return gw


@panel("bqgraph")
class BQGraphPanel(object):
  def __init__ (self, args, data):
    self._type = args[0]  # always default for now
    self._engine = "neato"
    self._data = data
    self._parse_args(args)

  def _parse_args (self, args):
    arg_pairs = args[1].split(";")
    for pair in arg_pairs:
      k,v = pair.split(",")
      if k == "engine":
        self._engine = v

  def widget (self):
    posmap = graph.getGraphvizPositions(self._data["manifest"], engine = self._engine)
    w = graph.buildBQWidget(posmap, self._data)
    return w


@panel("mtable")
class ManifestTablePanel(object):
  def __init__ (self, args, data):
    self._data = data
    self._widget = None
    self._build_widget(args)

  def _build_widget (self, args):
    target,kwargs = buildKwargs(args, self._data)

    tkls = tables.MTABLETYPES[target]
    tbl = tkls(self._data["manifest"], **kwargs)

    self._widget = widgets.Output(layout = self._data["_wlayout"])
    with self._widget:
      self._widget.append_display_data(tbl)

  def widget (self):
    return self._widget


@panel("mtables")
class ManifestTablesPanel(object):
  def __init__ (self, arg_sets, data):
    self._data = data
    self._widget = None
    self._build_widget(arg_sets)

  def _build_widget (self, arg_sets):
    self._widget = widgets.Output(layout = self._data["_wlayout"])

    for argstr in arg_sets:
      target,kwargs = buildKwargs(argstr, self._data)
      kwargs["sliverstatus"] = self._data["sliverstatus"]
      if "api:l2-switch:stp-info" in self._data:
        kwargs["stpinfo"] = self._data["api:l2-switch:stp-info"]
      if "api:uh.host:get-route-table" in self._data:
        kwargs["hostroutetables"] = self._data["api:uh.host:get-route-table"]
      if "vts:uh.quagga:get-route-table" in self._data:
        kwargs["rtrroutetables"] = self._data["vts:uh.quagga:get-route-table"]
      if "api:uh.host:get-arp-table" in self._data:
        kwargs["arpdata"] = self._data["api:uh.host:get-arp-table"]

      tkls = tables.MTABLETYPES[target]
      tbl = tkls(self._data["manifest"], **kwargs)

      with self._widget:
        self._widget.append_display_data(tbl)

  def widget (self):
    return self._widget


@panel("pcap-tables")
class PCAPTablesPanel(object):
  def __init__ (self, arg_sets, data):
    self._data = data
    self._widget = None
    self._build_widget(arg_sets)

  def _build_widget (self, arg_sets):
    self._widget = widgets.Output(layout = self._data["_wlayout"])

    for argstr in arg_sets:
      target,kwargs = buildKwargs(argstr, self._data)

      tkls = tables.PCAPTABLETYPES[target]
      tbl = tkls(self._data["pcap-roster"], self._data, **kwargs)

      with self._widget:
        self._widget.append_display_data(tbl)

  def widget (self):
    return self._widget

@panel("ns-tables")
class NSTablesPanel(object):
  def __init__ (self, arg_sets, data):
    self._data = data
    self._widget = None
    self._build_widget(arg_sets)

  def _build_widget (self, arg_sets):
    self._widget = widgets.Output(layout = self._data["_wlayout"])

    for argstr in arg_sets:
      target,kwargs = buildKwargs(argstr, self._data)

      tkls = tables.DNSTABLETYPES[target]
      tbl = tkls(self._data["ns-roster"], self._data, **kwargs)

      with self._widget:
        self._widget.append_display_data(tbl)

  def widget (self):
    return self._widget


@panel("mtable-stack")
class ManifestTableStack(object):
  def __init__ (self, arg_sets, data):
    self._data = data
    self._widget = None
    self._build_widget(arg_sets)

  def _build_widget (self, arg_sets):
    titles = []
    panes = []
    for argstr in arg_sets:
      target,kwargs = buildKwargs(argstr, self._data)
      kwargs["sliverstatus"] = self._data["sliverstatus"]
      if "api:l2-switch:stp-info" in self._data:
        kwargs["stpinfo"] = self._data["api:l2-switch:stp-info"]
      if "api:uh.host:get-route-table" in self._data:
        kwargs["hostroutetables"] = self._data["api:uh.host:get-route-table"]
      if "vts:uh.quagga:get-route-table" in self._data:
        kwargs["rtrroutetables"] = self._data["vts:uh.quagga:get-route-table"]
      if "api:uh.host:get-arp-table" in self._data:
        kwargs["arpdata"] = self._data["api:uh.host:get-arp-table"]

      tkls = tables.MTABLETYPES[target]
      tbl = tkls(self._data["manifest"], **kwargs)
      if isinstance(tbl, list):
        tbl.sort(key=lambda x: x._title)
        for item in tbl:
          titles.append(item._title)
          item._title = None
          tw = widgets.Output()
          with tw:
            tw.append_display_data(item)
          panes.append(tw)
      else:
        titles.append(tbl._title)
        tbl._title = None

        tw = widgets.Output()
        with tw:
          tw.append_display_data(tbl)
        panes.append(tw)

    self._widget = widgets.Accordion(children = panes, layout = self._data["_wlayout"])
    for idx,title in enumerate(titles):
      self._widget.set_title(idx, title)


  def widget (self):
    return self._widget
