# Copyright (c) 2018-2020  University of Houston Networking Lab

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import getpass
import json
import sys

from IPython.core.getipython import get_ipython
from IPython.display import display, Javascript
import ipywidgets as widgets
import requests

from geni.aggregate.apis import CreateSliverError

from . import creds
from .uhwidgets.pbar import ProgressStatusBar

class Scenario(object):
  def __init__ (self, spec):
    self.modid = spec["uuid"]
    self.request = spec["request"]
    self.sort = spec["sort"]
    self.uiname = spec["uiname"]
    self.slicename = spec["slicename"]

    self.pre_build = spec["pre-build"]
    self.post_build = spec["post-build"]
    self.topo_name = spec["topo-name"]


rkwargs = {"headers" : {"Content-Type" : "application/json"},
           "verify" : False}

class _ScenarioManager(object):
  def __init__ (self):
    self._uhm = None
    self._scenarios = {}
    self._advertisement = None

    # widgets
    self._w_picker_dd = None
    self._pbar = None

  @property
  def scenarios (self):
    if not self._scenarios:
      for scenid in requests.get("http://localhost:9008/scenarios/modules").json()["value"]:
        scenario = Scenario(requests.get("http://localhost:9008/scenarios/%s" % (scenid)).json()["value"])
        self._scenarios[scenario.uiname] = scenario
    return self._scenarios.values()

  @property
  def advertisement (self):
    if not self._advertisement:
      self._advertisement = self.uhm._site.listresources(self.uhm.context)
    return self._advertisement

  @property
  def uhm (self):
    return self._uhm

  @uhm.setter
  def uhm (self, val):
    self._uhm = val

  def picker (self, name):
    sl = self.scenarios
    sl.sort(key=lambda x: x.sort)

    labels = [x.uiname for x in sl]
    if name:
      self._w_picker_dd = widgets.Dropdown(options=labels, value=name)
    else:
      self._w_picker_dd = widgets.Dropdown(options=labels)
    sb = widgets.Button(description = "Build Scenario")
    sb.on_click(self.buildnet)

    outerbox = widgets.VBox([self._w_picker_dd, sb])
    return outerbox

  def resetProgress (self):
    if self._pbar:
      self._pbar.reset()
    else:
      self._pbar = ProgressStatusBar(0, 5, "Building: ", "Forming Request")
      self._pbar.display()

  def build (self, modspec):
    mod = compile(modspec.request, '', 'exec')
    _ip = get_ipython()
    env = {}
    env.update(globals())
    env.update(locals())
    env.update(_ip.user_ns)
    env.update({"_AD" : self.advertisement})
    eval(mod, env) # pylint: disable=W0123
    _ip.push({"request" : env["r"]})

    return (env["r"])

  def postbuild (self, modspec):
    mod = compile(modspec.post_build, '', 'exec')
    _ip = get_ipython()
    env = {}
    env.update(globals())
    env.update(locals())
    env.update(_ip.user_ns)
    eval(mod, env) # pylint: disable=W0123


  def buildnet (self, _):
    from . import util
    _ip = get_ipython()

    value = self._w_picker_dd.value
    scen = self._scenarios[value]

    self.resetProgress()

    r = self.build(scen)

    try:
      self.uhm._slice = scen.slicename

      self._pbar.setStatus("Creating scenario...", True)

      user = getpass.getuser()
      data = json.dumps({"user-cert" : creds.getUserCert(user), "project" : self.uhm.context.project})
      slicecred_bundle = requests.post("http://localhost:9008/course/_demo/creds/issue-slice/%s" % (scen.slicename),
                                       data, **rkwargs).json()["value"]
      creds.writeSCreds(user, slicecred_bundle, self.uhm.context.project, scen.slicename)

      m = self.uhm._site.createsliver(self.uhm.context, scen.slicename, r)

      data = json.dumps({"manifest" : m.text})
      requests.post("http://localhost:9008/module/%s/manifest/%s" % (scen.modid, self.uhm.context.userurn),
                    data, **rkwargs)
      _ip.push({"manifest" : m})
      self.uhm.manifest = m

      self._pbar.incrementProgress()

      if scen.post_build:
        self._pbar.setStatus("Post-build setup...")
        self.postbuild(scen)

    except CreateSliverError:
      self._pbar.setStatus("Getting existing scenario...")
      m = self.uhm._site.listresources(self.uhm.context, scen.slicename)
      if scen.topo_name == m.toponame:
        _ip.push({"manifest" : m})
        self.uhm.manifest = m
        sys.exc_clear()
      else:
        self._pbar.fail("ERROR: Existing non-matching topology found.")
        return

    try:
      ss = self.uhm._site.sliverstatus(self.uhm.context, scen.slicename)
      self.uhm.sliverstatus = ss
      requests.post("http://localhost:9008/module/%s/status/%s" % (scen.modid,
                   _ip.user_ns["context"].userurn), json.dumps({"status" : json.dumps(ss)}), **rkwargs)
    except Exception:
      pass

    self._pbar.setStatus("Writing SSH Config", True)
    util.writeSliceConfig(self.uhm._slice, m)
    self._pbar.success("Success!")

    display(Javascript('IPython.notebook.select_next()'))


ScenarioManager = _ScenarioManager()
