# Copyright (c) 2018-2021  University of Houston Networking Lab

import random

import IPy
from .oldgraph.util import buildFromManifest

from . import pcaputil

class ListTable(object):
  def __init__ (self, cols, row_template, data):
    self._data = data
    self._columns = cols
    self._row_template = row_template
    self._col_template = "<th><center>%s</center></th>"
    self._title = None
    self._opt_monospace = False
    self._opt_fontsize = 90 

    self.border = False

  def _repr_html_ (self):
    trlist = []
    for row in self._data:
      if isinstance(row, (list, tuple)):
        trlist.append(self._row_template.format(*row))
      elif isinstance(row, dict):
        trlist.append(self._row_template.format(**row))
      elif isinstance(row, str):
        trlist.append(self._row_template.format(row))

    collist = []
    for column in self._columns:
      collist.append(self._col_template % (column))

    tbl = ["table"]
    if self._opt_monospace:
      tbl.append("style='font-family: monospace; font-size:%d%%'" % (self._opt_fontsize))
    if self.border:
      tbl.append("border='1'")
    retl = ["<%s>" % (" ".join(tbl))]

    if self._title:
      retl.append("<caption><center><b><i>%s</i><b></center></caption>" % (self._title))
    if collist:
      retl.append("<tr>%s</tr>" % ("".join(collist)))
    retl.append("\n".join(trlist))
    retl.append("</table>")

    return "\n".join(retl)

MTABLETYPES = {}
PCAPTABLETYPES = {}
DNSTABLETYPES = {}

def manifesttable (typ):
  def decorator(k):
    MTABLETYPES[typ] = k
    return k
  return decorator

def pcaptable (typ):
  def decorator(k):
    PCAPTABLETYPES[typ] = k
    return k
  return decorator

def dnstable (typ):
  def decorator(k):
    DNSTABLETYPES[typ] = k
    return k
  return decorator


@pcaptable("l2-fields")
class L2Fields(ListTable):
  COLS = ["Source", "Dest", "TPID", "VLAN", "Type"]
  ROW = "<tr><td>{src}</td><td>{dst}</td><td>{tpid}</td><td>{tag}</td><td>{type}</td></tr>"
  def __init__ (self, roster, xrdata, item, **kwargs):
    super(L2Fields, self).__init__(L2Fields.COLS, L2Fields.ROW, [])

    self._item = item
    self._title = "%s, port %s" % (roster[item]["bridge"], roster[item]["port"])
    
    stream = xrdata[roster[item]["pcap-name"]]
    self._data = pcaputil.buildEthHeaderList(stream)

@pcaptable("arp-fields")
class ARPFields(ListTable):
  COLS = ["eth<br/>Source MAC", "eth<br/>Dest  MAC", "eth<br/>Type", "eth<br/>VLAN", "arp<br/>op", "arp<br/>sha", "arp<br/>spa", "arp<br/>tha", "arp<br/>tpa"]
  ROW = "<tr><td>{src}</td><td>{dst}</td><td>{type}</td><td>{tag}</td><td>{op}</td><td>{sha}</td><td>{spa}</td><td>{tha}</td><td>{tpa}</td></tr>"
  def __init__ (self, roster, xrdata, item, **kwargs):
    super(ARPFields, self).__init__(ARPFields.COLS, ARPFields.ROW, [])
    self._opt_monospace = True
    self._opt_fontsize = 85

    self._item = item
    self._title = "%s, interface %s" % (roster[item]["host"], roster[item]["interface"])
    
    stream = xrdata[roster[item]["pcap-name"]]
    self._data = pcaputil.buildARPHeaderList(stream)


@dnstable("ns-tree-table")
class NSTreeTable(ListTable):
  COLS = ["Name", "TTL", "Type", "Data"]
  ROW = "<tr><td>{name}</td><td>{ttl}</td><td>{type}</td><td>{value}</td>"
  def __init__ (self, roster, xrdata, item, **kwargs):
    super(NSTreeTable, self).__init__(NSTreeTable.COLS, NSTreeTable.ROW, [])
    self._opt_monospace = True
    self._opt_fontsize = 85

    self._item = item
    self._title = item
    nsdata = xrdata[roster[item]["key"]]
    self._data = dnsutil.buildTableFromNSDP(nsdata)


@manifesttable("interface-macs")
class TopologyInterfaceMACs(ListTable):
  COLS = ["Interface", "MAC Address"]
  ROW = "<tr><td>{0}</td><td>{1}</td></tr>"
  def __init__ (self, manifest, **kwargs):
    super(TopologyInterfaceMACs, self).__init__(TopologyInterfaceMACs.COLS,
                                                TopologyInterfaceMACs.ROW, [])
    self._manifest = manifest
    self._title = "Topology Interfaces (MAC)"

    for container in manifest.containers:
      for port in container.ports:
        self._data.append(("%s.%s" % (container.client_id, port.name), port.macaddress))


@manifesttable("interface-ipv4")
class TopologyInterfaceIPs(ListTable):
  COLS = ["Interface", "IPv4 Addresses"]
  ROW = "<tr><td>{0}</td><td>{1}</td></tr>"
  def __init__ (self, _, sliverstatus, **kwargs):
    super(TopologyInterfaceIPs, self).__init__(TopologyInterfaceIPs.COLS,
                                               TopologyInterfaceIPs.ROW, [])
    self._sliverstatus = sliverstatus
    self._title = "Topology Interfaces (IPv4)"

    for container in sliverstatus["vts_containers"]:
      for port in container["ports"]:
        for addr in port["ipv4-addresses"]:
          self._data.append(("%s.%s" % (container["client-id"], port["name"]), addr))


@manifesttable("bridge-port-config")
class BridgePortConfig(ListTable):
  COLS = ["Port", "VLAN ID"]
  ROW = "<tr><td>{0}</td><td>{1}</td></tr>"
  def __init__ (self, manifest, bridge, **kwargs):
    super(BridgePortConfig, self).__init__(BridgePortConfig.COLS, BridgePortConfig.ROW, [])
    self._title = "%s Port Config" % (bridge)

    for port in manifest[bridge].ports:
      self._data.append((port.name, self._vlanstr(port.vlan_id)))

    self._data.sort(key=lambda x: x[0])

  def _vlanstr (self, vlan):
    if vlan == 0:
      return "[any]"
    return str(vlan)


@manifesttable("bridge-port-config-adv")
class BridgePortConfigAdv(ListTable):
  COLS = ["Port", "Type", "Native"]
  ROW = "<tr><td>{0}</td><td>{1}</td><td>{2}</td></tr>"
  def __init__ (self, manifest, bridge, **kwargs):
    super(BridgePortConfigAdv, self).__init__(BridgePortConfigAdv.COLS, BridgePortConfigAdv.ROW, [])
    self._title = "%s Port Config" % (bridge)

    for port in manifest[bridge].ports:
      self._data.append((port.name, self._getType(port), self._vlanstr(port.vlan_id)))

    self._data.sort(key=lambda x: x[0])

  def _getType (self, port):
    if port.vlan_id:
      return "Access"
    else:
      return "Trunk"

  def _vlanstr (self, vlan):
    if vlan == 0:
      return "None"
    return str(vlan)


@manifesttable("dhcp-config")
class DHCPConfig(ListTable):
  COLS = ["Server", "Subnet"]
  ROW = "<tr><td>{0}</td><td>{1}</td></tr>"
  IMAGES = ["uh.simple-dhcpd", "uh.dnsdhcp"]
  def __init__ (self, manifest, sliverstatus, **kwargs):
    super(DHCPConfig, self).__init__(DHCPConfig.COLS, DHCPConfig.ROW, [])
    self._title = "DHCP Server Config"

    dhcp_servers = []
    for container in manifest.containers:
      if container.image in DHCPConfig.IMAGES:
        dhcp_servers.append(container.client_id)

    for container in sliverstatus["vts_containers"]:
      if container["client-id"] in dhcp_servers:
        for port in container["ports"]:
          for addr in port["ipv4-addresses"]:
            self._data.append((container["client-id"], str(IPy.IP(addr, make_net=True))))


@manifesttable("host-route-table")
class HostRouteTable(ListTable):
  COLS = ["Destination", "Interface", "Gateway IP"]
  ROW = "<tr><td>{0}</td><td>{1}</td><td>{2}</td></tr>"
  def __init__ (self, manifest, hostname = None, hostroutetables = None, **kwargs):
    super(HostRouteTable, self).__init__(HostRouteTable.COLS, HostRouteTable.ROW, [])

    self._manifest = manifest
    self._title = "%s Route Table" % (hostname)
    routetable = hostroutetables[hostname]
    for route in routetable:
      dest_subnet = IPy.IP("%s/%s" % (route["destination"], route["mask"])).strNormal()
      interface = route["device"]
      gateway = route["gateway"]
      self._data.append((dest_subnet, interface, gateway))

@manifesttable("all-host-route-tables")
def buildHostRouteTables (manifest, hostroutetables = None, **kwargs):
  tbls = []
  for container in manifest.containers:
    if container.image != "uh.quagga":
      tbls.append(HostRouteTable(manifest, container.client_id, hostroutetables))
  return tbls

@manifesttable("router-route-table")
class RouterRouteTable(ListTable):
  COLS = ["Type", "Destination", "Interface", "Gateway IP"]
  ROW = "<tr><td>{0}</td><td>{1}</td><td>{2}</td><td>{3}</td></tr>"
  def __init__ (self, manifest, rtrname = None, rtrroutetables = None, **kwargs):
    super(RouterRouteTable, self).__init__(RouterRouteTable.COLS, RouterRouteTable.ROW, [])

    self._manifest = manifest
    self._title = "%s Route Table" % (rtrname)
    routetable = rtrroutetables[rtrname]
    for route in routetable:
      if route["selected"]:
        conn_type = route["type"]
        dest_subnet = route["network"]
        interface = route["interface"]
        gateway = route["next-hop"]
        self._data.append((conn_type, dest_subnet, interface, gateway))

@manifesttable("all-router-route-tables")
def buildRouterRouteTables (manifest, rtrroutetables = None, **kwargs):
  tbls = []
  for container in manifest.containers:
    if container.image == "uh.quagga":
      tbls.append(RouterRouteTable(manifest, container.client_id, rtrroutetables))
  return tbls


@manifesttable("host-arp-table")
class HostARPTable(ListTable):
  COLS = ["IP Address", "MAC Address", "Interface"]
  ROW = "<tr><td>{0}</td><td>{1}</td><td>{2}</td></tr>"
  def __init__ (self, manifest, sliverstatus, hostname = None, stpinfo = None, arpdata = None, **kwargs):
    super(HostARPTable, self).__init__(HostARPTable.COLS, HostARPTable.ROW, [])

    self._manifest = manifest
    self._sliverstatus = sliverstatus
    self._hostname = hostname
    self._title = "%s ARP Table" % (self._hostname)

    if arpdata and (hostname in arpdata):
      for entry in arpdata[hostname]:
        if not entry["device"]:
          continue
        self._data.append((entry["ip-address"], entry["hw-address"], entry["device"]))
    else:
      g = buildFromManifest(manifest, site = "ignore")
      g.decorateSTPInfo(stpinfo = stpinfo)
      g.applySliverStatus(sliverstatus)
      self._data = g[hostname].computeARPTable()


@manifesttable("all-arp-tables")
def buildARPTables (manifest, sliverstatus, stpinfo = None, **kwargs):
  tbls = []
  for container in manifest.containers:
    tbls.append(HostARPTable(manifest, sliverstatus, container.client_id, stpinfo))
  return tbls


@manifesttable("host-interfaces")
class HostInterfaces(ListTable):
  COLS = ["Interface", "MAC Address", "IPv4 Addresses"]
  ROW = "<tr><td>{0}</td><td>{1}</td><td>{2}</td></tr>"
  def __init__ (self, manifest, sliverstatus, hostname, **kwargs):
    super(HostInterfaces, self).__init__(HostInterfaces.COLS, HostInterfaces.ROW, [])

    self._manifest = manifest
    self._sliverstatus = sliverstatus
    self._title = "Host Interfaces (%s)" % (hostname)

    intfs = {}
    for container in manifest.containers:
      if container.client_id == hostname:
        for port in container.ports:
          intfs[port.name] = [port.macaddress, []]

    for container in sliverstatus["vts_containers"]:
      if container["client-id"] == hostname:
        for port in container["ports"]:
          for addr in port["ipv4-addresses"]:
            intfs[port["name"]][1].append(addr)
      
    kl = sorted(intfs.keys())
    for iname in kl:
      idata = intfs[iname]
      self._data.append((iname, idata[0], ", ".join(idata[1])))


@manifesttable("all-interfaces")
class AllInterfaces(ListTable):
  COLS = ["Interface", "MAC Address", "IPv4 Addresses"]
  ROW = "<tr><td>{0}</td><td>{1}</td><td>{2}</td></tr>"
  def __init__ (self, manifest, sliverstatus, **kwargs):
    super(AllInterfaces, self).__init__(AllInterfaces.COLS, AllInterfaces.ROW, [])

    self._manifest = manifest
    self._sliverstatus = sliverstatus
    self._title = "Topology Interfaces (MAC/IPv4)"

    ipv4_map = {}
    for container in sliverstatus["vts_containers"]:
      for port in container["ports"]:
        for addr in port["ipv4-addresses"]:
          ipv4_map.setdefault("%s.%s" % (container["client-id"], port["name"]), []).append(addr)

    intfs = {}
    for container in manifest.containers:
      for port in container.ports:
        iname = "%s.%s" % (container.client_id, port.name)
        if iname in ipv4_map:
          data = ipv4_map[iname]
        else:
          data = []
        intfs[iname] = (port.macaddress, ", ".join(data))

    kl = sorted(intfs.keys())
    for iname in kl:
      idata = intfs[iname]
      self._data.append((iname, idata[0], idata[1]))


@manifesttable("l2-table")
class Layer2Table(ListTable):
  COLS = ["Port", "MAC Address", "Age"]
  ROW = "<tr><td>{0}</td><td>{1}</td><td>{2}</td></tr>"
  VLANCOLS = ["Port", "MAC Address", "VLAN ID", "Age"]
  VLANROW = "<tr><td>{0}</td><td>{1}</td><td>{2}</td><td>{3}</td></tr>"
  def __init__ (self, manifest, bridge, stpinfo = None, hide = None, vlan = False, **kwargs):
    if vlan:
      super(Layer2Table, self).__init__(Layer2Table.VLANCOLS, Layer2Table.VLANROW, [])
    else:
      super(Layer2Table, self).__init__(Layer2Table.COLS, Layer2Table.ROW, [])

    if not isinstance(hide, list):
      hide = [hide]

    self._manifest = manifest
    self._bridge = bridge
    self._title = "%s L2 Table" % (self._bridge)

    if stpinfo:
      g = buildFromManifest(manifest, site = "ignore")
      g.decorateSTPInfo(stpinfo = stpinfo)
      macs = g[bridge].computeMACTable(include_vlans = vlan, hide = hide)
      if vlan:
        self._data = [(x[0], x[1], x[2], random.randint(0, 300)) for x in macs]
      else:
        self._data = [(x[0], x[1], random.randint(0, 300)) for x in macs]

    else:
      for port in manifest[bridge].ports:
        for rport in manifest[port.remote_dpname].ports:
          if port.remote_dpname in hide:
            continue
          if rport.client_id == port.remote_client_id:
            try:
              if vlan:
                self._data.append((port.name, rport.macaddress, port.vlanid, random.randint(0, 300)))
              else:
                self._data.append((port.name, rport.macaddress, random.randint(0, 300)))
            except AttributeError:
              continue # No MAC address

    self._data.sort(key=lambda x: x[0])


@manifesttable("all-l2-tables")
def buildL2Tables (manifest, stpinfo = None, hide = None, vlan = False, **kwargs):
  tbls = []
  for dp in manifest.datapaths:
    tbls.append(Layer2Table(manifest, dp.client_id, stpinfo, hide, vlan))
  return tbls


@manifesttable("all-bridge-port-configs")
def buildBPCs (manifest, **kwargs):
  tbls = []
  for dp in manifest.datapaths:
    tbls.append(BridgePortConfig(manifest, dp.client_id))
  return tbls
