# Copyright (c) 2019  University of Houston Networking Lab

from __future__ import absolute_import

import functools

import ipywidgets as widgets # pylint: disable=F0401

from .. import inputtable as IT
from ..widgetutils import getHostInterfaces

class ARPTab(object):
  def __init__ (self, mgr, idx):
    self.mgr = mgr
    self._tabidx = idx

    self._hostname = None
    self._hostintfs = []

    self._w_hostdd = None
    self._w_vbox = None
    self._input_table = None

    self._build()

  def _build (self):
    self._w_hostdd = widgets.Dropdown(options = self.mgr.hostlist, description = "Host:")
    self._w_hostdd.observe(self.hostChoice)

    self._w_deletebtn = widgets.Button(description = "Remove Tab", icon = "minus")
    self._w_deletebtn.on_click(functools.partial(self.mgr.deleteTab, idx = self.tabidx))

    self._w_vbox = widgets.VBox([self._w_hostdd, self._w_deletebtn])

  @property
  def hostname (self):
    return self._hostname

  @property
  def tabidx (self):
    return self._tabidx

  @tabidx.setter
  def tabidx (self, val):
    self._tabidx = val
    if self._input_table:
      self._input_table.dpindex = self.tabidx

  def hostChoice (self, change):
    if change['type'] == 'change' and change['name'] == 'value':
      self._hostname = change['new']
      self.mgr.setTabTitle(self.tabidx, self._hostname)
      self._hostintfs = self.mgr.getHostInterfaces(self._hostname)

      cols = []
      cols.append(IT.TextInputColumn("IP Address"))
      cols.append(IT.TextInputColumn("MAC Address"))
      cols.append(IT.DropdownColumn("Interface", values = self._hostintfs, allow_none = IT.DropdownColumn.MULTIPLE_ONLY))
      self._input_table = IT.InputTable(cols, {"index" : self.tabidx}, IT.TableOptions(allow_reorder = False))

      self._w_vbox.children = [self._w_hostdd, self._input_table.display(), self._w_deletebtn]

  def display (self):
    return self._w_vbox

  def getAnswerData (self):
    ans = {"hostname" : self._hostname , "table" : None}
    if self._input_table:
      ans["table"] = self._input_table.getAnswerData()
    return ans


class ARPTabManager(object):
  def __init__ (self, data):
    self._data = data
    self._hostlist = []
    self._hostintfs = {}

    self.dpindex = data["index"]
    self._wtab = widgets.Tab()

    self._wbutton = None
    if not ("suppress-button" in self._data) or (not self._data["suppress-button"]):
      if "button-text" in self._data:
        self._wbutton = widgets.Button(description = self._data["button-text"], icon = "plus")
      else:
        self._wbutton = widgets.Button(description = "Add ARP Table", icon = "plus")

    if self._wbutton:
      self._wbutton.on_click(self.addTab)
      self._vbox = widgets.VBox([self._wtab, self._wbutton])
    else:
      self._vbox = widgets.VBox([self._wtab])

    self._tabs = []
    self._tabdata = []
    self._uidata = None

    self.addTab()

  @property
  def hostlist (self):
    if not self._hostlist:
      for container in self._data["manifest"].containers:
        self._hostlist.append(container.client_id)
        for port in container.ports:
          self._hostintfs.setdefault(container.client_id, []).append(port.name)
      self._hostlist.sort()
      self._hostlist.insert(0, " -  ")
    return self._hostlist

  def getHostInterfaces (self, hostname):
    return self._hostintfs[hostname]

  def setTabTitle (self, idx, title):
    self._wtab.set_title(idx, title)

  def deleteTab (self, button, idx):
    self._tabdata.pop(idx)
    for tidx,tmgr in enumerate(self._tabdata):
      tmgr.tabidx = tidx
    self._tabs = [x.display() for x in self._tabdata]
    self._wtab.children = self._tabs
    for tab in self._tabdata:
      self.setTabTitle(tab.tabidx, tab.hostname)

  def addTab (self, *args, **kwargs):
    arptab = ARPTab(self, len(self._tabs))
    self._tabdata.append(arptab)
    self._tabs.append(arptab.display())
    self._wtab.children = self._tabs
    self._wtab.selected_index = len(self._tabs) - 1

  def display (self):
    return self._vbox

  def getAnswerData (self):
    data = [{"_uidata" : self._uidata}]
    for arptab in self._tabdata:
      data.append(arptab.getAnswerData())
    return data
