# Copyright (c) 2020  University of Houston

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import ipywidgets as widgets

from uhnet.types import IPv4Address

from ..dplib import dns
from ..log import Logger
from ..widgetutils import Layouts

GREY = "style='border: 1px solid black; background-color: #f5f5f5'"
WHITE = "style='border: 1px solid black; background-color: #ffffff'"

class QueryOutput(object):
  def __init__ (self):
    self._rows = []
    self._qidx = 0
      
  def addQueryOutput (self, server, rtype, key, response, error = False):
    self._qidx += 1
    if error:
      self._rows.insert(0, "<tr %s><td colspan=9>%s</td></tr>" % (WHITE, str(error)))
    else:
      if "additional" in response:
        ra = response["additional"]
        self._rows.insert(0, "<tr %s><td>&#8853</td><td>Key:</td><td>%s</td><td>TTL:</td><td>%s</td><td>Type:</td><td>%s</td><td>Value:</td><td>%s</td></tr>" % (
          WHITE, ra["key"], ra["ttl"], ra["type"], ra["value"]))
      self._rows.insert(0, "<tr %s><td>&nbsp;</td><td>Key:</td><td>%s</td><td>TTL:</td><td>%s</td><td>Type:</td><td>%s</td><td>Value:</td><td>%s</td></tr>" % (
        WHITE, response["key"], response["ttl"], response["type"], response["value"]))
    self._rows.insert(0, "<tr %s><td>%d</td><td colspan=3>Nameserver:</td><td>%s</td><td>Type:</td><td>%s</td><td>Key:</td><td>%s</td></tr>" % (GREY, self._qidx, server, rtype, key))
    if len(self._rows) > 20:
      self._rows = self._rows[:20]
          
  def _repr_html_ (self):
    tds = "border: 1px solid black"
    retl = []
    retl.append("<table style='font-family: monspace; font-size:90; %s'>" % (tds))
    retl.extend(self._rows)
    retl.append("</table>")
    return "\n".join(retl)
  
class InfoTable(object):
  def __init__ (self, dp, rtype, num_keys):
    self.__dp = dp
    self.__rtype = rtype
    self.__num_keys = num_keys
      
  def _repr_html_ (self):
    hlist = []
    hlist.append("<table style='border: 1px black solid; bgcolor: #ffffff'>")
    hlist.append("<tr><th>Root Nameserver</th><td>%s</td></tr>" % (self.__dp.getRoot().v4address))
    keys = self.__dp.pickRandomKeys(self.__rtype, self.__num_keys)
    hlist.append("<tr><th rowspan=%d>Records to Find</th><td></td></tr>" % (len(keys)+1))
    for key in keys:
      kval = ".".join(key.split(".")[::-1])[:-1]
      Logger.log("dns key: %s" % (kval))
      hlist.append("<tr><td>%s (%s)</td></tr>" % (kval, self.__rtype))
    hlist.append("</table>")
    return "\n".join(hlist)
                                                                               
class InteractiveExperience(object):
  def __init__ (self, zdp, num_keys = 3, rtype = "A"):
    self.__dp = zdp
    self.__num_keys = num_keys
    self.__rtype = rtype
    
    self._w_dd_srv = None
    self._w_dd_type = None
    self._w_input_key = None
    self._hbox_query = None
    self._response_output = QueryOutput()
    self._response_html = None
    self._full_vbox = None
    self._iw = None

    self._build()
      
  def _build(self):
    # Information portion
    infotable = InfoTable(self.__dp, self.__rtype, self.__num_keys)
    self._iw = widgets.Output()
    with self._iw:
      self._iw.append_display_data(infotable)
                                                                               
    # Query portion
    s_ips = [IPv4Address(ns.v4address) for ns in self.__dp.nameservers]
    s_ips.sort(key = lambda x: x._ip)
    types = ["A", "NS", "PTR", "MX", "TXT", "AAAA"]
    types.sort()

    w_lbl_srv = widgets.Label("Nameserver", layout = Layouts.AUTOWIDTH)
    self._w_dd_srv = widgets.Dropdown(options = [str(x) for x in s_ips], layout = Layouts.AUTOWIDTH)
    w_vb_srv = widgets.VBox([w_lbl_srv, self._w_dd_srv])

    w_lbl_type = widgets.Label("Type", layout = Layouts.AUTOWIDTH)
    self._w_dd_type = widgets.Dropdown(options = types, layout = Layouts.AUTOWIDTH)
    w_vb_type = widgets.VBox([w_lbl_type, self._w_dd_type])

    w_lbl_key = widgets.Label("Key", layout = Layouts.AUTOWIDTH)
    self._w_input_key = widgets.Text(layout = Layouts.AUTOWIDTH)
    w_vb_key = widgets.VBox([w_lbl_key, self._w_input_key])
    
    w_lbl_query = widgets.Label("")
    w_btn_query = widgets.Button(description = "Query", layout = Layouts.AUTOWIDTH)
    w_btn_query.on_click(self._query)
    w_vb_query = widgets.VBox([w_lbl_query, w_btn_query])

    self._hbox_query = widgets.HBox([w_vb_srv, w_vb_type, w_vb_key, w_vb_query])
    
    gw = widgets.Output()
    with gw:
      gw.append_display_data(self._response_output)
        
    self._full_vbox = widgets.VBox([self._iw, self._hbox_query, gw])
      
  def _query (self, button):
    nsip = self._w_dd_srv.value
    rtype = self._w_dd_type.value
    ns = self.__dp.getNameserverByIPv4(nsip)
    key = self._w_input_key.value
    try:
      Logger.log("dns-query: %s, %s, %s" % (nsip, rtype, key))
      resp = ns.lookup(key, rtype)
      self._response_output.addQueryOutput(nsip, rtype, key, resp)
    except dns.DNSError as e:
      self._response_output.addQueryOutput(nsip, rtype, key, None, e)

    gw = widgets.Output()
    with gw:
      gw.append_display_data(self._response_output)
    self._full_vbox.children = [self._iw, self._hbox_query, gw]

  def display (self):            
    gw = widgets.Output()
    with gw:
      gw.append_display_data(self._response_output)
    self._full_vbox.children = [self._iw, self._hbox_query, gw]
    return self._full_vbox
