# Copyright (c) 2020  University of Houston

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import json
import random

import ipywidgets as widgets

from ..log import Logger
from ..widgetutils import Layouts

GREY = "style='border: 1px solid black; background-color: #f5f5f5'"
WHITE = "style='border: 1px solid black; background-color: #ffffff'"

BITS = [4, 8, 12, 16]

class NumberSystem(object):
  BASES = { "b" : 2, "o" : 8, "d" : 10, "n" : 10, "x" : 16, "X" : 16}
  TT = "".maketrans({x:"" for x in [" ", ".", "_", "-"]})

  def __init__ (self, name, fmt):
    self.name = name
    self.format = fmt
    self.base = NumberSystem.BASES[fmt]

  def display (self, val):
    return format(val, self.format)

  def parse (self, val):
    val = val.translate(NumberSystem.TT)
    return int(val, self.base)


SYSTEMS = [
  NumberSystem("Binary", "b"),
  NumberSystem("Decimal", "d"),
  NumberSystem("Hexadecimal", "X")
  ]

class HistoryOutput(object):
  def __init__ (self):
    self._rows = []
    self._qidx = 0

  def addOutput (self, query, answer):
    self._qidx += 1

    answer["class"] = WHITE
    if answer["aa"]:
      answer["aa"] = '<i class="fas fa-check"></i>'
    else:
      answer["aa"] = '<i class="fas fa-exclamation-circle"></i>'

    if answer["ba"]:
      answer["ba"] = '<i class="fas fa-check"></i>'
    else:
      answer["ba"] = '<i class="fas fa-exclamation-circle"></i>'


    self._rows.insert(0, '<tr %(class)s><td style="width:10%%">&nbsp;</td><td>%(an)s</td><td>%(av)s</td><td style="width:10%%">%(aa)s</td></tr>' % (answer))
    self._rows.insert(0, '<tr %(class)s><td style="width:10%%">&nbsp;</td><td>%(bn)s</td><td>%(bv)s</td><td style="width:10%%">%(ba)s</td></tr>' % (answer))

    query["class"] = GREY
    query["idx"] = self._qidx
    self._rows.insert(0, '<tr %(class)s><td style="width:10%%">%(idx)d</td><td>%(name)s</td><td>%(val)s</td><td style="width: 10%%">&nbsp</td></tr>' % (query))

    if len(self._rows) > 15:
      self._rows = self._rows[:15]
          
  def _repr_html_ (self):
    tds = "border: 1px solid black"
    retl = []
    retl.append("<table style='font-family: monspace; font-size:90; %s'>" % (tds))
    retl.extend(self._rows)
    retl.append("</table>")
    return "\n".join(retl)


class InteractiveExperience(object):
  def __init__ (self):
    self._w_dd_bits = None
    self._history_output = HistoryOutput()
    self._full_vbox = None
    self._w_hb_bits = None
    self._w_hb_input = None

    self._w_lbl_query_system = None
    self._w_lbl_query_value = None

    self._w_lbl_input_sysA = None
    self._w_txt_input_sysA = None
    self._w_lbl_input_sysB = None
    self._w_txt_input_sysB = None

    self._build()

  def _build (self):
    w_lbl_bits = widgets.Label("Max Bits", layout = Layouts.AUTOWIDTH)
    self._w_dd_bits = widgets.Dropdown(options = [str(x) for x in BITS], layout = Layouts.AUTOWIDTH)
    w_btn_go = widgets.Button(description = "Go", layout = Layouts.AUTOWIDTH)
    w_btn_go.on_click(self._go_button)
    self._w_hb_bits = widgets.HBox([w_lbl_bits, self._w_dd_bits, w_btn_go])

    self._w_lbl_query_system = widgets.Label("", layout = Layouts.AUTOWIDTH)
    self._w_lbl_query_value = widgets.Label("", layout = Layouts.AUTOWIDTH)

    self._w_lbl_input_sysA = widgets.Label("", layout = Layouts.AUTOWIDTH)
    self._w_txt_input_sysA = widgets.Text(layout = Layouts.AUTOWIDTH)
    self._w_lbl_input_sysB = widgets.Label("", layout = Layouts.AUTOWIDTH)
    self._w_txt_input_sysB = widgets.Text(layout = Layouts.AUTOWIDTH)
    w_btn_check = widgets.Button(description = "Check", layout = Layouts.AUTOWIDTH)
    w_btn_check.on_click(self._check_button)

    w_vb_ilbl = widgets.VBox([self._w_lbl_query_system, self._w_lbl_input_sysA, self._w_lbl_input_sysB])
    w_vb_itxt = widgets.VBox([self._w_lbl_query_value, self._w_txt_input_sysA, self._w_txt_input_sysB])
    self._w_hb_input = widgets.HBox([w_vb_ilbl, w_vb_itxt, widgets.VBox([w_btn_check])])

    self._full_vbox = widgets.VBox([self._w_hb_bits, self._w_hb_input])

  def _go_button (self, button):
    self._w_txt_input_sysA.value = ""
    self._w_txt_input_sysB.value = ""

    bits = int(self._w_dd_bits.value)
    Logger.log("%%number: Go, %d bits" % (bits))
    val = random.randint(0, 2**bits)
    system = random.choice(SYSTEMS) 
    self._w_lbl_query_system.value = system.name
    self._w_lbl_query_value.value = system.display(val)
    Logger.log("%%number: Query: %s, %s" % (system.name, system.display(val)))

    for sysa in SYSTEMS:
      if sysa != system:
        self._w_lbl_input_sysA.value = sysa.name
        break

    for sysb in SYSTEMS:
      if sysb != system and sysb != sysa:
        self._w_lbl_input_sysB.value = sysb.name

  def _check_button (self, button):
    # Somewhere an alley is being deprived of its dumpster fire
    sysa = [x for x in SYSTEMS if x.name == self._w_lbl_input_sysA.value][0]
    sysb = [x for x in SYSTEMS if x.name == self._w_lbl_input_sysB.value][0]
    sysval = [x for x in SYSTEMS if x.name == self._w_lbl_query_system.value][0]

    realval = sysval.parse(self._w_lbl_query_value.value)
    sysacheck = (sysa.parse(self._w_txt_input_sysA.value) == realval)
    sysbcheck = (sysb.parse(self._w_txt_input_sysB.value) == realval)

    query = {"name" : sysval.name, "val" : self._w_lbl_query_value.value}
    answer = {"an" : sysa.name, "av" : self._w_txt_input_sysA.value, "aa" : sysacheck,
              "bn" : sysb.name, "bv" : self._w_txt_input_sysB.value, "ba" : sysbcheck}
    Logger.log("%%number: Check: %s" % (json.dumps(answer)))
    self._history_output.addOutput(query, answer)

    gw = widgets.Output()
    with gw:
      gw.append_display_data(self._history_output)
    self._full_vbox.children = [self._w_hb_bits, self._w_hb_input, gw]


  def display (self):
    gw = widgets.Output()
    with gw:
      gw.append_display_data(self._history_output)
    self._full_vbox.children = [self._w_hb_bits, self._w_hb_input, gw]
    return self._full_vbox

