# Copyright (c) 2018-2020  University of Houston

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import ipywidgets as widgets # pylint: disable=F0401
import requests

import grader.client

from ..exercise import ExerciseManager
from ..log import Logger
from ..widgetutils import Layouts

class PracticeController(object):
  def __init__ (self, exlist):
    self._exlist = exlist
    self._exmap = {} # { ui_name : { q_title : (exobj, qidx), ...}, ...}
    self._exm = None
    self._cur_qidx = None
    self._cur_xrcs = None

    self._buildDDData()

    self._w_ddhbox = None
    self._w_framebox = None
    dummy = ["  -"]
    dummy.extend(sorted(self._exmap.keys()))
    self._w_exdd = widgets.Dropdown(options = dummy, layout = Layouts.AUTOWIDTH)
    self._w_qdd = widgets.Dropdown(options = ["  -"], layout = Layouts.AUTOWIDTH)
    self._w_sbutton = widgets.Button(description = "Show")
    self._w_sbutton.on_click(self._buttonShow)
    self._w_sbar = widgets.Text("", layout = widgets.Layout(width="99%"))

    self._w_exdd.observe(self._exChoice)

  def _buildDDData (self):
    for exobj in self._exlist:
      if not exobj["ui-name"] or exobj["disable-practice"]:
        continue  # Skip exercises without names

      tmap = {}
      for idx,qinfo in enumerate(exobj["question-data"], 1):
        tmap[qinfo["title"]] = (exobj, idx)
      self._exmap[exobj["ui-name"]] = tmap

  def _exChoice (self, change):
    if change['type'] == 'change' and change['name'] == 'value':
      if change['new'] == "  -":
        self._w_qdd.options = ["  -"]
      else:
        dummy = ["  -"]
        dummy.extend(sorted(self._exmap[change['new']].keys()))
        self._w_qdd.options = dummy
      self._w_qdd.value = "  -"

  def _buttonShow (self, button):
    if self._w_exdd.value == "  -" or self._w_qdd.value == "  -":
      return
    Logger.log("%%practice: Loading %s: %s" % (self._w_exdd.value, self._w_qdd.value))
    (self._cur_xrcs, self._cur_qidx) = self._exmap[self._w_exdd.value][self._w_qdd.value]
    self._exm = ExerciseManager(self._cur_xrcs, self._gradeCallback, self._cur_qidx)
    self._w_sbar.value = ""
    self._w_framebox.children = [self._w_ddhbox, self._exm.display(), self._w_sbar]

  def _getAnswerKey (self):
    res = requests.get("http://localhost:9008/answer-key/name/%s" % (self._cur_xrcs["akey-name"]))
    return res.json()["value"]

  def _getDatapool (self, idx):
    res = requests.get("http://localhost:9008/datapool/%s/get/%d" % (
                        self._cur_xrcs["question-data"][self._cur_qidx-1]["data-pool"], idx))
    return res.json()["value"]

  def _gradeCallback (self, exm, button):
    exmgr = exm._qpanel_data[1][0] # Get the manager for the only tab
    if isinstance(exmgr, dict):
      adata = {}
      for k,v in exmgr.items():
        if k == "_uidata":
          adata[k] = v
          dpindex = v["index"]
        else:
          adata[k] = v.value
      answers = { str(self._cur_qidx) : [adata]}
    else:
      answers = { str(self._cur_qidx) : {"index" : exmgr.dpindex, "values" : exmgr.getAnswerData()} }
      dpindex = exmgr.dpindex

    akey = self._getAnswerKey()
    dp = self._getDatapool(dpindex)
    dp["use-vlans"] = self._cur_xrcs["question-data"][self._cur_qidx-1]["use-vlans"]
    data = {str(self._cur_qidx) : dp}

    ginfo = grader.client.evalQuestions(data, akey, answers)
    if ginfo[str(self._cur_qidx)]["score"] == 1:
      self._w_sbar.value = "Correct!"
    else:
      self._w_sbar.value = "Your answer is incorrect.  Try again!"

    internal_score = "%0.2f%%" % (ginfo[str(self._cur_qidx)]["score"]*100)
    Logger.log("%%practice: grade %s" % (internal_score))

  def display (self):
    if not self._w_framebox:
      self._w_ddhbox = widgets.HBox([self._w_exdd, self._w_qdd, self._w_sbutton])
      self._w_framebox = widgets.VBox([self._w_ddhbox])
    return self._w_framebox
