# Copyright (c) 2017-2021  University of Houston

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import getpass
import json
import os
import os.path
import random
import sys

import geni._coreutil as GCU
import geni.aggregate.vts as VTSAM
from IPy import IP
import requests

from . import creds
from .log import Logger

HOME = os.path.expanduser("~")
DEFAULT_PATH = HOME + "/.ssh/%s.config"

class LabserveError(Exception):
  def __init__ (self, msg):
    super(LabserveError, self).__init__()
    self._msg = msg
  def __str__ (self):
    return self._msg

def writeSliceConfig (slicename, manifests, pkey_path = None, path = None, user = None):
  from geni.rspec.vtsmanifest import Manifest as VTSM
  from geni.rspec.pgmanifest import Manifest as PGM

  if not path:
    path = DEFAULT_PATH % (slicename)

  if not isinstance(manifests, (list, set)):
    manifests = [manifests]

  if not os.path.exists("%s/.ssh" % (HOME)):
    os.makedirs("%s/.ssh" % (HOME))

  # Build up a mapping of pubkeys to private key paths
  kmaps = {}
  sshdirs = [os.path.expanduser("~/.ssh/"), GCU.getDefaultDir()]
  for dirpath in sshdirs:
    if not os.path.exists(dirpath):
      continue
    for name in os.listdir(dirpath):
      if name.endswith(".pub"):
        pvkpath = "%s/%s" % (dirpath, name[:-4])
        if os.path.exists(pvkpath):
          pkdata = open("%s/%s" % (dirpath, name), "r").read().strip()
          kmaps[pkdata] = pvkpath

  config = []
  config.append("IdentitiesOnly yes")
  config.append("Protocol 2")
  config.append("")

  for manifest in manifests:
    if isinstance(manifest, VTSM):
      objs = manifest.containers
    elif isinstance(manifest, PGM):
      objs = manifest.nodes

    for node in objs:
      if not node.logins:
        continue

      linfo = node.logins[0]
      for user in node.users:
        if user.public_key in kmaps:
          config.append("Host %s" % (node.client_id))
          config.append("  Hostname %s" % (linfo.hostname))
          config.append("  User %s" % (user.login))
          config.append("  Port %d" % (linfo.port))
          config.append("  IdentityFile %s" % (kmaps[user.public_key]))
          config.append("")
          break


  f = open(path, "w")
  f.write("\n".join(config))
  f.close()

def userHasSliceCred (context, sname, project = None):
  if not project:
    project = context.project
  spath = "%s/%s-%s-%s-scred.xml" % (context.datadir, context.cf.name, project, sname)
  return os.path.exists(spath)

rkwargs = {"headers" : {"Content-Type" : "application/json"},
           "verify" : False}

def issueUserSliceCred (course, slice_data):
  for project, user_data in slice_data.items():
    for user in user_data:
      [slicename, user_cert] = user_data[user]
      Logger.log("Issuing slice credential for %s" % (slicename))
      data = json.dumps({'user-cert': user_cert.decode("utf-8"), 'project': project})
      slicecred_bundle = requests.post("http://localhost:9008/course/%s/creds/issue-slice/%s" % (course, slicename),
                                       data, **rkwargs).json()["value"]
      creds.writeSCreds(user, slicecred_bundle, project, slicename)


def getLabCourse ():
  user = getpass.getuser()
  labdata = requests.get("http://localhost:9008/user/%s/courses" % (user)).json()['value']
  courses = labdata["courses"]
  if courses:
    return (courses[0], labdata["shortuid"], labdata["labserve-url"])

  bindir = os.path.dirname(sys.executable)
  CPATH = "%s/labsrv-info.json" % (bindir)

  if not os.path.exists(CPATH):
    return "hs-cyber"

  data = json.loads(open(CPATH, "rb").read())
  return (data["course"], user, "http://localhost:9008")

def getUserSite (course, userurn):
  site = requests.get("http://localhost:9008/course/%s/site/%s" % (course, userurn)).json()["value"]
  if isinstance(site, str):
    return VTSAM.name_to_aggregate(site)
  elif isinstance(site, (tuple,list)):
    return VTSAM.VTS(site[0], site[1])


### Utils for rspec generation / randomization

def getRandomNetwork (prefix_len, exclude_nets = None):
  if exclude_nets is None:
    exclude_nets = []
  not_found = True
  while not_found:
    not_found = False
    nint = random.randint(1, 3758096383)
    for net in exclude_nets:
      if nint in net:
        not_found = True
        break
  ip1 = IP(nint)
  ip2 = IP("%s/%d" % (ip1.strNormal(), prefix_len), make_net = True)
  return ip2

def getSubnets (min_prefix, max_prefix, count, exclude_nets = None):
  subnets = []
  while len(subnets) != count:
    plen = random.randint(min_prefix, max_prefix)
    net = getRandomNetwork(plen, exclude_nets)
    try:
      for subnet in subnets:
        if net.overlaps(subnet):
          raise ValueError
    except ValueError:
      continue
    subnets.append(net)
  return subnets

class InsufficientHostSpaceError(Exception):
  """Exception raised for errors when there is not enough number
  of host IP addresses in the subnet address range given.
  """
  def __init__ (self,num_hosts, net_len):
    super(InsufficientHostSpaceError, self).__init__()
    self._num_hosts = num_hosts
    self._net_len = net_len

  def __str__ (self):
    return "There are not sufficient number of IPs (= %d) for this many host IPs (= %d)" % (self._net_len, self._num_hosts)


def getNetworkHosts (net, num_hosts, exclude_ips = None):
  hostIPs = set([])
  max_num_ips = net.len() - 2
  if num_hosts > max_num_ips:
    raise InsufficientHostSpaceError(num_hosts, max_num_ips)

  if exclude_ips is None:
    exclude_ips = set()

  for i in range(num_hosts):
    # reserve exclude set
    new_ip = net[random.randint(1, net.len()-2)]
    while (new_ip in hostIPs) or (new_ip in exclude_ips):
      new_ip = net[random.randint(1, net.len()-2)]
    hostIPs.add(new_ip)

  return list(hostIPs)

def randomNames (name_prefix, no_names = 2, rand_range = 20):
  NAMEList = []
  for idx in range(1, rand_range):
    NAMEList.append("%s%02d" % (name_prefix, idx))
  random.shuffle(NAMEList)
  return NAMEList[0:no_names]

class NetworkSequence(object):
  def __init__ (self, subnetstr, prefix_len):
    self.subnet = IP(subnetstr, make_net=True)
    self.prefix_len = prefix_len

  def __iter__ (self):
    return NetSeqIter(self)


class NetSeqIter(object):
  def __init__ (self, ns):
    self._ns = ns
    self.iter = 0
    self.netsize = 2**(32-self._ns.prefix_len)

  def __iter__ (self):
    return self

  def next (self):
    try:
      ret = IP("%s/%d" % (self._ns.subnet[self.iter * self.netsize], self._ns.prefix_len))
      self.iter += 1
      return ret
    except IndexError:
      raise StopIteration
