# Copyright (c) 2018-2020  University of Houston

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import functools
import random

import ipywidgets as widgets # pylint: disable=F0401
import IPy
import six

DNSRECTYPES = ["A",
               "AAAA",
               "CNAME",
               "MX",
               "NS",
               "PTR"]

class Layouts(object):
  TINYBUTTON = widgets.Layout(width="25px", padding="0px")
  ROWNUMCOL = widgets.Layout(width="50px")
  AUTOWIDTH = widgets.Layout(width="auto")
  DDWIDTH = widgets.Layout(width="200px")


class HeaderRow(object):
  def __init__ (self, _):
    self.col_row = widgets.HTML("&nbsp;", layout=Layouts.AUTOWIDTH)
    self.col_vertices = widgets.HTML("<b>Vertex</b>", layout=Layouts.AUTOWIDTH)
    self.col_actions = widgets.HTML("<b>Action</b>", layout=Layouts.AUTOWIDTH)
    self.col_data = widgets.HTML("<b>Data</b>", layout=Layouts.AUTOWIDTH)
    self.col_controls = widgets.HTML("&nbsp;", layout=Layouts.AUTOWIDTH)


class FooterRow(object):
  def __init__ (self, manager):
    self.manager = manager

    addbutton = widgets.Button(icon="plus-circle", layout=Layouts.ROWNUMCOL, tooltip = "Add Row")
    addbutton.on_click(self.manager.buttonAddRow)
    self.col_row = addbutton
    self.col_vertices = widgets.VBox([])
    self.col_actions = widgets.VBox([])
    self.col_data = widgets.VBox([])
    self.col_controls = widgets.VBox([])


class AnswerRow(object):
  def __init__ (self, manager, rownum):
    self.manager = manager
    self._rownum = rownum
    self._col_rows = []
    self._col_vertices = []
    self._col_actions = []
    self._col_data = []
    self._col_controls = []
    self._hidden_controls = []
    self._buildEmpty()

  @property
  def rownum (self):
    return self._rownum

  @rownum.setter
  def rownum (self, val):
    self._rownum = val
    self._col_rows[0].value = str(val+1)
    for button in self._col_controls[0].children:
      button._row = val
    for widget in self._col_actions:
      widget._row = val

  @property
  def col_row (self):
    return self._col_rows

  @property
  def col_vertices (self):
    return self._col_vertices

  @property
  def col_actions (self):
    return self._col_actions

  @property
  def col_data (self):
    return self._col_data

  @property
  def col_controls (self):
    return self._col_controls

  @property
  def hide_controls (self):
    return self._hidden_controls

  @hide_controls.setter
  def hide_controls (self, val):
    if val == False and self._hidden_controls:
      self._col_controls = self._hidden_controls[:]
      self._hidden_controls = []
    elif val == True:
      self._hidden_controls = self._col_controls[:]
      self._col_controls = [widgets.HBox()]

  def getAnswerData (self):
    data = {}
    data["vertex"] = self._col_vertices[0].value
    data["action"] = self._col_actions[0].value
    ddata = {}
    for x in self._col_data[0].children:
      if isinstance(x, widgets.VBox):
        ddata = {w.description : w.value for w in x.children}
      else:
        ddata[x.description] = x.value
    data["data"] = ddata
    return data


  def _buildEmpty (self):
    dnames = ["  -"]
    dnames.extend(sorted(self.manager.vertices.keys()))
    dactions = ["  -"]
#        dactions.extend(self.manager.actions)
    namedd = widgets.Dropdown(options=dnames, layout=Layouts.DDWIDTH)
    namedd.observe(functools.partial(self.vertexChoice, dropdown=namedd))
    actiondd = widgets.Dropdown(options=dactions, layout=Layouts.DDWIDTH)
    actiondd._row = self.rownum
    actiondd.observe(functools.partial(self.actionChoice, dropdown=actiondd))
    databox = widgets.HBox([widgets.Label("")])
    rowdelbutton = widgets.Button(icon="minus-circle", layout=Layouts.TINYBUTTON, tooltip = "Delete Row")
    rowdelbutton._row = self.rownum
    rowdelbutton.on_click(self.manager.buttonDeleteRow)
    rowupbutton = widgets.Button(icon="caret-up", layout=Layouts.TINYBUTTON, tooltip = "Move Row Up")
    rowupbutton._row = self.rownum
    rowupbutton.on_click(self.manager.buttonMoveRowUp)
    rowdownbutton = widgets.Button(icon="caret-down", layout=Layouts.TINYBUTTON, tooltip = "Move Row Down")
    rowdownbutton._row = self.rownum
    rowdownbutton.on_click(self.manager.buttonMoveRowDown)
    rowaddbutton = widgets.Button(icon="plus-circle", layout=Layouts.TINYBUTTON, tooltip = "Add Row Before")
    rowaddbutton._row = self.rownum
    rowaddbutton.on_click(self.manager.buttonAddRowBefore)
    self._col_rows.append(widgets.Label(str(self.rownum+1), layout = Layouts.AUTOWIDTH))
    self._col_vertices.append(namedd)
    self._col_actions.append(actiondd)
    self._col_data.append(databox)
    self._col_controls.append(widgets.HBox([rowdelbutton, rowupbutton, rowdownbutton, rowaddbutton]))

  def actionChoice (self, change, dropdown):
    row = dropdown._row
    if change['type'] == 'change' and change['name'] == 'value':
      self.makeDataboxElements(change['new'])

  def vertexChoice (self, change, dropdown):
    if change['type'] == 'change' and change['name'] == 'value':
      self._populateActionDropdown(change['new'])
      self.makeDataboxElements(None)

  def _populateActionDropdown (self, vertex):
    actions = ["  -"]
    try:
      (vtype, vimage) = self.manager.vertices[vertex]
      if vimage in self.manager.actions[vtype]:
        actions.extend(self.manager.actions[vtype][vimage])
      else:
        actions.extend(self.manager.actions[vtype]["*"])
    except KeyError:
      pass
    finally:
      self._col_actions[0].options = actions
      self._col_actions[0].value = "  -"


  def _setRowHeight (self, height):
    if len(self._col_rows) > height:
      self._col_rows = self._col_rows[:height]
      self._col_vertices = self._col_vertices[:height]
      self._col_actions = self._col_actions[:height]
      self._col_controls = self._col_controls[:height]
    elif len(self._col_rows) < height:
      for x in range(height - len(self._col_rows)):
        self._col_rows.append(widgets.Label(""))
        self._col_vertices.append(widgets.Label(""))
        self._col_actions.append(widgets.Label(""))
        self._col_controls.append(widgets.Label(""))

  def makeDataboxElements(self, atype):
    if atype == "forward":
      args = {"use-dest-mac" : False, "use-dest-ip" : False, "port-title" : "Port",
              "use-port" : True}
      vtype, vimage = self.manager.vertices[self._col_vertices[0].value]
      try:
        typeargs = self.manager.action_args[vtype]
        if vimage not in typeargs:
          args.update(typeargs["*"]["forward"])
        args.update(typeargs[vimage]["forward"])
      except KeyError:
        pass

      row_height = 0

      inputs = []

      if args["use-port"]:
        ports = []
        try:
          ports.extend(self.manager._dp_ports[self._col_vertices[0].value])
        except KeyError:
          pass
        ports.sort()
        ports.insert(0, "  -")
        inputs.append(widgets.Dropdown(options=ports, description="%s:" % (args["port-title"])))
        row_height += 1

      if self.manager.require_vlans:
        vlans = ["  -"]
        vlans.extend(getVLANs(self.manager.manifest))
        vlans.append("[none]")
        inputs.append(widgets.Dropdown(options=vlans, description = "VLAN:"))
        row_height += 1

      if args["use-dest-mac"]:
        macs = ["  -"]
        macs.extend(self.manager.macs)
        inputs.append(widgets.Dropdown(options=macs, description="Dest. MAC:"))
        row_height += 1

      if args["use-dest-ip"]:
        ips = ["  -"]
        ips.extend(self.manager._ip_addresses)
        inputs.append(widgets.Dropdown(options=ips, description="Dest. IP:"))
        row_height += 1

#       inputs.append(widgets.BoundedIntText(value=0, min=0, max=4092, description="VLAN:", layout=widgets.Layout(width="200px")))
      self._setRowHeight(row_height)
      new_children = [widgets.VBox(inputs)]
    elif atype == "arp-request":
      ips = ["  -"]
      ips.extend(self.manager._ip_addresses)
      new_children = [widgets.Dropdown(options=ips, description="Target IP:")]
      self._setRowHeight(1)
    elif atype == "arp-reply":
      macs = ["  -"]
      macs.extend(self.manager.macs)
      dd = widgets.Dropdown(options=macs, description="Source MAC:")
      new_children = [dd]
      self._setRowHeight(1)
    elif atype == "dns-request":
      types = ["  -"]
      types.extend(DNSRECTYPES)
      ips = ["  -"]
      ips.extend(self.manager._ip_addresses)
      dd = widgets.Dropdown(options=types, description="Type:")
      ipw = widgets.Dropdown(options=ips, description="Nameserver IP:")
      vw = widgets.Text(description="Value:")
      new_children = [dd, vw, ipw]
      self._setRowHeight(3)
    elif atype == "dns-response":
      types = ["  -"]
      types.extend(DNSRECTYPES)
      dd = widgets.Dropdown(options=types, description="Type:")
      vw = widgets.Text(description="Value:")
      new_children = [dd, vw]
    elif atype == "send-packet":
      macs = ["  -"]
      macs.extend(self.manager.macs)
      dd = widgets.Dropdown(options=macs, description="Dest. MAC:")
      new_children = [dd]
      self._setRowHeight(1)
    elif atype == "echo-request":
      ips = ["  -"]
      ips.extend(self.manager._ip_addresses)
      macs = ["  -"]
      macs.extend(self.manager.macs)
      inputs = []
      inputs.append(widgets.Dropdown(options=ips, description="Dest. IP:"))
      inputs.append(widgets.Dropdown(options=macs, description="Dest. MAC:"))
      new_children = [widgets.VBox(inputs)]
      self._setRowHeight(2)
    elif atype == "echo-reply":
      ips = ["  -"]
      ips.extend(self.manager._ip_addresses)
      macs = ["  -"]
      macs.extend(self.manager.macs)
      inputs = []
      inputs.append(widgets.Dropdown(options=ips, description="Dest. IP:"))
      inputs.append(widgets.Dropdown(options=macs, description="Dest. MAC:"))
      new_children = [widgets.VBox(inputs)]
      self._setRowHeight(2)
    elif atype == "flood":
      new_children = [widgets.Text(description="Ports:")]
      self._setRowHeight(1)
    else:
      new_children = [widgets.Label("")]
      self._setRowHeight(1)
    self._col_data[0].children = new_children
    self.manager.display()


class AnswerManager(object):
  def __init__ (self, data, actions):
    self.manifest = data["manifest"]
    self.sliverstatus = data["sliverstatus"]
    self.dpindex = data["index"]
    self._uidata = None
    self.require_vlans = False
    self.vertices = {}
    self._dp_ports = {}
    self.actions = {}
    self.action_args = {}
    self.macs = []
    self.rows = []
    self.header = None
    self.footer = None
    self._hbox = None
    self._vbox_rows = None
    self._vbox_vertices = None
    self._vbox_actions = None
    self._vbox_data = None
    self._vbox_controls = None
    self._hidden_controls = False

    self._buildActions(actions)
    self._build()

  def _buildActions (self, actions):
    if isinstance(actions, (list, tuple)):
      self.actions.setdefault("container", {})["*"] = actions
      self.actions.setdefault("datapath", {})["*"] = actions
    elif isinstance(actions, dict):
      for k,v in actions.items():
        vtype,vimage = k.split(":", 1)
        action_names = []
        for item in v:
          if isinstance(item, (six.text_type)):
            action_names.append(item)
          elif isinstance(item, dict):
            # TODO: Fixed for py3, but also, wtf does this actually do?
            keys = list(item.keys())
            values = list(item.values())
            action_names.append(keys[0])
            self.action_args.setdefault(vtype, {}).setdefault(vimage, {})[keys[0]] = values[0]
        self.actions.setdefault(vtype, {})[vimage] = action_names
    # TODO: else raise something

  def _build (self):
    for dp in self.manifest.datapaths:
      self.vertices[dp.client_id] = ("datapath", "none")
      self._dp_ports[dp.client_id] = [x.name for x in dp.ports]
    for node in self.manifest.containers:
      for port in node.ports:
        self.vertices["%s.%s" % (node.client_id, port.name)] = ("container", node.image)
      self._dp_ports[node.client_id] = [x.name for x in node.ports]
    self.macs = ["ff:ff:ff:ff:ff:ff"]
    for node in self.manifest.containers:
      for port in node.ports:
        self.macs.append(port.macaddress)
    self.macs.sort()
    self._ip_addresses = []
    for cdata in self.sliverstatus["vts_containers"]:
      for port in cdata["ports"]:
        for ip in port["ipv4-addresses"]:
          self._ip_addresses.append(ip.split("/")[0])
    self._ip_addresses.sort()

    self.header = HeaderRow(self)
    self.footer = FooterRow(self)
    self.addEmptyRow()

  def getAnswerData (self):
    vals = [{"_uidata" : self._uidata, "index" : self.dpindex}]
    for row in self.rows:
      vals.append(row.getAnswerData())
    return vals

  def setRowMetadata (self):
    for idx,row in enumerate(self.rows):
      row.rownum = idx

  def buttonAddRow (self, _):
    self.addEmptyRow()
    self.display()

  def buttonDeleteRow (self, button):
    if len(self.rows) <= 1:
      return
    del self.rows[button._row]

    self.setRowMetadata()
    self.display()

  def buttonMoveRowDown (self, button):
    if len(self.rows) <= button._row+1:
      return
    oldrow = self.rows[button._row+1]
    self.rows[button._row+1] = self.rows[button._row]
    self.rows[button._row] = oldrow

    self.setRowMetadata()
    self.display()

  def buttonMoveRowUp (self, button):
    if button._row == 0:
      return
    oldrow = self.rows[button._row-1]
    self.rows[button._row-1] = self.rows[button._row]
    self.rows[button._row] = oldrow

    self.setRowMetadata()
    self.display()

  def buttonAddRowBefore (self, button):
    self.rows.insert(button._row, AnswerRow(self, button._row))
    self.setRowMetadata()
    self.display()

  def addEmptyRow (self):
    self.rows.append(AnswerRow(self, len(self.rows)))

  def composeColumns (self):
    rcol = [self.header.col_row]
    for row in self.rows:
      rcol.extend(row.col_row)
    rcol.append(self.footer.col_row)

    vcol = [self.header.col_vertices]
    for row in self.rows:
      vcol.extend(row.col_vertices)
    vcol.append(self.footer.col_vertices)

    acol = [self.header.col_actions]
    for row in self.rows:
      acol.extend(row.col_actions)
    acol.append(self.footer.col_actions)

    dcol = [self.header.col_data]
    for row in self.rows:
      dcol.extend(row.col_data)
    dcol.append(self.footer.col_data)

    ccol = [self.header.col_controls]
    for row in self.rows:
      ccol.extend(row.col_controls)
    ccol.append(self.footer.col_controls)

    return (rcol, vcol, acol, dcol, ccol)

  @property
  def hide_controls (self):
    return self._hidden_controls

  @hide_controls.setter
  def hide_controls (self, val):
    if val == True:
      for row in self.rows:
        row.hide_controls = True
    elif val == False:
      for row in self.rows:
        row.hide_controls = False

    (rcol, vcol, acol, dcol, ccol) = self.composeColumns()
    self._vbox_rows.children = rcol
    self._vbox_vertices.children = vcol
    self._vbox_actions.children = acol
    self._vbox_data.children = dcol
    self._vbox_controls.children = ccol

  def display (self):
    (rcol, vcol, acol, dcol, ccol) = self.composeColumns()

    if self._hbox:
      self._vbox_rows.children = rcol
      self._vbox_vertices.children = vcol
      self._vbox_actions.children = acol
      self._vbox_data.children = dcol
      self._vbox_controls.children = ccol
    else:
      self._vbox_rows = widgets.VBox(rcol, layout=widgets.Layout(align_content="center"))
      self._vbox_rows.layout.display = 'flex'
      self._vbox_rows.layout.flex_flow = 'column'
      self._vbox_rows.layout.align_items = 'center'
      self._vbox_vertices = widgets.VBox(vcol)
      self._vbox_actions = widgets.VBox(acol)
      self._vbox_data = widgets.VBox(dcol)
      self._vbox_controls = widgets.VBox(ccol)
      self._hbox = widgets.HBox([self._vbox_rows, self._vbox_vertices,
                                 self._vbox_actions, self._vbox_data, self._vbox_controls])
    return self._hbox


def getMACs (manifest):
  macs = ["ff:ff:ff:ff:ff:ff"]
  for node in manifest.containers:
    for port in node.ports:
      macs.append(port.macaddress)
  macs.sort()
  return macs

def getBridgePorts (manifest, client_id):
  dp = manifest[client_id]
  ports = [x.name for x in dp.ports]
  ports.sort()
  return ports

def getVLANs (manifest, include_zero = False):
  vlans = set([])
  for dp in manifest.datapaths:
    for port in dp.ports:
      vlans.add(port.vlan_id)
  if include_zero:
    vlans.add(0)
  else:
    vlans.discard(0)
  return sorted(vlans)

def getIPs (sliverstatus):
  ips = []
  for ctr in sliverstatus["vts_containers"]:
    for port in ctr["ports"]:
      ips.extend(port["ipv4-addresses"])
  ips.sort(key = lambda x: IPy.IP(x.split("/")[0]).int())
  return ips

def getInterfaces (manifest):
  intfs = []
  for ctr in manifest.containers:
    for port in ctr.ports:
      intfs.append("%s.%s" % (ctr.client_id, port.name))
  intfs.sort()
  return intfs

def getHostInterfaces (manifest, host):
  intfs = []
  for ctr in manifest.containers:
    if ctr.client_id == host:
      for port in ctr.ports:
        intfs.append(port.name)
  intfs.sort()
  return intfs

def getVLANIP (manifest, sliverstatus, host, valid_hosts = None):
  hobj = None
  for hobj in sliverstatus["vts_containers"]:
    if hobj["client-id"] == host:
      break
  hintf = random.choice(hobj["ports"])
  ip = IPy.IP(hintf["ipv4-addresses"][0], make_net=True)

  if not valid_hosts:
    valid_hosts = [x.client_id for x in manifest.containers if len(x.ports) == 1]

  for c in sliverstatus["vts_containers"]:
    if c["client-id"] == host or c["client-id"] not in valid_hosts:
      continue
    for port in c["ports"]:
      for ipstr in port["ipv4-addresses"]:
        test_ip_net = IPy.IP(ipstr, make_net=True)
        if test_ip_net == ip:
          return ipstr.split("/")[0]

def getExternalIP (manifest, sliverstatus, host, valid_hosts = None):
  hobj = None
  for hobj in sliverstatus["vts_containers"]:
    if hobj["client-id"] == host:
      break
  hintf = random.choice(hobj["ports"])
  ip = IPy.IP(hintf["ipv4-addresses"][0], make_net=True)

  if not valid_hosts:
    valid_hosts = [x.client_id for x in manifest.containers if len(x.ports) == 1]

  for c in sliverstatus["vts_containers"]:
    if c["client-id"] == host or c["client-id"] not in valid_hosts:
      continue
    for port in c["ports"]:
      for ipstr in port["ipv4-addresses"]:
        test_ip_net = IPy.IP(ipstr, make_net=True)
        if test_ip_net != ip:
          return ipstr.split("/")[0]

def getSubnetIP (manifest, sliverstatus, host, valid_hosts = None):
  return getVLANIP(manifest, sliverstatus, host, valid_hosts)

def generateRouteIPs (manifest, sliverstatus, host, default, nondefault):
  pass

