# Copyright (c) 2020-2021  University of Houston Networking Lab

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import itertools
import random

import dateutil

from . import exceptions

class Course(object):
  class UnknownUUIDError(exceptions.LabserveError):
    def __init__ (self, euuid):
      super(Course.UnknownUUIDError, self).__init__(1001)
      self.msg = "Unknown UUID: %s" % (euuid)

  class UnknownCourseError(exceptions.LabserveError):
    def __init__ (self, cname):
      super(Course.UnknownCourseError, self).__init__(1002)
      self.msg = "Unknown course: %s" % (cname)

  def __init__ (self, name, labdb):
    self._type = "scheduled"

    self.name = name
    self._db = labdb
    self.labserve_url = "http://localhost:9008"

    self._modules = []     # [<uuid>, ...]

    self._exercises = []   # [<xrcsobj>, ...]
    self._xrcsmap = {}

    self._exams = []       # [<exobj>, ...]
    self._exammap = {}

    self._instructors = [] # [<urn>, ...]
    self._sites = []       # [<site-name>, ...]
    self._certinfo = None
    self._users = []       # [<short-uid>, ...]
    self._user_sites = {}  # {<userurn> : <site-name>, ...}
    self._realms = []      # [<string>, ...]

    self._user_dates = {}   # { <urn> : { <uuid> : {"hard-start" : <date>, "hard-stop" : <date> }, ... }, ... }

  @property
  def lab_modules (self):
    return self._modules

  @property
  def short_uids (self):
    return self._users

  def load (self):
    return

  def addLabModule (self, lmuuid):
    self._modules.append(lmuuid)

  def addExam (self, examobj):
    self._exams.append(examobj)
    self._exammap[examobj.uuid] = examobj

  def addExercise (self, xrcsobj):
    self._exercises.append(xrcsobj)
    self._xrcsmap[xrcsobj.uuid] = xrcsobj

  def addUserID (self, uid):
    self._users.append(uid)

  def removeUserID (self, uid):
    self._users.remove(uid)

  def setUserSite (self, urn, site):
    self._user_sites[urn] = site

  def setInstructors (self, ilist):
    self._instructors = ilist

  def isInstructor (self, urn):
    return urn in self._instructors

  def setSites (self, slist):
    self._sites = slist

  def setCertInfo (self, certinfo):
    self._certinfo = certinfo

  def setUserDate (self, user_urn, euuid, hstart, hstop):
    # Probably need a faster random search here
    for xr in itertools.chain(self._exams, self._exercises):
      if xr.uuid == euuid:
        startdate = dateutil.parser.parse(hstart)
        stopdate = dateutil.parser.parse(hstop)

        self._user_dates.setdefault(user_urn, {})[euuid] = {"hard-start" : startdate, "hard-stop" : stopdate}
        return

    raise Course.UnknownUUIDError(euuid)

  def getUserURNs (self):
    return self._user_sites.keys()

  def getUserSite (self, uurn):
    if uurn not in self._user_sites:
      site = random.choice(self._sites)
      self._user_sites[uurn] = site
      self._db._write_users()
    return self._user_sites[uurn]

  def getExercisesBefore (self, stopdate):
    retlist = []
    for exinfo in self._exercises:
      if exinfo.hard_stop and exinfo.hard_stop < stopdate:
        retlist.append(exinfo)
    return retlist

  def getExercises (self):
    return self._exercises

  def hasExerciseByUUID (self, exuuid):
    for ex in self._exercises:
      if ex.uuid == exuuid:
        return True
    return False

  def getExams (self):
    return self._exams

  def getExerciseForDate (self, curdate, user_urn):
    if user_urn not in self._user_dates:
      for exinfo in self._exercises:
        if exinfo.hard_start <= curdate and exinfo.hard_stop > curdate:
          return exinfo
    else:
      uddata = self._user_dates[user_urn]
      for exinfo in self._exinfos:
        if exinfo.uuid in uddata:
          if uddata[exinfo.uuid]["hard-start"] <= curdate and curdate < uddata[exinfo.uuid]["hard-stop"]:
            return exinfo
        else:
          if exinfo.hard_start <= curdate and curdate < exinfo.hard_stop:
            return exinfo
    return None

  def certsEnabled (self):
    return bool(self._certinfo)

  def getCertInfo (self):
    return self._certinfo

  def setRealms (self, realms_list):
    self._realms = realms_list

  def getShortUID (self, uid):
    parts = uid.split("@")
    shortuid = parts[0]
    if len(parts) > 1:
      if parts[1] not in self._realms:
        shortuid = uid
    return shortuid

  def hasUser (self, uid):
    shortuid = self.getShortUID(uid)

    if shortuid in self._users:
      return True

    return False

  def hasUserByURN (self, urn):
    return urn in self._user_sites

  def getExerciseByName (self, name):
    for xrcs in self._exercises:
      if isinstance(xrcs.toponame, (list, tuple)):
        if name in xrcs.toponame:
          return xrcs
      else:
        if name == xrcs.toponame:
          return xrcs

  def getExamForDate (self, dtime, user_urn):
    if user_urn not in self._user_dates:
      for exam in self._exams:
        if exam.hard_start <= dtime and dtime < exam.hard_stop:
          return exam
    else:
      uddata = self._user_dates[user_urn]
      for exam in self._exams:
        if exam.uuid in uddata:
          if uddata[exam.uuid]["hard-start"] <= dtime and dtime < uddata[exam.uuid]["hard-stop"]:
            return exam
        else:
          if exam.hard_start <= dtime and dtime < exam.hard_stop:
            return exam
    return None
