# Copyright (c) 2020-2021  University of Houston Networking Lab

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import datetime
import random
import time

import dateutil.parser
from flask import request

import six

from . import creds
from .exceptions import LabserveError
from .json import jsonify_ephemeral as jsonify
from .json import jsonValidate
from .restapi import Dispatcher, route

if six.PY3:
  unicode = str

class ExamStartError(LabserveError):
  def __init__ (self):
    super(ExamStartError, self).__init__(code = 101)
    self.msg = "The exam period for this course has not yet started."

class CourseTypeError(LabserveError):
  def __init__ (self):
    super(CourseTypeError, self).__init__(code = 201)
    self.msg = "Incorrect course type for request"

class CourseServer(Dispatcher):
  def __init__ (self, labsvc, app):
    super(CourseServer, self).__init__("Course Service", app.logger, app)
    self._labsvc = labsvc
    self._db = labsvc._db

  def validate (self, rjson, types):
    return jsonValidate(rjson, types, self._log)

  @route('/courses', methods=["GET"])
  def getClasses (self):
    return jsonify(list(self._db._courses.keys()))

  @route('/courses/2', methods=["GET"])
  def getClassesV2 (self):
    retl = []
    for cname,course in self._db._courses.items():
      retl.append((cname, course._type))
    return jsonify(retl)

  @route('/course/name/<cname>/instructor/<urn>', methods=["GET"])
  def getIsClassInstructor (self, cname, urn):
    course = self._db._courses[cname]
    return jsonify(course.isInstructor(urn))


  ########################
  ### Admin Management ###
  ########################

#  @route('/course/create', methods=["POST"])
#  @route('/course/name/<cname>/add-user/<user_urn>', methods=["GET"])
#  @route('/course/name/<cname>/add-lab/<lmod_uuid>', methods=["GET","POST"])
#  @route('/course/name/<cname>/add-exercise/<xrcs_uuid>', methods=["GET","POST"])
#  @route('/course/name/<cname>/add-exam/<exam_uuid>', methods=["GET", "POST"])
#  @route('/course/name/<cname>/set-order', methods=["POST"])

  #######################
  ### Ordered Actions ###
  #######################

  @route('/course/name/<cname>/status/<user_urn>', methods=["GET"])
  def getClassUserStatus (self, cname, user_urn):
    course = self._db._courses[cname]
    if course._type != "ordered":
      raise CourseTypeError()

    upos = course.getPosition(user_urn)
    data = {"user-pos" : upos, "order-summary" : course.getOrderSummary()}
    return jsonify(data)

  @route("/course/name/<cname>/current-attempt/<user_urn>", methods=["GET"])
  def getCheckUserCurrentAttempt (self, cname, user_urn):
    course = self._db._courses[cname]
    if course._type != "ordered":
      raise CourseTypeError()

    obj = course.getExerciseForDate(None, user_urn)
    if obj:
      atts = self._labsvc.generateAttempts(obj.uuid, user_urn)
      if atts:
        return jsonify(True)
    
    obj = course.getExamForDate(None, user_urn)
    if obj:
      atts = self._labsvc.generateAttempts(obj.uuid, user_urn, exam=True)
      if atts:
        return jsonify(True)

    return jsonify(False)

  @route('/course/name/<cname>/next/<user_urn>', methods=["GET"])
  def getClassUserNext (self, cname, user_urn):
    course = self._db._courses[cname]
    if course._type != "ordered":
      raise CourseTypeError()

    (ttype, task) = course.nextTask(user_urn)
    return jsonify({"type" : ttype, "ui-name" : task.ui_name, "time-limit" : task.time_limit})

  #################
  ### Exercises ###
  #################

  @route('/course/name/<cname>/exercises', methods=["GET"])
  def getCourseExercises (self, cname):
    course = self._db._courses[cname]
    return jsonify(course.getExercises())

  @route('/course/name/<cname>/exams', methods=["GET"])
  def getCourseExams (self, cname):
    course = self._db._courses[cname]
    return jsonify(course.getExams())

  ###################
  ### Lab Modules ###
  ###################

  @route('/course/<name>/modules/topo/<toponame>', methods=["GET"])
  def getClassModuleByToponame (self, name, toponame):
    course = self._db._courses[name]
    for mod in self._db.modules.values():
      try:
        if mod.topo_name == toponame:
          return jsonify(mod)
      except Exception:
        continue

  @route('/course/<name>/modules/<user_urn>', methods=["GET", "POST"])
  def getModulesForClass (self, name, user_urn):
    now = datetime.datetime.now()
    instructor = False

    course = self._db._courses[name]

    instructor = course.isInstructor(user_urn)

    rdate = None
    if request.method == "POST":
      if not request.json:
        return jsonify(None, -1, "Missing json POST data")
      if not instructor:
        return jsonify(None, -1, "Non-instructors may not provide alternate module filters")
      rdate = dateutil.parser.parse(request.json["date"])
      self._log.info("Modules request from admin (%s) for date (%s)" % (name, request.json["date"]))

    retlist = []
    for modid in course.lab_modules:
      mod = self._db.modules[modid]
      if mod.instr_only and not instructor:
        continue
      if mod.hard_start:
        if rdate:
          if mod.hard_start <= rdate:
            retlist.append(modid)
        else:
          if mod.hard_start <= now:
            retlist.append(modid)
      else:
        retlist.append(modid)
    return jsonify(retlist)

  @route('/course/<name>/site/<user>', methods=["GET"])
  def getSiteForUser (self, name, user):
    if name not in self._db._courses:
      return jsonify(None)

    course = self._db._courses[name]
    return jsonify(course.getUserSite(user))

  @route('/course/<name>/exercise/<ename>', methods=['GET'])
  def getCourseExerciseByName (self, name, ename):
    if name not in self._db._courses:
      return jsonify(None)

    course = self._db._courses[name]
    xrcs = course.getExerciseByName(ename)
    return jsonify(xrcs)

  @route('/course/<name>/exam/<user_urn>', methods=['GET', 'POST'])
  def getExam (self, name, user_urn):
    if name not in self._db._courses:
      return jsonify(None)

    course = self._db._courses[name]

    admin = False
    now = None
    if request.method == "POST":
      if not request.json:
        return jsonify(None, -1, "Missing json POST data")
      admin = request.json["admin"]
      now = datetime.datetime.strptime(request.json["date"], "%Y%m%d%H%M")
      self._log.info("Exam requested from admin (%s) for date (%s)" % (admin, request.json["date"]))

    try:
      if not now:
        now = datetime.datetime.now()
      exam = course.getExamForDate(now, user_urn)
      if not exam:
        raise ExamStartError()
      return jsonify(exam)
    except LabserveError as e:
      return jsonify(False, e.code, e.msg)
    except KeyError:
      return jsonify(None)

  ###########################
  ### Credential Handling ###
  ###########################

  @route('/course/<cname>/creds/enabled', methods=["GET"])
  def credsEnabled (self, cname):
    if cname not in self._db._courses:
      return jsonify(None)
    try:
      return jsonify(self._db._courses[cname].certsEnabled())
    except KeyError:
      return jsonify(None)

  @route('/course/<cname>/creds/issue-user/<user>', methods=["GET"])
  def credsIssueUser (self, cname, user):
    if cname not in self._db._courses:
      return jsonify(None)

    course = self._db._courses[cname]

    cert_info = course.getCertInfo()
    ucert = {}
    userkey, usercert, cert = creds.makeusercert(user, cert_info)
    ucert["certkey"] = userkey.decode('utf-8')
    ucert["usercert"] = usercert.decode('utf-8')
    ucert["usercred"] = creds.makeusercred(cert_info, cert).decode('utf-8')

    uurn = creds.returnUserURN(usercert)
    course.getUserSite(uurn)

    return jsonify(ucert)

  @route('/course/<cname>/creds/issue-slice/<slicename>', methods=["POST"])
  def credsIssueSlice (self, cname, slicename):
    if not request.json:
      return jsonify(None, -1, "Missing json POST data")

    o = self.validate(request.json, [("user-cert", (unicode, str)),
                                     ("project", (unicode, str))])

    cert_info = self._db._courses[cname].getCertInfo()
    userurn = creds.returnUserURN(o["user-cert"].encode('utf-8'))
    self._log.info("Received user cert for %s in project %s" % (userurn, o["project"]))
    scert = {}

    tscert = creds.makeslicecert(slicename, cert_info, o["project"])
    scert["slicecred"] = creds.makeslicecred(o["user-cert"].encode('utf-8'), tscert, cert_info).decode('utf-8')
    scert["slicecert"] = tscert.decode('utf-8')
    return jsonify(scert)

  ### Users ###

  @route('/course/<name>/users', methods=["GET"])
  def getCourseUserURNs (self, name):
    if name not in self._db._courses:
      return jsonify(None)

    course = self._db._courses[name]
    return jsonify(list(course.getUserURNs()))

  @route('/course/<cname>/short-uids', methods=["GET"])
  def getCourseUserIDs (self, cname):
    if cname not in self._db._courses:
      return jsonify(None)

    course = self._db._courses[cname]
    return jsonify(list(course.short_uids))


