# Copyright (c) 2018-2021  University of Houston

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import datetime
import uuid

from cryptography import x509
from cryptography.x509.oid import NameOID, ExtendedKeyUsageOID
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from lxml import etree as ET
import six
import xmlsec

# TODO logging of cred assignments
#logc = log.KeyAdapter("cred")

if six.PY3:
  from builtins import str

consts = xmlsec.constants

def _buildCSRAndKey (subject, san):
  key = rsa.generate_private_key(public_exponent = 65537, key_size = 2048, backend = default_backend())
  csr = x509.CertificateSigningRequestBuilder().subject_name(subject).add_extension(san, critical=False)
  csr = csr.add_extension(x509.KeyUsage(True, False, False, False, False, False, False, False, False), critical=True)
  csr = csr.sign(key, hashes.SHA256(), default_backend())
  return (csr, key)


def _csrIssueCert (cacert, cakey, csr, duration):
  builder = x509.CertificateBuilder().serial_number(x509.random_serial_number()).issuer_name(cacert.issuer)
  builder = builder.not_valid_before(datetime.datetime.utcnow()).not_valid_after(
    datetime.datetime.utcnow() + datetime.timedelta(days=duration))
  builder = builder.public_key(csr.public_key()).subject_name(csr.subject)
  for ext in csr.extensions:
    builder = builder.add_extension(ext.value, ext.critical)
  builder = builder.add_extension(x509.ExtendedKeyUsage([ExtendedKeyUsageOID.CLIENT_AUTH]), critical = True)
  builder = builder.add_extension(x509.BasicConstraints(False, None), critical=True)
  cert = builder.sign(cakey, hashes.SHA256(), default_backend())
  return cert


def _loadCAData (capath, cakey):
  with open(cakey, "rb") as keyfile:
    cakey = serialization.load_pem_private_key(keyfile.read(), password=None, backend=default_backend())
  with open(capath, "rb") as cafile:
    cacert = x509.load_pem_x509_certificate(cafile.read(), backend=default_backend())
  return (cakey, cacert)


def makeusercert (username, rootcert_info):
  capath = rootcert_info["cert"]
  cakey = rootcert_info["key"]
  duration = rootcert_info["duration"]
  (cakey, cacert) = _loadCAData(capath, cakey)
  # *We* don't issue roots with more than one CN, but it's not illegal, so this is a bit fragile
  fqdn = cacert.issuer.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value
  user_uuid = str(uuid.uuid4())
  subject = x509.Name([
    x509.NameAttribute(NameOID.COMMON_NAME, u"%s" % (user_uuid))
    ])
  san = x509.SubjectAlternativeName([
    x509.UniformResourceIdentifier(u"urn:publicid:IDN+%s+user+%s" % (fqdn, username)),
    x509.UniformResourceIdentifier(u"urn:uuid:%s" % (user_uuid))
    ])
  # Make the user CSR and save it and the key
  csr, userkey = _buildCSRAndKey(subject, san)
  # Sign that shit
  cert = _csrIssueCert(cacert, cakey, csr, duration)
  #return (userkey, csr, cert)
  ukey = userkey.private_bytes(encoding = serialization.Encoding.PEM,
                               format = serialization.PrivateFormat.TraditionalOpenSSL,
                               encryption_algorithm = serialization.NoEncryption())
  ucert = cert.public_bytes(serialization.Encoding.PEM)
  return (ukey, ucert, cert)


def makeslicecert (slicename, rootcert_info, project):
  capath = rootcert_info["cert"]
  cakey = rootcert_info["key"]
  duration = rootcert_info["duration"]
  (cakey, cacert) = _loadCAData(capath, cakey)
  # *We* don't issue roots with more than one CN, but it's not illegal, so this is a bit fragile
  fqdn = cacert.issuer.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value
  slice_uuid = str(uuid.uuid4())
  subject = x509.Name([
    x509.NameAttribute(NameOID.COMMON_NAME, u"%s//%s.slice.%s" % (fqdn, project, slicename))
    ])
  san = x509.SubjectAlternativeName([
    x509.UniformResourceIdentifier(u"urn:publicid:IDN+%s:%s+slice+%s" % (fqdn, project, slicename)),
    x509.UniformResourceIdentifier(u"urn:uuid:%s" % (slice_uuid))
    ])
  # Make the user CSR and save it and the key
  csr, slicekey = _buildCSRAndKey(subject, san)
  # Sign that shit
  cert = _csrIssueCert(cacert, cakey, csr, duration)
  serialcert = cert.public_bytes(serialization.Encoding.PEM)
  return serialcert

def _addPrivilege (element, priv, can_delegate):
  pelem = ET.SubElement(element, "privilege")
  name_elem = ET.SubElement(pelem, "name")
  name_elem.text = u"%s" % (str(priv))
  delegate_elem = ET.SubElement(pelem, "can_delegate")
  if can_delegate:
    delegate_elem.text = u"true"
  else:
    delegate_elem.text = u"false"


def genicred (cakey, capath, owner_gid, owner_urn, target_gid, target_urn, expires, privs):
  XSI = "http://www.w3.org/2001/XMLSchema-instance"
  XML = "http://www.w3.org/XML/1998/namespace"

  credr = ET.Element("signed-credential", attrib = {"{%s}noNamespaceSchemaLocation" % (XSI) : "http://www.planet-lab.org/resources/sfa/credential.xsd"})
  crede = ET.SubElement(credr, "credential", attrib = {"{%s}id" % (XML) : "ref0"})
  typee = ET.SubElement(crede, "type")
  typee.text = u"privilege"
  ser = ET.SubElement(crede, "serial")
  ser.text = u"8"
  ogide = ET.SubElement(crede, "owner_gid")
  ogide.text = u"%s" % (owner_gid.decode('utf-8'))
  ourne = ET.SubElement(crede, "owner_urn")
  ourne.text = u"%s" % (str(owner_urn))
  tgide = ET.SubElement(crede, "target_gid")
  tgide.text = u"%s" % (target_gid.decode('utf-8'))
  turne = ET.SubElement(crede, "target_urn")
  turne.text = u"%s" % (str(target_urn))
  uuide = ET.SubElement(crede, "uuid")
  expirese = ET.SubElement(crede, "expires")
  expirese.text = u"%s" % (str(expires.strftime("%Y-%m-%dT%H:%M:%SZ")))

  pvse = ET.SubElement(crede, "privileges")
  for priv in privs:
    _addPrivilege(pvse, priv, False)

  sigse = ET.SubElement(credr, "signatures")

  # Sign this mess
  # pylint: disable=E1101
  sign = xmlsec.template.create(crede, consts.TransformExclC14N, consts.TransformRsaSha1)
  sign.attrib["{%s}id" % (XML)] = u"Sig_ref0"
  sigse.append(sign)
  ref = xmlsec.template.add_reference(sign, consts.TransformSha1, uri="#ref0")
  xmlsec.template.add_transform(ref, consts.TransformEnveloped)
  ki = xmlsec.template.ensure_key_info(sign)
  xmlsec.template.add_x509_data(ki)

  ctx = xmlsec.SignatureContext()
  ctx.key = xmlsec.Key.from_file(cakey, format=consts.KeyDataFormatPem)
  ctx.key.load_cert_from_file(capath, consts.KeyDataFormatPem)
  ctx.sign(sign)
  # pylint: enable=E1101

  return credr


def makeusercred (rootcert_info, ucert):
  capath = rootcert_info["cert"]
  cakey = rootcert_info["key"]
  duration = rootcert_info["duration"]
  # Dig the user URN out of the client cert
  sanext = ucert.extensions.get_extension_for_oid(x509.SubjectAlternativeName.oid)
  urilist = sanext.value.get_values_for_type(x509.UniformResourceIdentifier)
  for uri in urilist:
    if uri.count("+user+"):
      userurn = uri
      break

  expires = datetime.datetime.now() + datetime.timedelta(days = duration)
  gidbytes = ucert.public_bytes(serialization.Encoding.PEM)
  credroot = genicred(cakey, capath, gidbytes, userurn, gidbytes, userurn, expires, ["refresh", "resolve", "info"])
  #return credroot
  return ET.tostring(credroot, pretty_print = False)

def returnUserURN (user_cert):
  # Dig the user URN out of the client cert
  ucert = x509.load_pem_x509_certificate(user_cert, backend = default_backend())
  sanext = ucert.extensions.get_extension_for_oid(x509.SubjectAlternativeName.oid)
  urilist = sanext.value.get_values_for_type(x509.UniformResourceIdentifier)
  for uri in urilist:
    if uri.count("+user+"):
      userurn = uri
      break
  return userurn

def makeslicecred (user_cert, slicecert, rootcert_info):
  capath = rootcert_info["cert"]
  cakey = rootcert_info["key"]
  duration = rootcert_info["duration"]
  ucert = x509.load_pem_x509_certificate(user_cert, backend = default_backend())
  scert = x509.load_pem_x509_certificate(slicecert, backend = default_backend())
  # Dig the user URN out of the client cert
  usanext = ucert.extensions.get_extension_for_oid(x509.SubjectAlternativeName.oid)
  urilist = usanext.value.get_values_for_type(x509.UniformResourceIdentifier)
  for uri in urilist:
    if uri.count("+user+"):
      userurn = uri
      break

  # Now lets find ourselves a slice urn
  ssanext = scert.extensions.get_extension_for_oid(x509.SubjectAlternativeName.oid)
  urilist = ssanext.value.get_values_for_type(x509.UniformResourceIdentifier)
  for uri in urilist:
    if uri.count("+slice+"):
      sliceurn = uri
      break

  expires = datetime.datetime.now() + datetime.timedelta(days = duration)
  ownergid_bytes = ucert.public_bytes(serialization.Encoding.PEM)
  targetgid_bytes = scert.public_bytes(serialization.Encoding.PEM)

  credroot = genicred(cakey, capath, ownergid_bytes, userurn, targetgid_bytes, sliceurn, expires, ["*"])
  return ET.tostring(credroot, pretty_print = False)
