# Copyright (c) 2020  University of Houston Networking Lab

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import uuid

import dateutil.parser

from .datafile import DataFile
from .exceptions import LabserveError, UnsupportedVersionError

TIMEFMT = "%Y-%m-%d %H:%M"

class Exercise(DataFile):
  class NoStartDateError(LabserveError):
    def __init__ (self):
      super(Exercise.NoStartDateError, self).__init__(2001)
      self.msg = "Exercises must contain start dates"

  class StartStopMismatchError(LabserveError):
    def __init__ (self):
      super(Exercise.StartStopMismatchError, self).__init__(2002)
      self.msg = "Exercise stop date must be after start date"

  EXTENSION='.xrcs'

  def __init__ (self):
    super(Exercise, self).__init__()
    self.uuid = str(uuid.uuid4())
    self.actions = []
    self.toponame = None
    self.course = None
    self.question_data = []
    self.version = 1
    self.layout_engine = "neato"
    self.global_graph = True
    self.time_limit = None
    self.hard_stop = None
    self.hard_start = None
    self.akey_name = None
    self.ui_name = None
    self.disable_practice = False

  def __json__ (self):
    obj = {}
    obj["uuid"] = self.uuid
    obj["version"] = self.version
    obj["course"] = self.course
    if self.version >= 3:
      obj["lab-topo-names"] = self.toponame
    else:
      obj["topo-name"] = self.toponame
    obj["question-data"] = self.question_data
    obj["layout-engine"] = self.layout_engine
    obj["global-graph"] = self.global_graph
    obj["time-limit"] = self.time_limit
    obj["hard-stop"] = self.hard_stop
    obj["hard-start"] = self.hard_start

    if self.hard_stop:
      obj["hard-stop"] = obj["hard-stop"].strftime(TIMEFMT)
    if self.hard_start:
      obj["hard-start"] = obj["hard-start"].strftime(TIMEFMT)

    if self.akey_name:
      obj["akey-name"] = self.akey_name

    obj["ui-name"] = self.ui_name
    obj["disable-practice"] = self.disable_practice

    return obj

  @classmethod
  def jconstruct (cls, obj):
    ex = cls()
    ex.time_limit = obj.get("time-limit")

    if obj["version"] >= 8:
      raise UnsupportedVersionError(obj["version"])

    if obj["version"] >= 1:
      if "uuid" in obj:
        ex.uuid = obj["uuid"]

      ex.version = obj["version"]
      ex.course = obj["course"]
      ex.toponame = obj.get("topo-name")
      ex.layout_engine = obj.get("layout-engine")
      ex.disable_practice = obj.get("disable-practice", False)

      ex.question_data = obj["question-data"]

    if obj["version"] >= 2:
      ex.global_graph = obj["global-graph"]

    if obj["version"] >= 3:
      ex.toponame = obj["lab-topo-names"]

    if obj["version"] >= 4:
      if "hard-stop" in obj and obj["hard-stop"]:
        ex.hard_stop = dateutil.parser.parse(obj["hard-stop"])
      if "hard-start" in obj and obj["hard-start"]:
        ex.hard_start = dateutil.parser.parse(obj["hard-start"])

    if obj["version"] >= 5:
      ex.akey_name = obj["akey-name"]

    if obj["version"] >= 6:
      ex.ui_name = obj.get("ui-name")

    if obj["version"] >= 7:
      ex.hard_start = dateutil.parser.parse(obj["hard-start"])

    if not ex.hard_start:
      raise Exercise.NoStartDateError()

    if ex.hard_start >= ex.hard_stop:
      raise Exercise.StartStopMismatchError()

    return ex
