# Copyright (c) 2017-2021  University of Houston Networking Lab

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import base64
import datetime
import os
import os.path
import random

from . import json
from .answerkey import AnswerKey
from .course import Course
from .datafile import DataFile
from .exam import Exam
from .exceptions import LabserveError
from .exercise import Exercise
from .log import KeyAdapter
from .module import Module
from .orderedcourse import OrderedCourse
from .scenario import Scenario

TEXT_EXTS = set(["json", "xml"])
DATEFMT = "%Y%m%d%H%M"

def getLookupTable():
  #lookup={v.EXTENSION:v for v in DataFile.__subclasses__()}
  lookup = {}
  for v in globals().values():
    if not isinstance(v, type):
      continue
    if issubclass(v, DataFile):
      lookup[v.EXTENSION] = v
  return lookup

class UnknownDatapoolError(LabserveError):
  def __init__ (self, name):
    super(UnknownDatapoolError, self).__init__(code = 100)
    self.name = name
    self.msg = str(self)
  def __str__ (self):
    return "Requested datapool (%s) is not known here" % (self.name)

class DB(object):
  def __init__ (self, cfg, logger):
    self._cfg = cfg
    self._root = "%s/data" % (cfg["basedir"])
    self._log = KeyAdapter("DB", logger)
    self._imports = "%s/IMPORT" % (self._root)
    self._acctdir = "%s/accounting" % (cfg["basedir"])
    self._user_data = "%s/user_data.json" % (self._root)
    self._user_course = "%s/user_course.json" %(self._root)
    self._site_data = "%s/site_data.json" % (self._root)
    self._user_exceptions = "%s/user_exceptions.json" % (self._root)

    self.modules = {}
    self.scenarios = {} # { <scen-id> : <scenario-obj>, ...}
    self.data_pools = {}  # { <dpname> : { 'count' : <int>, 'path' : <str> }, ... }
    self.answer_keys = {}

    # additional indexes
    self.akey_by_name = {}

    self._courses = {}

    # Configuration
    self._disabled_lmods = set()
    self._disabled_exercises = set()
    self._accounting_umask = 0o0700
    self.config = {}

    self._load()

  def _write_users (self):
    udata = {}
    for course in self._courses.values():
      udata[course.name] = {}
      for uurn,site in course._user_sites.items():
        udata[course.name][uurn] = site
    with open(self._user_data, "w+") as f:
      f.write(json.dumps(udata))

  def _load (self):
    if not os.path.exists(self._imports):
      self._log.info("Creating import directory (%s)" % (self._imports))
      os.makedirs(self._imports)

    self._doImports()

    self._loadConfiguration()
    self._loadSiteData()
    self._loadData()
    self._loadUserData()
    self._loadUserCourse()
    self._loadDataPools()
    self._loadUserExceptions()
    self._loadCourses()


  def _loadConfiguration (self):
    cpath = "%s/configdata.json" % (self._root)
    if not os.path.exists(cpath):
      return

    self._log.info("Loading configuration metadata from %s" % (cpath))

    obj = json.loads(open(cpath, "rb").read())
    if "disabled-xrcs" in obj:
      self._disabled_exercises = set(obj["disabled-xrcs"])
    if "disabled-lmod" in obj:
      self._disabled_lmods = set(obj["disabled-lmod"])
    if "accounting-umask" in obj:
      self._accounting_umask = int(obj["accounting-umask"], 8)
    if "view-root" in obj:
      self.config["view-root"] = obj["view-root"]


  def _loadData (self):
    lookup = getLookupTable()

    for fname in os.listdir(self._root):
      if fname[-5:] not in lookup:
        continue

      obj = json.loads(open("%s/%s" % (self._root, fname), "r").read())
      try:
        instance = lookup[fname[-5:]].jconstruct(obj)
      except Exception as e:
        self._log.exception("Exception loading %s" % (fname))
        continue

      if isinstance(instance, Module):
        self.modules[instance.uuid] = instance
        if instance.uuid not in self._disabled_lmods:
          self._courses[instance.course].addLabModule(instance.uuid)
        else:
          self._log.debug("Skipping index for disabled lmod (%s)" % (instance.uuid))
      elif isinstance(instance, Exam):
        self._courses[instance.course].addExam(instance)
      elif isinstance(instance, Exercise):
        if not instance.hard_start:
          self._log.warning("Skipping exercise without hard-start (%s)" % (instance.uuid))

        if instance.uuid not in self._disabled_exercises:
          self._courses[instance.course].addExercise(instance)
        else:
          self._log.debug("Skipping index for disabled xrcs (%s)" % (instance.uuid))

      elif isinstance(instance, Scenario):
        self.scenarios[instance.uuid] = instance
      elif isinstance(instance, AnswerKey):
        self.answer_keys[instance.uuid] = instance
        self.akey_by_name[instance.info['akey-name']] = instance


  def _loadUserData (self):
    if os.path.exists(self._user_data):
      for cname,usermap in json.loads(open(self._user_data, "r").read()).items():
        for uurn,site in usermap.items():
          self._courses.setdefault(cname, Course(cname, self)).setUserSite(uurn, site)

  def _loadUserCourse (self):
    if os.path.exists(self._user_course):
      ucmap = json.loads(open(self._user_course, "r").read())
      for user,clist in ucmap.items():
        for cname in clist:
          self._courses.setdefault(cname, Course(cname, self)).addUserID(user)

  def _writeUserCourse (self):
    ucmap = {}
    for cname,course in self._courses.items():
      for uid in course.short_uids:
        ucmap.setdefault(uid, []).append(cname)
    with open(self._user_course, "w+") as ucf:
      ucf.write(json.dumps(ucmap))

  def _loadUserExceptions (self):
    # This file contains all the per-user exceptions that will be made to previously
    # loaded course information (exam dates, etc.)  It's hinky and we should do
    # something better eventually.
    if os.path.exists(self._user_exceptions):
      uelist = json.loads(open(self._user_exceptions, "r").read())
      for idx,entry in enumerate(uelist):
        # (<course>, <userurn>, <type>, <type_data>)
        if entry[0] not in self._courses:
          self._log.warning("User Exception (%d) references unknown course (%s)" % (
            idx, entry[0]))
          continue
        course = self._courses[entry[0]]

        if not course.hasUserByURN(entry[1]):
          self._log.warning("User Exception (%d) references unknown user (%s) in course (%s)" % (
            idx, entry[1], entry[0]))
          continue

        if entry[2] == "edate":
          try:
            course.setUserDate(entry[1], **entry[3])
          except LabserveError as e:
            self._log.error("User Exception (%d) error: %s" % (idx, str(e)))
            continue
        else:
          self._log.warning("User Exception (%d) is of unknown type (%s)" % (
            idx, entry[2]))
          continue

        self._log.info("User Exception (%d) of type (%s) processed: [%s][%s]" % (
          idx, entry[2], entry[0], entry[1]))

  def _doImports (self):
    lookup = getLookupTable()
    for fname in os.listdir(self._imports):
      self._log.debug("Found import file (%s)" % (fname))

      try:
        self._log.info("Importing (%s)" % (fname))
        obj = json.loads(open("%s/%s" % (self._imports, fname), "r").read())
        cls = lookup[fname[-5:]].jconstruct(obj)
        with open("%s/%s%s" % (self._root, cls.uuid, cls.EXTENSION), "w+") as impf:
          data = json.dumps(cls)
          impf.write(data)
        os.unlink("%s/%s" % (self._imports, fname))
      except Exception:
        self._log.exception("Exception loading lab module JSON")
        continue

  def _loadSiteData (self):
    if not os.path.exists(self._site_data):
      self._log.info("No site_data.json found")
      return

    sdobj= json.loads(open(self._site_data, "r").read())
    version = 1
    if "_version" in sdobj:
      version = sdobj["_version"]

    # v3 format
    # {"_version" : 3,
    #  "cn4421" :
    #    { "sites": [ [ "vts-local", "127.0.0.1"], "vts-gpo", ... ],
    #      "root-cert" : "vts-4421",
    #      "cred-max-duration" : 150,
    #      "instructors" : [ <urn>, ...] },
    #   ...
    # }

    if version <= 2:
      self._log.error("site_data versions before v3 are not supported")
    if version >= 3:
      for k,course_data in sdobj.items():
        if k == "_version":
          continue
        if course_data.get("ordered", False):
          self._log.debug("Loading ordered course [%s]" % (k))
          course = self._courses.setdefault(k, OrderedCourse(k, self))
        else:
          self._log.debug("Loading course [%s]" % (k))
          course = self._courses.setdefault(k, Course(k, self))

        course.setInstructors(course_data.get("instructors", []))
        course.setSites(course_data.get("sites", []))
        if "root-cert" in course_data:
          self._log.debug("Loading root certificates for [%s]" % (k))
          certinfo = {}
          certinfo["cert"] = "%s/certs/%s.crt" % (self._cfg["basedir"], course_data["root-cert"])
          certinfo["key"] = "%s/certs/%s.key" % (self._cfg["basedir"], course_data["root-cert"])
          certinfo["duration"] = course_data["cred-max-duration"]
          course.setCertInfo(certinfo)
        if "realms" in course_data:
          course.setRealms(course_data["realms"])
    if version >= 4:
      for k,course_data in sdobj.items():
        if k == "_version":
          continue

        course = self._courses[k]
        course.labserve_url = course_data["labserve-url"]

        if course_data["ordered"]:
          course.start_date = datetime.datetime.strptime(course_data["start-date"], DATEFMT)
          course.stop_date = datetime.datetime.strptime(course_data["stop-date"], DATEFMT)


  def _loadDataPools (self):
    dpdir = "%s/datapools" % (self._root)
    if not os.path.exists(dpdir):
      self._log.warning("No data pool directory found")
      return

    for dirname in os.listdir(dpdir):
      pool_path = "%s/%s" % (dpdir, dirname)
      if not os.path.isdir(pool_path):
        continue

      pool_size = len(os.listdir(pool_path))
      self._log.info("Loading data pool (%s) of size (%d)" % (dirname, pool_size))
      self.data_pools[dirname] = {"count" : pool_size, "path" : pool_path}

  def _loadCourses (self):
    for course in self._courses.values():
      course.load()

  def getDatapoolValue (self, dpname, index=None):
    if dpname not in self.data_pools:
      raise UnknownDatapoolError(dpname)

    dpinfo = self.data_pools[dpname]
    if index is None:
      index = random.randint(0, dpinfo["count"] - 1)

    dpv = {"index" : index}
    datadir = "%s/datapools/%s/%d" % (self._root, dpname, index)
    for fname in os.listdir(datadir):
      parts = fname.split(".")
      if parts[-1] in TEXT_EXTS:
        with open("%s/%s" % (datadir, fname), "r") as fhandle:
          fdata = fhandle.read()
          dpv[fname.rsplit(".", 1)[0]] = (fdata, parts[-1])
      else:
        with open("%s/%s" % (datadir, fname), "rb") as fhandle:
          fdata = fhandle.read()
          dpv[fname.rsplit(".", 1)[0]] = (base64.b64encode(fdata).decode("utf-8"), parts[-1])
    return dpv

