# Copyright (c) 2021  University of Houston Networking Lab

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from flask import request

from .db import TeamsDB
from ..course import Course
from ..exceptions import LabserveError
from ..json import jsonify_ephemeral as jsonify
from ..json import jsonValidate
from ..restapi import Dispatcher, route

class MSTeamsServer(Dispatcher):
  def __init__ (self, labsvc, app):
    super(MSTeamsServer, self).__init__("MS Teams Service", app.logger, app)
    self._labsvc = labsvc
    self._ldb = labsvc._db
    self._db = TeamsDB(self._ldb, app.logger)

  def validate (self, rjson, types):
    return jsonValidate(rjson, types, self._log)

  @route('/msteams/course/<cname>/uuid', methods = ["GET", "POST"])
  def getpostCourseTeamsUUID (self, cname):
    if cname not in self._ldb._courses:
      raise Course.UnknownCourseError(cname)

    if request.method == "POST":
      if not request.json:
        return jsonify(None, -1, "Missing json POST data")
      o = self.validate(request.json, [("uuid", str)])
      self._db.setCourseUUID(cname, o["uuid"])
      return jsonify(True)
    else:
      return jsonify(self._db.getCourseUUID(cname))

  @route('/msteams/course/<cname>/exercise/<euuid>/assignment', methods = ["GET","POST"])
  def getpostExerciseAssignmentData (self, cname, euuid):
    if cname not in self._ldb._courses:
      raise Course.UnknownCourseError(cname)
    course = self._ldb._courses[cname]

    if not course.hasExerciseByUUID(euuid):
      raise Course.UnknownUUIDError(euuid)

    if request.method == "POST":
      if not request.json:
        return jsonify(None, -1, "Missing json POST data")
      o = self.validate(request.json, [("uuid", str)])
      self._db.setCourseAssignment(cname, euuid, o["uuid"])
      return jsonify(True)
    else:
      return jsonify(self._db.getCourseAssignment(cname, euuid))

  @route('/msteams/course/<cname>/users', methods = ["GET"])
  def getCourseUsers (self, cname):
    if cname not in self._ldb._courses:
      raise Course.UnknownCourseError(cname)

    course = self._ldb._courses[cname]
    ret = []
    for suid in course.short_uids:
      try:
        ret.append((suid, self._db._users[suid]))
      except KeyError:
        ret.append((suid, None))
    return jsonify(ret)

  @route('/msteams/user/<shortid>/uuid/<teamsuuid>', methods = ["PUT"])
  def putUserUUID (self, shortid, teamsuuid):
    if shortid in self._db._users:
      self._log.debug("Setting new UUID (%s) for known user (%s)" % (teamsuuid, shortid))
    self._db._users[shortid] = teamsuuid
    self._db._writeUsers()
    return jsonify(True)


