# Copyright (c) 2020-2021  University of Houston Networking Lab

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import os.path

from . import json
from . import course
from .exam import Exam
from .exercise import Exercise

# NOTE: There is no "accelerating" your position in OrderedCourse (e.g. perspective date)
# Everything is processed based on your current position

class OrderedCourse(course.Course):
  def __init__ (self, name, labdb):
    super(OrderedCourse, self).__init__(name, labdb)
    self._type = "ordered"

    self._start_date = None
    self._stop_date = None
    self._order = []
    self._user_positions = {"_version" : 1}
    self._prog_path = "%s/%s.prog.oc" % (self._db._root, self.name)
    self._order_path = "%s/%s.order.oc" % (self._db._root, self.name)

  @property
  def start_date (self):
    return self._start_date

  @start_date.setter
  def start_date (self, val):
    self._start_date = val

  @property
  def stop_date (self):
    return self._stop_date

  @stop_date.setter
  def stop_date (self, val):
    self._stop_date = val

  def load (self):
    if not os.path.exists(self._order_path):
      return
    orderdata = json.loads(open(self._order_path, "r").read())
    for typ,euuid in orderdata["tasks"]:
      if typ == Exercise.EXTENSION:
        self._order.append(self._xrcsmap[euuid])
      elif typ == Exam.EXTENSION:
        self._order.append(self._exammap[euuid])

    if not os.path.exists(self._prog_path):
      return
    progdata = json.loads(open(self._prog_path, "r").read())
    if progdata["_version"] == 1:
      self._user_positions = progdata

  def _writeProgress (self):
    with open(self._prog_path, "w") as pf:
      pf.write(json.dumps(self._user_positions))

  def getOrderSummary (self):
    osum = []
    for obj in self._order:
      osum.append({"type" : obj.EXTENSION, "ui-name" : obj.ui_name, "uuid" : obj.uuid})
    return osum

  def getUserURNs (self):
    return [x for x in self._user_positions.keys() if x != "_version"]

  def getPosition (self, user_urn):
    return self._user_positions.setdefault(user_urn, 0)

  def getExamForDate (self, dtime, user_urn):
    cur = self._order[self.getPosition(user_urn)]
    if isinstance(cur, (Exam)):
      return cur
    return None

  def getExercises (self):
    return [x for x in self._order if not isinstance(x, Exam)]

  def getExerciseForDate (self, dtime, user_urn):
    cur = self._order[self.getPosition(user_urn)]
    if isinstance(cur, (Exercise)):
      return cur
    return None

  def nextTask (self, user_urn):
    newpos = self.getPosition(user_urn) + 1
    self._user_positions[user_urn] = newpos
    self._writeProgress()
    return (self._order[newpos].EXTENSION, self._order[newpos])
