# Copyright (c) 2017-2021  University of Houston Networking Lab

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import base64
import datetime
import errno
import json
import os
import os.path
import random
import time

import dateutil.parser
from flask import request
import six

from . import labdb, creds
from .adminserve import AdminServer
from .courseserve import CourseServer
from .exceptions import LabserveError
from .json import jsonify_ephemeral as jsonify
from .json import jsonValidate
from .restapi import Dispatcher, route
from .msteams.serve import MSTeamsServer

if six.PY3:
  unicode = str

def parseBool(val):
  TRUE_STRS = ["t", "true", "y", "yes"]
  FALSE_STRS = ["f", "false", "n", "no"]
  if val.lower() in TRUE_STRS:
    return True
  elif val.lower() in FALSE_STRS:
    return False
  else:
    raise ValueError("Give Correct boolean values")


class ModuleServer(Dispatcher):
  def __init__ (self, app, cfg):
    super(ModuleServer, self).__init__("UHED Module Server", app.logger, app)
    self._db = labdb.DB(cfg, app.logger)
    if not os.path.exists(self._db._acctdir):
      os.makedirs(self._db._acctdir, self._db._accounting_umask)
    self._courseserve = CourseServer(self, app)
    self._adminserve = AdminServer(self, app)
    self._teamserve = MSTeamsServer(self, app)

  def validate (self, rjson, types):
    return jsonValidate(rjson, types, self._log)

  @route('/module/<uuid>', methods=["GET"])
  def getModule (self, uuid):
    return jsonify(self._db.modules[uuid])

  ###################
  ### Site Config ###
  ###################

  @route('/config/<key>', methods=["GET"])
  def getConfigKeyValue (self, key):
    if key in self._db.config:
      return jsonify(self._db.config[key])
    else:
      return jsonify(None)

  #################
  ### Scenarios ###
  #################

  @route('/scenarios/modules', methods=["GET"])
  def getScenarioModules (self):
    return jsonify(self._db.scenarios.keys())

  @route('/scenarios/<uuid>', methods=["GET"])
  def getScenario (self, uuid):
    return jsonify(self._db.scenarios[uuid])

  #################
  ### Exercises ###
  #################

  # YYYYMMDDhhmm
  @route('/exercises/<course>/stop-before/<datestr>/user/<user_urn>', methods=['GET'])
  def getExercisesAfterStop (self, course, datestr, user_urn):
    sdate = datetime.datetime.strptime(datestr, "%Y%m%d%H%M")
    retlist = self._db._courses[course].getExercisesBefore(sdate)
    return jsonify(retlist)

  @route('/exercises/<cname>/current/<datestr>/user/<urn>', methods=['GET'])
  def getExerciseForDate (self, cname, datestr, urn):
    sdate = datetime.datetime.strptime(datestr, "%Y%m%d%H%M")
    course = self._db._courses[cname]
    return jsonify(course.getExerciseForDate(sdate, urn))

  @route('/answer-key/uuid/<uuid>', methods=['GET'])
  def getAKeyByUUID (self, uuid):
    try:
      return jsonify(self._db.answer_keys[uuid])
    except KeyError:
      return jsonify(None)

  @route('/answer-key/name/<name>', methods=['GET'])
  def getAKeyByName (self, name):
    try:
      return jsonify(self._db.akey_by_name[name])
    except KeyError:
      return jsonify(None)

  def generateAttempts (self, euuid, user_urn, exam=False):
    files = []
    modules = []
    if exam:
      folder = "%s/exams/%s/%s" % (self._db._acctdir, euuid, user_urn)
    else:
#      if euuid not in self._db.modules_by_euuid:
#        self._log.info("Got an unknown Exercise UUID %s" % (euuid))

      folder = "%s/exercises/%s/%s" % (self._db._acctdir, euuid, user_urn)

    try:
      for fname in os.listdir(folder):
        fullname = '%s/%s' %(folder,fname)
        timestamp = fname[12:]
        files.append((timestamp,fullname))
    except OSError as e:
      if e.errno == errno.ENOENT:
        self._log.info("Folder does not exist %s" % (folder))

    files.sort(key=lambda x: x[0])

    result = []
    for index,(ts,fname) in enumerate(files):
      result.append((index,ts,fname))

    return result

  @route('/exercise/<euuid>/user/<user_urn>/attempts/', methods=['GET'])
  def getAttempts (self, euuid, user_urn):
    result = self.generateAttempts(euuid, user_urn)
    result = [(x,y) for x,y,_ in result]
    return jsonify(result)

  @route('/exercise/<euuid>/user/<user_urn>/attempt/<int:index>', methods=['GET'])
  def geAttemptByIndex (self, euuid, user_urn, index):
    result = self.generateAttempts(euuid, user_urn)
    attempt = result[index]
    with open(attempt[2], "r") as f:
      return jsonify(f.read())

# TODO: get rid of muuid on client side
  @route('/module/<muuid>/exercise/<euuid>/user/<user_urn>', methods=['POST'])
  def postExerciseAnswers (self, muuid, euuid, user_urn): # pylint: disable=R0201
    if not request.json:
      return jsonify(None, -1, "Missing json POST data")
    ddir = "%s/exercises/%s/%s" % (self._db._acctdir, euuid, user_urn)
    if not os.path.exists(ddir):
      os.makedirs(ddir, self._db._accounting_umask)

    dpath = "%s/answer.json.%.6f" % (ddir, time.time())
    with open(dpath, "w+") as f:
      f.write(json.dumps(request.json))
    return jsonify(True)


  ################
  ### Practice ###
  ################

  #################
  ### Datapools ###
  #################

  @route('/datapool/<dpname>/acquire/<user_urn>', methods=['GET'])
  def getDatapoolValues (self, dpname, user_urn):
    try:
      data = self._db.getDatapoolValue(dpname)
      self._log.info("Retrieved (%s:%d) for %s" % (dpname, data["index"], user_urn))
      return jsonify(data)
    except LabserveError as e:
      self._log.exception("Datapool acquire error")
      return jsonify(False, e.code, e.msg)
    except Exception as e:
      self._log.exception("Datapool acquire error")
      return jsonify(False, -2, str(e))

  @route('/datapool/<dpname>/get/<int:index>', methods=['GET'])
  def getDatapoolByIndex (self, dpname, index):
    try:
      data = self._db.getDatapoolValue(dpname, index)
      self._log.info("Retrieved (%s:%d) by index" % (dpname, index))
      return jsonify(data)
    except LabserveError as e:
      return jsonify(False, e.code, e.msg)
    except Exception as e:
      return jsonify(False, -2, str(e))

  @route('/datapools', methods=['GET'])
  def getDatapoolsMetadata (self):
    try:
      retl = self._db.data_pools.keys()
      return jsonify(retl)
    except LabserveError as e:
      return jsonify(False, e.code, e.msg)
    except Exception as e:
      return jsonify(False, -2, str(e))

  #############
  ### Exams ###
  #############

  @route('/exam/<examid>/user/<user_urn>', methods=['POST'])
  def postExamAnswers (self, examid, user_urn): # pylint: disable=R0201
    if not request.json:
      return jsonify(None, -1, "Missing json POST data")
    ddir = "%s/exams/%s/%s" % (self._db._acctdir, examid, user_urn)
    if not os.path.exists(ddir):
      os.makedirs(ddir, self._db._accounting_umask)

    dpath = "%s/answer.json.%f" % (ddir, time.time())
    with open(dpath, "w+") as f:
      f.write(json.dumps(request.json))
    return jsonify(True)

  @route('/exam/<euuid>/user/<user_urn>/attempts/', methods=['GET'])
  def getExamAttempts (self, euuid, user_urn):
    result = self.generateAttempts(euuid, user_urn, exam=True)
    result = [(x,y) for x,y,_ in result]
    return jsonify(result)

  @route('/exam/<euuid>/user/<user_urn>/attempt/<int:index>', methods=['GET'])
  def getExamAttemptByIndex (self, euuid, user_urn, index):
    result = self.generateAttempts(euuid, user_urn, exam=True)
    attempt = result[index]
    with open(attempt[2], "r") as f:
      return jsonify(f.read())

  ###########################
  ### Credential Handling ###
  ###########################

  @route('/user/<user>/courses', methods=["GET"])
  def userCourses (self, user):
    # TODO: We can only have one shortuid for the user across multiple courses
    ret = {"courses" : [], "shortuid" : None}
    uclist = []
    shortuid = None
    lsurl = None
    for course in self._db._courses.values():
      if course.hasUser(user):
        uclist.append(course.name)
        shortuid = course.getShortUID(user)
        lsurl = course.labserve_url
    if not uclist:
      self._log.warning("No courses found for requesting user '%s'" % (user))
    # TODO: report lsurl per course
    ret["courses"] = uclist
    ret["shortuid"] = shortuid
    ret["labserve-url"] = lsurl
    return jsonify(ret)

  #######################
  ### Accounting Data ###
  #######################

  @route('/module/<uuid>/extra/<user_urn>', methods=["POST"])
  def postExtra (self, uuid, user_urn): # pylint: disable=R0201
    if not request.json:
      return jsonify(None, -1, "Missing json POST data")
    mdir = "%s/%s" % (self._db._acctdir, uuid)
    if not os.path.exists(mdir):
      os.makedirs(mdir, self._db._accounting_umask)

    mpath = "%s/%s-extra.%f" % (mdir, user_urn, time.time())
    with open(mpath, "w+") as f:
      f.write(json.dumps(request.json))
    return jsonify(True)

  @route('/module/<uuid>/manifest/<user_urn>', methods=["POST"])
  def postManifest (self, uuid, user_urn):
    # We really need a forensic accounting manager, but it will live here for now
    if not request.json:
      return jsonify(None, -1, "Missing json POST data")
    o = self.validate(request.json, [("manifest", (unicode, str))])
    mdir = "%s/%s" % (self._db._acctdir, uuid)
    if not os.path.exists(mdir):
      os.makedirs(mdir, self._db._accounting_umask)

    mpath = "%s/%s-manifest.%f" % (mdir, user_urn, time.time())
    with open(mpath, "w+") as f:
      f.write(o["manifest"])
    return jsonify(True)

  @route('/module/<uuid>/status/<user_urn>', methods=["POST"])
  def postStatus (self, uuid, user_urn):
    if not request.json:
      return jsonify(None, -1, "Missing json POST data")
    o = self.validate(request.json, [("status", (unicode, str))])

    mdir = "%s/%s" % (self._db._acctdir, uuid)
    if not os.path.exists(mdir):
      os.makedirs(mdir, self._db._accounting_umask)

    mpath = "%s/%s-status.%f" % (mdir, user_urn, time.time())
    with open(mpath, "w+") as f:
      f.write(o["status"])
    return jsonify(True)


def setup (app, cfg):
  api = ModuleServer(app, cfg)
  return api
